////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//																							F A R M     -  Flair advanced rendering module
//
//
// FARM is a render engine based on the principles of physically based ray-tracing. It solves the light transport equation,
// taking into account the interaction of visible photons with objects by using Monte Carlo techniques to integrate over
// bi-directional scattering functions or analytical solutions of those. Parametrization of measured reflectance & transmission
// data is supported and included in a number of standard interaction models included in the system. The design of the extensible
// plugin architecture of materials & BxDFs is strongly based on the outstanding PBRT1 by Matt Pharr, Greg Humphreys & Wenzel Jakob.
//
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
//					Version 1.0			08/05/2020
//
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "viewer/d3layer.h"
#include "geometry/voxel.h"
#include "farm.h"


#include <iostream>
#include <string>

#include <unistd.h>

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CFARM::CFARM() : m_bConfigRead(false), m_bShadeData( false ), m_pVisLayer( nullptr ), m_pVoxelLayer(nullptr)

// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CFARM::~CFARM()

// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CFARM::Init( D3Layer* pVisLayer, VVoxel* pVoxelLayer,  Geometry* pGeometry )

// basic initialization
// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_pVisLayer = pVisLayer;
	m_pVoxelLayer = pVoxelLayer;

	// set up integrators
	m_SurfaceIntegrator.Init( this, pVisLayer, nullptr , pGeometry );
	m_VoxelIntegrator.Init( this, pVisLayer, pVoxelLayer, pGeometry );
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CFARM::Shade( GeometryEngine* pGeoEngine, Ray* pShadingRay, const int RenderPass, int Light0 )

// treat passed shading ray
// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color RetCol;

	// NOTE: In the future this could be a sum which also accommodates, global illumination, emission of light + a volume integrator for voxels
	// Li = direct illumination component
	RetCol = m_SurfaceIntegrator.Li( pGeoEngine, pShadingRay, RenderPass, Light0 );

	return RetCol;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CFARM::ShadeVoxel( GeometryEngine* pGeoEngine, Ray* pShadingRay, const int RenderPass, int Light0 )

// voxel shading
// [6/9/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color RetCol;
	Ray   MarchingRay( *pShadingRay );	// as we march the ray through the voxels we need to do it with a copy to avoid side-effects with USRBIN
																			// and other renderings

	MarchingRay.moveby( SMALL3D3 );
	RetCol = m_VoxelIntegrator.Li( pGeoEngine, &MarchingRay, RenderPass, Light0 );
	if( RetCol == Color::Black ) {
		pShadingRay->stop();
	}

	return RetCol;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool	CFARM::IsVoxelDrawnTransparent( Ray* pShadingRay )

// [6/10/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( pShadingRay->voxelreg < 0 )
		return false;

	return m_VoxelIntegrator.RayToTransparency( pShadingRay ) > Color::Black ? true : false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CFARM::GetVoxelTransparency( Ray* pShadingRay )

// [6/12/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( pShadingRay->voxelreg < 0 )
		return Color::Black;

	return m_VoxelIntegrator.RayToTransparency( pShadingRay );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CFARM::ReadConfigFile( const std::string& Path )

// Read the optional configuration file to set up the renderer
// [6/29/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	string FileName( Path );

	// FIXME: For the moment we get the path from a font that is loaded.
	//				Thus, we have to do some gymnastics to get the flair installation path. Ideally we should get the path of FLAIR here
	// NOTE:	With C++17 we would get the <filesystem> STL to do this but it's not yet in GCC7 :-(
	//				Therefore, we do it ourselves

	string Directory;
	const size_t LastSlashPos = FileName.rfind( "/fonts" );
	if( std::string::npos != LastSlashPos ) {
		Directory = FileName.substr( 0, LastSlashPos);
	}
	Directory += "/farm.ini";

	//std::clog << "FARM: reading config file = " << Directory;

	if( !m_ConfigParser.Load( Directory ) ) {
		//std::clog << " ... FAILED! Using defaults" << endl;
		return false;
	}

	m_bConfigRead = true;

	//std::clog << " ... OK!" << endl;

	//
	// initialize some basic FARM data members from the read config file
	m_bShadeData = m_ConfigParser.GetValue( "usrbin-shadow", 0 );

	// initialize integrators with optional config file options
	m_SurfaceIntegrator.InitFromConfig();
	m_VoxelIntegrator.InitFromConfig();

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CFARM::ShadeData( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// Return the shadow factor for data textures.
// Shadowing is only provided during the beauty pass & if configured in the farm.ini file. By default
// it is disabled.
//
// [6/29/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( !m_bShadeData)
		return 1.0;

	return m_SurfaceIntegrator.GetLightIntensityAtHit( pEngine, pRay, RenderPass, Light0 );
}



// vim:ts=2
