////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "farmsurfaceintegrator.h"
#include "geometry/engine.h"
#include "viewer/d3layer.h"
#include "viewer/viewer.h"
#include "math/randsampler.h"
#include "rendermaterial.h"

#include <string>

using namespace std;

CSurfaceIntegrator::CSurfaceIntegrator() : CBaseIntegrator(), m_AOCMinIntensity( 0.5), m_AOCMaxIntensity(1.0), m_AOCMaxDist(100), m_NrOfAOCSamples(20), m_MaxAOCDepth(0),
m_bUseLightMap(false), m_pLightMap(nullptr), m_NrOfLightMapSamples(32)
{
  SetupBxDFLUT();

	// TEST: under development
	m_bUseLightMap = false;
}

CSurfaceIntegrator::~CSurfaceIntegrator( void )
{
   for( size_t i = 0; i < m_BxDFLUT.size(); ++i )
     delete m_BxDFLUT[i];

   m_BxDFLUT.clear();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  CSurfaceIntegrator::SetupBxDFLUT()

// Create the look-up table of all availabe BxDF's.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_BxDFLUT.push_back( new CMatPhong( this ) );
	m_BxDFLUT.push_back( new CMatPlastic( this, 0.1) );                   // standard roughness = 1.0 - very broad highlight
	m_BxDFLUT.push_back( new CMatPlastic( this, 5.0f ) );             // 5 = higher roughness -> sharper highlight
	m_BxDFLUT.push_back( new CMatPlastic( this, 15.f ) );
	m_BxDFLUT.push_back( new CMatShinyMetal( this, 0.1 ) );                // standard roughness = 1.0 - very broad highlight
	m_BxDFLUT.push_back( new CMatShinyMetal( this, 5.0f ) );        // 5.0 = higher roughness -> sharper highlight
	m_BxDFLUT.push_back( new CMatShinyMetal( this, 15.0f ) );
	m_BxDFLUT.push_back( new CMatBrushedMetal( this ) );
	m_BxDFLUT.push_back( new CMatFresnelBlend( this ) );              // Fresnel blended substrate with microfacet distributions
	m_BxDFLUT.push_back( new CMatFresnelBlend( this, true ) );				// Fresnel blended substrate with perfect specular reflection
	m_BxDFLUT.push_back( new CMatBluePaint( this ) );
	m_BxDFLUT.push_back( new CMatSkin( this ) );
	m_BxDFLUT.push_back( new CMatBlinn( this ) );
	m_BxDFLUT.push_back( new CMatMinnaert( this ) );
	m_BxDFLUT.push_back( new CMatToon( this ) );
	m_BxDFLUT.push_back( new CMatTranslucent( this ) );
	m_BxDFLUT.push_back( new CMatGlass( this, 1.40f ) );
	m_BxDFLUT.push_back( new CMatClearGlass( this ) );								// plexi-glass
	m_BxDFLUT.push_back( new CMatDispersiveClearGlass( this, 30.0 ) );
	m_BxDFLUT.push_back( new CMatDispersiveClearGlass( this, 80.0 ) );

	//m_BxDFLUT.push_back( new CMatFresnelBlend( this, 4, 50) );              // Fresnel blended substrate with highly anisotropic microfacet distributions

	/*
	m_BxDFLUT.push_back( new CMatOrenNayar( this ) );
	m_BxDFLUT.push_back( new CMatOrenNayar( this, 45 ) );			// sigma of 45 degrees,
  m_BxDFLUT.push_back( new CMatClay( this ) );
  m_BxDFLUT.push_back( new CMatFelt( this ) );
  m_BxDFLUT.push_back( new CMatPrimer( this ) );
  */
  m_pBxDF = m_BxDFLUT[0];
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// main integrator routine returning the luminance for the passed ray & rendering pass
// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Point	 Hit;
	const	 Material* pMaterial = nullptr;
	double ambiOcc = 1.0;
	Color	 ShadingColor;
	Ray		 lightRay;

	// on end of ray, error or blackbody
	if( !m_pVisLayer->nextIntersection( pEngine, pRay ) ) {
		if( m_bUseLightMap && pRay->depth > 0 ) {							// take care of env. reflection
			return m_pLightMap->LightEmission( pRay->dir() ) * 1.0;
		}
		else {
			return ShadingColor;	// on end of ray, error or blackbody
		}
	}

		// Get region/voxel color
	if( pRay->voxelreg > 0 ) {
		ShadingColor.set( m_pVisLayer->viewer.voxel().roiColor( pRay->voxelreg ) );
	}
	else {
		ShadingColor.set( pRay->hitRegion()->color() );			// region color
		pMaterial = pRay->hitRegion()->region()->material();
	}

		// If we are in the start of the plane do not apply any shadow
	if( pRay->start() ) return ShadingColor;

//	if( pMaterial && pMaterial->fuzz() > 0.0 ) {
//		pRay->normal += pMaterial->fuzz() *
//			Vector( pEngine->random( -1.0, 1.0 ), pEngine->random( -1.0, 1.0 ), pEngine->random( -1.0, 1.0 ) );
//	}

	//
	// calculate light intensity
	//
	double totalIntensity = ShadeHitPoint( pEngine, pRay, RenderPass, Light0 );

#if 0
	int log2depth;
	frexp( hit.z, &log2depth );
#endif
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Environmental infinite light maps
	//
	if( m_bUseLightMap && pMaterial ) {
		ShadingColor += Li_LightMap( pEngine, pRay) ;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Ambient occlusion shading
	// CT 10/7/19
	if( m_pVisLayer->shadows and m_NrOfAOCSamples>0 &&
			pRay->depth <= m_MaxAOCDepth ) {
		// NOTE: To avoid artifacts the AOC distance should be reasonably small
		// (configurable?), false = no AOC sample scrambling
		ambiOcc = CalcAmbientOcclusion( pEngine, pRay,
																		m_NrOfAOCSamples,
																		m_AOCMaxDist,
																		m_AOCMinIntensity,
																		m_AOCMaxIntensity,
																		true );

		// ShadingColor.set(ambiOcc, ambiOcc, ambiOcc); // debug to get only AOC pass
		ShadingColor *= totalIntensity + (float)m_pVisLayer->ambient / 255.0;
		ShadingColor *= ambiOcc;
	}
	else	// standard shading
		ShadingColor *= (totalIntensity + (float)m_pVisLayer->ambient / 255.0);

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Reflection & refraction
	//
	if(	pRay->depth < pRay->max_depth &&
			pMaterial != nullptr ) {

		int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;

		// reflection
		if( pMaterial->specular() > 0.0 && m_BxDFLUT[BxDFIndex]->MatchesType( CBxDFBase::BSDF_REFLECTION ) ) {
			ShadingColor += Li_Reflection( pEngine, pRay, RenderPass, Light0 );
		}

		// refraction & transmission
		if( !pRay->hitRegion()->notTransparent() ) {
			if( m_BxDFLUT[BxDFIndex]->MatchesType( CBxDFBase::BSDF_TRANSMISSION ) ) {
				ShadingColor += Li_Transmission( pEngine, pRay, RenderPass, Light0 );
			}
		}
	}

	ShadingColor.clamp();

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CSurfaceIntegrator::CalcAmbientOcclusion( GeometryEngine* eng, Ray* pixelRay,
																								 int nrOfSamples, const double cMaxDist, const double cMinIntensity,
																								 const double cMaxIntensity, bool bScramble /* = false */ )


//	ambient occlusion integrator
// [6/3/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Point    hit;
	Vector   sampleDir, bodyNormal;
	Ray      ambiOccRay;
	float    nonShadowedSamples = 0.f, shadowedSamples = 0.f, transparency = 0.f;
	double   randNr1, randNr2;

	int      scramble1 = bScramble ? eng->random.integer() : 0;
	int      scramble2 = bScramble ? eng->random.integer() : 0;

	// Find body normal
	hit = pixelRay->segment( pixelRay->n ).hit( -SMALL3D3 );
	bodyNormal = pixelRay->normal;

	// check that the normal is opposite to the direction
	if( bodyNormal * pixelRay->segment( pixelRay->n ).dir > 0.0 )
		bodyNormal.negate();

	// ATTENTION: This is still experimental and not yet ready !!!!
	// The problem is that this method is already spawned from pixel-based threads.
	// Parallelizing requires that each sub-thread that we spawn here has its own engine object
	// to prevent race conditions. However, cloning the engines on the fly becomes terribly slow
	// and we lose all benefit. Sub-pools might be an option but first tests were not not
	// showing significant benefits as we start to carry around lots of baggage. Maybe we could
	// make the geometry engines thread safe by implementing mutex locks in the tracer?
	//
	// CT 19/8/2019
#if 0
#ifdef THREAD
	///////////////////////////////////////////////////////////////////////////////////////////
	// PARALLEL ambient occlusion sampling
	//
	// define task

	// TEST: clone kernel threaded engines --> this becomes extremely slow!!!!!
	// Can we go for sub-pools of threaded engines??
	// FIXME: Ideally the engines would be thread safe so we would not need to pool them

	nrOfSamples = kernel.nthreads();
	GeometryEngine aocEngines[8];

	for( int i = 0; i < 8; i++ )
		aocEngines[i] = *eng;

	//-----
	auto ambiOccSampler = [] ( int i, GeometryEngine& geoEngine, const Vector& curBodyNormal,
														 const Point& hitPoint, const double maxAmbiDist, VZone& prevZone ) {
		Ray    curRay;
		double retVal = 0.0;
		Vector rayDir = CBaseSampler::uniformSampleHemisphere( curBodyNormal,
																													 VanDerCorput( i, 395732523 ),
																													 Sobol2( i, 146146721 ) );

		RaySegment new_segment( hitPoint, rayDir, &prevZone );
		new_segment.tmin = SMALL3D3;
		new_segment.tmax = maxAmbiDist * 2.0;
		curRay.push( new_segment );

		int oldBodyCheckId = geoEngine.incBodyCheckId();

		// return false = intersection occurred!
		if( !geoEngine.intersectRay( &curRay, false ) ) {
			VZone* pZone = curRay.hitZone();
			if( pZone != nullptr &&
				(pZone->region()->type() == GRegion::Type::NORMAL ||
					pZone->region()->type() == GRegion::Type::VOXEL) ) {
				geoEngine.bodyCheckId( oldBodyCheckId );
				// TODO: correctly account for transparent objects
				retVal = (curRay.T() <= maxAmbiDist) ? 0.0 : 1.0;
			}
			else
				retVal = 1.0;
		}
		else
			retVal = 1.0;
		geoEngine.bodyCheckId( oldBodyCheckId );
		return retVal;
	}; // ambiOccSampler

	// set up parallel execution of the task
	if( kernel.nthreads() ) {
		if( kernel.nBodies() ) {
			std::future<double> results[64];
			if( nrOfSamples > 64 || nrOfSamples > kernel.nthreads() ) {
				nrOfSamples = std::min( 64, kernel.nthreads() );
				std::cerr << "AmbiOcc: WARNING: too many samples/pixel. Limiting to "
					<< nrOfSamples << "!" << std::endl;
			}

			// start threads
			for( int t = 0; t < nrOfSamples; ++t )
				results[t] = std::async( std::launch::async | std::launch::deferred,
																 ambiOccSampler,
																 t, std::ref( aocEngines[t] ),
																 bodyNormal, hit,
																 cMaxDist,
																 std::ref( *pixelRay->prevZone() ) );
		}

		// gather results
		for( int t = 0; t < nrOfSamples; t++ )
			nonShadowedSamples += results[t].get();
	}
	else
#endif
#endif
	///////////////////////////////////////////////////////////////////////////////////////////
	// STANDARD sequential calculation
	{
		for( int i = 0; i < nrOfSamples; i++ ) {
			randNr1 = CBaseSampler::VanDerCorputDbl( i, scramble1 );
			randNr2 = CBaseSampler::Sobol2Dbl( i, scramble2 );

			sampleDir = CBaseSampler::uniformSampleHemisphere( bodyNormal, randNr1, randNr2 );

			ambiOccRay.use_clip = pixelRay->use_clip;
			RaySegment new_segment( hit, sampleDir, pixelRay->prevZone() );
			new_segment.tmin = SMALL3D3;
			new_segment.tmax = cMaxDist * 2.0;
			ambiOccRay.push( new_segment );

			int oldBodyCheckId = eng->incBodyCheckId();

			if( !eng->intersectRay( &ambiOccRay, false) ) {
				// return false = intersection occurred!
				VZone* pZone = ambiOccRay.hitZone();
				if( pZone != nullptr &&
					(pZone->region()->type() == GRegion::Type::NORMAL ||
						pZone->region()->type() == GRegion::Type::VOXEL) ) {
					eng->bodyCheckId( oldBodyCheckId );
					if( ambiOccRay.T() <= cMaxDist ) {
						// ATTENTION: the higRegion->transparent() function only returns true for FULLY transparent regions --> we need to check alpha!!
						transparency = ambiOccRay.hitRegion()->alpha() / 255.0;
						shadowedSamples += (1.0 - transparency );		// take transparency into account when calculating AOC
						continue;
					}
					else
						nonShadowedSamples++;
				}
				else
					nonShadowedSamples++;
			}
			else
				nonShadowedSamples++;
			eng->bodyCheckId( oldBodyCheckId );
			ambiOccRay.pop();
		}
	} // end of sequential

// #ifdef _DUMP
//	std::cerr << "AOC = " << (float)nonShadowedSamples / (float)nrOfSamples << std::endl;
// #endif

	// and account for transparency
	return cMinIntensity + (cMaxIntensity - cMinIntensity)
		* (nrOfSamples - shadowedSamples) / (float)nrOfSamples;


} // calcAmbientOcclusion


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_Reflection( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [4/28/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						ShadingColor, TmpColor;
	const Material* pMaterial = pRay->hitRegion()->region()->material();
	Vector					R, ReflNormal1, ReflNormal2;
	Ray							ReflRay;
	float						Pdf = 0.0;
	double					Rho = 0.0, Dot = 0.0;

	int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
	int NrOfSamples = (m_BxDFLUT[BxDFIndex]->RequiresMCSampling() || pMaterial->fuzz() > 0.0) ? std::min( 256, GetNrOfMCSamples() ) : 1;   // restrict the nr. of samples to 32. otherwise we might end up with extreme rendering times

	int Scramble1 = pEngine->random.integer();
	int Scramble2 = pEngine->random.integer();

	for( int i = 0; i < NrOfSamples; ++i ) {


		// sample reflection BxDF to obtain direction of incoming light + the corresponding rho
		Rho = m_BxDFLUT[BxDFIndex]->Sample_f( *pRay, ReflRay, Pdf, pRay->normal, *pMaterial, pEngine->random, CBxDFBase::BxDFType::BSDF_REFLECTION );

		// disable clipping bodies for reflection rays, to avoid visualization artifacts of background bleeding
		ReflRay.use_clip = ReflRay.clip_insight = false;

		// handle diffuse reflection if requested
		if( pMaterial->fuzz() ) {
			Vector distortedDir = CBaseSampler::uniformSampleCone( ReflRay.dir() , std::min( pMaterial->fuzz() / 10.0, 1.0 ), CBaseSampler::VanDerCorput( i, Scramble1 ), CBaseSampler::Sobol2( i, Scramble2 ) );
			ReflRay.segment().dir = distortedDir;
		}

		// calculate incoming reflected luminance
		Dot = fabs( pRay->normal.dot( ReflRay.dir() ) );
		if( Rho > 0 && Dot > 0 && Pdf > 0.0 ) {
			pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction
			ShadingColor += Li( pEngine, &ReflRay, RenderPass, Light0 ) * pMaterial->specular() * Dot * Rho / Pdf;	// NOTE: for MC sampled rays the PDF should include the factor 1/Pi already!
		}
	}

	// normalize
	ShadingColor /= (NrOfSamples * 1.0);      // calculate the total reflected contribution

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [5/4/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						RefractionColor, TransparencyColor( 1.0, 1.0, 1.0);
	const Material* pMaterial = pRay->hitRegion()->region()->material();
	Vector					R, ReflNormal1, ReflNormal2;
	Ray							RefractionRay;
	float						Pdf = 0.0;
	double					Rho = 0.0, Dot = 0.0;

	int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
	int NrOfSamples = (m_BxDFLUT[BxDFIndex]->RequiresMCSampling() || pMaterial->fuzz() > 0.0) ? std::min( 256, GetNrOfMCSamples() ) : 1;   // restrict the nr. of samples to 32. otherwise we might end up with extreme rendering times

	int Scramble1 = pEngine->random.integer();
	int Scramble2 = pEngine->random.integer();

	for( int i = 0; i < NrOfSamples; ++i ) {

		// sample refraction BxDF to obtain direction of incoming light + the corresponding rho
		Rho = m_BxDFLUT[BxDFIndex]->Sample_f( *pRay, RefractionRay, Pdf, pRay->normal, *pMaterial, pEngine->random, CBxDFBase::BxDFType::BSDF_TRANSMISSION );

		// disable clipping bodies for refraction rays, to avoid visualization artifacts of background bleeding
		RefractionRay.use_clip = RefractionRay.clip_insight = false;

		// handle diffuse transmission if requested
		if( pMaterial->fuzz() ) {
			Vector distortedDir = CBaseSampler::uniformSampleCone( RefractionRay.dir(), std::min( pMaterial->fuzz() / 10.0, 10.0 ), CBaseSampler::VanDerCorput( i, Scramble1 ), CBaseSampler::Sobol2( i, Scramble2 ) );
			RefractionRay.segment().dir = distortedDir;
		}

		// calculate incoming reflected luminance
		Dot = fabs( -pRay->normal.dot( RefractionRay.dir() ) );
		pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction
		RefractionColor += Li( pEngine, &RefractionRay, RenderPass, Light0 ) * Dot * Rho / Pdf;	// NOTE: for MC sampled rays the PDF should include the factor 1/Pi already!
	}

	// normalize
	RefractionColor /= (NrOfSamples * 1.0);      // calculate the total reflected contribution

	//
	// apply Beer's law
	//
	if( pRay->hitRegion()->alpha() != 0 ) {
		double MatDensity = -0.005 * (pRay->T());       // TODO: Get this parameter from the material
		Color Absorbance( pRay->hitRegion()->color() );
		Absorbance *= MatDensity;
		TransparencyColor.set( expf( Absorbance.red() ), expf( Absorbance.green() ), expf( Absorbance.blue() ) );
	}

	return RefractionColor * TransparencyColor * (pRay->hitRegion()->alpha() / 255.0) * (1.0 - pMaterial->specular());
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_LightMap( GeometryEngine* pEngine, Ray* pRay )

// Return the incoming irradiance from an environmental lighmap
//
// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color		ShadingColor, CurLightCol;
	int			NrOfSamples = std::min( 256, m_NrOfLightMapSamples );   // restrict the nr. of samples to 256. otherwise we might end up with extreme rendering times
	double	Rho = 1.0;
	float   Transmission = 1.0;
	Ray			LightRay( *pRay );
	Vector	LightDir;

	const Material* pMaterial = pRay->hitRegion()->region()->material();
	int							BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;

	// sample light map
	for( int i = 0; i < NrOfSamples; ++i ) {

		// TODO: replace by biased lightmap
		// sample incoming irradiance
		CurLightCol	= m_pLightMap->SampleLightMap( pEngine, i, *pRay, LightRay, Transmission );

		// calculate color
		Rho = m_BxDFLUT[BxDFIndex]->f( *pRay, LightRay, pRay->normal, *pMaterial, CurLightCol, 1.0 );
		double Dot = std::max( 0.0, pRay->normal.dot( LightRay.dir() ) );
		ShadingColor += CurLightCol * Rho * Dot * Transmission;
	}

	// normalize
	ShadingColor /= (NrOfSamples * 1.0);

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CSurfaceIntegrator::InitFromConfig()

// [8/19/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert( m_pRenderer );	// ensure the base-class Init() has been called before
	string FileName;

	// show config flags
	if( !m_pRenderer->GetConfig().Empty() ) {
#if _DEBUG>1
		m_pRenderer->GetConfig().Show();
#endif
  }

	// set up light maps
	m_bUseLightMap = m_pRenderer->GetConfig().GetValue<int>( "lightmap", 0 );
	if( m_bUseLightMap && !m_pLightMap ) {

		m_pLightMap = new CBiasedLightMap;
		FileName = m_pRenderer->GetConfig().GetValue<string>( "lightmap_file", "" );

		if( !m_pLightMap->LoadMap( FileName ) ) {
			std::cerr << "FARM ERROR: no image light found" << std::endl;
			m_bUseLightMap = false;
		}

		m_pLightMap->SetDistance( m_pRenderer->GetConfig().GetValue<double>( "lightmap_distance", 1.0 ) );		// NOTE: larger distances will blur out detailed reflections
		std::cout << "FARM: Lightmap initialized: " << FileName << " - " << m_pLightMap->GetDistance() << std::endl;

		// set up projection
		std::string LightMapProjection = m_pRenderer->GetConfig().GetValue<string>( "lightmap_projection", "xy" );
		if( LightMapProjection == "yz" )
			m_pLightMap->SetUVProjection( Vector( 0, 1, 0 ), Vector( 0, 0, 1 ) );
		else if( LightMapProjection == "zy" )
			m_pLightMap->SetUVProjection( Vector( 0, 0, 1 ), Vector( 0, 1, 0 ) );
		else if( LightMapProjection == "zx" )
			m_pLightMap->SetUVProjection( Vector( 0, 0, 1 ), Vector( 1, 0, 0 ) );
		else if( LightMapProjection == "xz" )
			m_pLightMap->SetUVProjection( Vector( 1, 0, 0 ), Vector( 0, 0, 1 ) );
		else if( LightMapProjection == "yx" )
			m_pLightMap->SetUVProjection( Vector( 0, 1, 0 ), Vector( 1, 0, 0 ) );
		else // default = xy
			m_pLightMap->SetUVProjection( Vector( 1, 0, 0 ), Vector( 0, 1, 0 ) );

		// set lightmap intensity
		m_pLightMap->SetIntensity( m_pRenderer->GetConfig().GetValue<double>( "lightmap_intensity", 1.0 ) );
	}

	// set up nr. of samples
	m_NrOfMCSamples				= m_pRenderer->GetConfig().GetValue<int>( "mc_samples", 32 );
	m_NrOfLightMapSamples = m_pRenderer->GetConfig().GetValue<int>( "lightmap_samples", 32 );

	// set up ambient occlusion
	m_AOCMinIntensity = m_pRenderer->GetConfig().GetValue<double>( "aoc_min", 0.5 );
	m_AOCMaxIntensity = m_pRenderer->GetConfig().GetValue<double>( "aoc_max", 1.0 );
	m_AOCMaxDist = m_pRenderer->GetConfig().GetValue<double>( "aoc_dist", 100.0 );
	m_MaxAOCDepth = m_pRenderer->GetConfig().GetValue<int>( "aoc_maxdepth", 0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CSurfaceIntegrator::ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [8/19/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( RenderPass );

	Point	 Hit;
	const	 Material* pMaterial = nullptr;
	double randNr[2];
	Ray		 lightRay;
	double Rho = 1.0;
	double pdf = 1.0;

	// All light rays should start from outside the object
	// Move a bit backwards
	// Should be bigger than what is inside distance
	Hit = pRay->hit( -SMALL3D3 );

	// Calculate illumination
	double totalIntensity = 0.0, curIntensity = 0.0;
	Vector dir, curLPos;

	for( int i = 0; i < Min( pRay->nlights, m_pGeometry->nlights ); i++ ) {
		const Light& L = m_pGeometry->light[Light0 + i];	// Geometry Light properties
		const VLight& VL = m_pVisLayer->light[i];				// Viewer light coordinate & direction
		double  dist, power, penumbraScaling = 0.0;

		const int nsamples = L.nsamples;
		//assert(InRange(0,nsamples,100));
		curIntensity = 0.0;

		// volumetric lights
		for( int lightSample = 0; lightSample < nsamples; ++lightSample ) {
			switch( L.type ) {
				case Light::Type::Sun:
					dir = -VL.dir;
					dist = INFINITE;
					power = L.power;
					if( L.shadow && pRay->shadow ) {
						double dot = m_pVisLayer->view().matrix( 0, 2 ) * dir.x +
							m_pVisLayer->view().matrix( 1, 2 ) * dir.y +
							m_pVisLayer->view().matrix( 2, 2 ) * dir.z;
						if( dot > SMALL3D3 && m_pVisLayer->view().projection != ViewPort::Projection::Perspective ) {
							dist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
											 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
											 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
								dot;
							if( dist < SMALL3D3 ) continue;
						}
					}
					break;

				case Light::Type::Omni:
				case Light::Type::Spot:
					// move the light position for volumetric light sources
					curLPos = VL.pos;
					if( lightSample > 0 ) {
						//pEngine->random.realqmcND( 2, randNr );
						randNr[0] = CBaseSampler::VanDerCorput( lightSample, pEngine->random.integer() );
						randNr[1] = CBaseSampler::Sobol2( lightSample, pEngine->random.integer() );
						dir = (VL.pos - Hit);
						dir.normalize();
						curLPos += CBaseSampler::uniformSampleDisk( dir, L.radius, randNr[0], randNr[1] );
					}
					dir = curLPos - Hit;
					dist = dir.normalize();

					if( dist < SMALL3D3 ) continue;
					if( dist > L.distance ) continue;
					if( L.falloff != Light::FallOff::Constant ) {
						// FIXME could be converted to switch ()
						if( L.falloff == Light::FallOff::Inverse )
							power = L.power * 100.0 / dist;
						else	// must be Light::FallOff::Inverse2
							power = L.power * 10000.0 / Sq( dist );
						if( power < 0.001 ) continue;
					}
					else
						power = L.power;

					if( L.shadow && pRay->shadow ) {
						double dot = m_pVisLayer->view().matrix( 0, 2 ) * dir.x +
							m_pVisLayer->view().matrix( 1, 2 ) * dir.y +
							m_pVisLayer->view().matrix( 2, 2 ) * dir.z;
						if( dot > SMALL3D3 ) {
							double pdist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
															 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
															 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
								dot;
							dist = Min( dist, pdist );
							if( dist < SMALL3D3 ) continue;
						}
					}
					power /= (double)nsamples;   // normalize the light power w.r.t. light samples
					break;

				default:
					continue;
			}

			// set up light ray & find intersection with light
			lightRay.use_clip = pRay->use_clip;
			RaySegment new_segment( Hit, dir, pRay->prevZone() );
			new_segment.tmin = SMALL3D3 * pRay->segment().tmin;
			new_segment.tmax = dist;
			lightRay.push( new_segment );

			double dot = dir * pRay->normal;
			if( dot > 0.00001 ) {
				if( L.shadow && pRay->shadow ) {
					int oldBodyCheckId = pEngine->incBodyCheckId();

					if( !pEngine->intersectRay( &lightRay, false ) ) {
						// Check if we hit a black hole or an error
						VZone* z = lightRay.hitZone();
						if( z != nullptr ) {
							if( (z->region()->type() == GRegion::Type::NORMAL ||
										z->region()->type() == GRegion::Type::VOXEL) ) {
								pEngine->bodyCheckId( oldBodyCheckId );

								// account for "light shadows" due to transparent objects
								double Transparency = z->region()->alpha() / 255.0;
								penumbraScaling += Transparency / (nsamples);		// take transparency into account when calculating AOC
								continue;
							}
							else
								if( z->region()->type() == GRegion::Type::BLACKHOLE )
									penumbraScaling = 1.0;
						}

////////////

					/* WIP: -- multi-step tracking for semi-opaque objects that cast light shadows
					if( !pEngine->intersectRay( &lightRay, true ) ) {
						// Check if we hit a black hole or an error
						VZone* z = lightRay.hitZone();
						if( z != nullptr ) {
							if( (z->region()->type() == GRegion::Type::NORMAL ||
										z->region()->type() == GRegion::Type::VOXEL) ) {

								// account for "light shadows" due to transparent objects
								double Transparency = z->region()->alpha() / 255.0;

								if( Transparency > 0.0 && Transparency < 1.0 ) {
									lightRay.moveby( SMALL3D4 * 10.0 );
									if( !pEngine->intersectRay( &lightRay, false ) ) {
										if( (z->region()->type() == GRegion::Type::NORMAL ||
													z->region()->type() == GRegion::Type::VOXEL) &&
												z->region()->type() != GRegion::Type::BLACKHOLE ) {
											penumbraScaling = 0.0;
										}
									}
								}

								pEngine->bodyCheckId( oldBodyCheckId );
								penumbraScaling += Transparency / (nsamples);		// take transparency into account when calculating AOC
								continue;
							}
							else
								if( z->region()->type() == GRegion::Type::BLACKHOLE )
									penumbraScaling = 1.0;
						}
						*/
///////////////////////////////
				}
					else {
						if( L.type == Light::Type::Sun )
							penumbraScaling = 1.0;	// no penumbra by the sun. we only use 1
																			// sample for the sunlight for reasons of efficiency
						else
							penumbraScaling += 1.0 / (double)nsamples;
					}
					pEngine->bodyCheckId( oldBodyCheckId );
				}
				else // no shadowing requested
					penumbraScaling = 1.0;
			}

			////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
			// Evaluate BxDF for current material
			//
			if( pMaterial ) {
				int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
				Rho = m_BxDFLUT[BxDFIndex]->f( *pRay, lightRay, pRay->normal, *pMaterial, Color( 1.0, 1.0, 1.0 ), power );
			}

			// take BxDF for current sample into account
			curIntensity += power * dot * Rho / pdf;

			// bail out early after 1 sample in case of sunlight
			if( L.type == Light::Type::Sun && lightSample == 0 )
				break;
		} // end of sampling for one light source

		totalIntensity += curIntensity * penumbraScaling;

	} // end of light loop

	return totalIntensity;
}
// vim:ts=2
