////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "matplastic.h"

CMatPlastic::CMatPlastic( CBaseIntegrator* pInt, const float cRoughness /* = 1.0 */ ) : CPhysMatInterface( pInt, static_cast<CBxDFBase::BxDFType>( CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_SPECULAR | CBxDFBase::BSDF_REFLECTION | CBxDFBase::BSDF_TRANSMISSION ), true /* true = MC sampling required for the microfacet distributions for example */  )
{
  // roughness higher values indicate a higher roughness -> sharp specular highlight, lower values -> broad specular highlight

  m_pBxDFDiffuse    = new CBxDFPhong;
  m_pFresnelDiel    = new CFresnelPreDefDiel( 1.5, 1.0);
  m_pBlinnMicroDist = new CBlinnMicroDist( cRoughness );
  m_pBxDFSpecular   = new CBxDFMicrofacet( m_pFresnelDiel, m_pBlinnMicroDist );
  m_pBxDFSpecReflection = new CBxDFSpecReflection( m_pFresnelDiel );
  m_pBxDFSpecTransmission = new CBxDFSpecTransmission;
}

CMatPlastic::~CMatPlastic(void)
{
  delete m_pFresnelDiel;
  delete m_pBlinnMicroDist;
  delete m_pBxDFDiffuse;
  delete m_pBxDFSpecular;
  delete m_pBxDFSpecReflection;
  delete m_pBxDFSpecTransmission;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CMatPlastic::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  double Diffuse, Specular;

  m_pBlinnMicroDist->SetExp( std::max( 1.0, RenderMat.shine() ) );
  Diffuse   = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );
  Specular  = m_pBxDFSpecular->f( Out, In, Normal, RenderMat, LightColor, Shade );

  return Diffuse * (1.0 - RenderMat.specular() ) + Specular * RenderMat.specular();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CMatPlastic::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  if( Type == CBxDFBase::BSDF_REFLECTION )
    //return m_pBxDFSpecReflection->Sample_f( Out, In, Pdf, Normal, RenderMat, Sampler, Type, LightColor, Shade );
    return m_pBxDFSpecular->Sample_f( Out, In, Pdf, Normal, RenderMat, Sampler, Type, LightColor, Shade );
  else if( Type == CBxDFBase::BSDF_TRANSMISSION )
    return m_pBxDFSpecTransmission->Sample_f( Out, In, Pdf, Normal, RenderMat, Sampler, Type, LightColor, Shade );

  return CPhysMatInterface::Sample_f( Out, In, Pdf, Normal, RenderMat, Sampler, Type, LightColor, Shade );
}
// vim:ts=2
