/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 *
 * Author:	Christian.Theis@cern.ch
 *
 * Date: 18-Nov-2020
 *
 *
 * Code to construct a hierarchical bounding volume based GRegion object that can be queried efficiently in which actual region a point is located.
 *
 * This version uses a relatively simple midpoint splitting strategy choosing the axis where the bounding box is largest to minimize overlaps of sub-volumes. This gives reasonable trees but they
 * clearly can be optimized. One could think about calculating Morton codes for each volume in the current bounding volume and order them by the Morton codes. The splitting point could then be determined
 * by looking at the index where the highest bit starts to flip for the first time. This is repeated recursively and we would automatically have a quite well balanced tree. This could also be combined
 * with cost functions based on SAH like Pantaleoni et al. did for Weta's Pantaray. To be investigated.
 *
 */

#include "bvh.h"

#include "geometry/ray.h"

#include <vector>
#include <cassert>


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, std::vector<VRegion*> RegionList, size_t  NrOfNodes, bool bVerbose /* = false */, int LogLevel /*= 0*/ ) :
                                              m_pLeftLeaf(nullptr), m_pRightLeaf(nullptr), m_pGeoRegion(nullptr), m_NrOfLeaves(0), m_bVerbose( bVerbose ), m_LogLevel( LogLevel ),
                                              m_NodeList( RegionList ), m_pEngine( pEngine )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector CurBoxMin, CurBoxMax, NextBoxMin, NextBoxMax, TmpBoxMin, TmpBoxMax;
  double Pivot;

  m_bVerbose && std::cerr << "BVH: starting construction for " << NrOfNodes << std::endl;
  if( NrOfNodes < 1 )
    // throw std::runtime_error( "BVH: insufficient geometry regions to build BV tree" );
    return;

  if( NrOfNodes == 1 ) {
    *this = CBoundingVolHierarchy( m_pEngine, m_NodeList[0] );
  }

  if( NrOfNodes >= 2 ) {
    // Obtain the centroid of the first bounding box that will be used as a pivot for the qsplit routine
    BBox CurBBox = m_NodeList[0]->region()->bbox();
    for( size_t i = 1; i < NrOfNodes; ++i ) {
      BBox NextBBox = m_NodeList[i]->region()->bbox();
      CurBBox = CurBBox | NextBBox;
    }

    // set up the bounding box of the whole union
    m_BoundingBox = CurBBox;

//	  Pivot = CurBBox.centroid( cXAxis );
//		size_t MidPoint = QSplit( m_NodeList.data(), NrOfNodes, Pivot, cXAxis );

    // split along the axis which has most coverage and thus, least probability of overlaps
    int Axis = CurBBox.size().dominant();
    Pivot = CurBBox.centroid( Axis );
    size_t MidPoint = QSplit( m_NodeList.data(), NrOfNodes, Pivot, Axis );

    // Create a new bounding volume & define the next axis as the new split axis
    m_pLeftLeaf = BuildBranch( m_NodeList.data(), MidPoint );
    m_pRightLeaf = BuildBranch( &m_NodeList.data()[MidPoint], NrOfNodes - MidPoint);
    m_pGeoRegion = nullptr;
  }

  m_bVerbose && std::cerr << "BVH: Total nr of tree leaves = " << m_NrOfLeaves << std::endl;
  m_bVerbose && std::cerr << "BVH: BBox (" << m_BoundingBox.low() << ") -> (" << m_BoundingBox.high() << ")" << std::endl;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, VRegion* pGeoRegion ) : m_pLeftLeaf( nullptr), m_pRightLeaf( nullptr),
                                              m_pGeoRegion( pGeoRegion), m_pEngine( pEngine)

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_BoundingBox = m_pGeoRegion->region()->bbox();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2 ) :
                                              m_pLeftLeaf(pNode1), m_pRightLeaf( pNode2), m_pGeoRegion(nullptr), m_pEngine( pEngine )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector Box1Min, Box2Min, Box1Max, Box2Max;

  if( m_pRightLeaf ) {
    m_BoundingBox = pNode1->bbox() | pNode2->bbox();
  }
  else
    m_BoundingBox = pNode1->bbox();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2, const BBox& CurBBox ) :
                                              m_pGeoRegion( nullptr ), m_pEngine( pEngine )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_pLeftLeaf  = pNode1;
  m_pRightLeaf = pNode2;
  m_BoundingBox = CurBBox;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy* CBoundingVolHierarchy::BuildBranch( VRegion** ppNodeList, size_t NrOfSubBranchNodes, int Axis /* = cXAxis */ )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  BBox   CurBBox, NextBBox;
  double Pivot;

  m_NrOfLeaves++;

  if( NrOfSubBranchNodes == 1)
    return new CBoundingVolHierarchy( m_pEngine, ppNodeList[0] );

  // Obtain the centroid of the bounding box that will be used as a pivot for the qsplit routine
  CurBBox = ppNodeList[0]->region()->bbox();
  for( size_t i = 1; i < NrOfSubBranchNodes; ++i ) {
    NextBBox = ppNodeList[i]->region()->bbox();
    CurBBox = CurBBox | NextBBox;
  }

  // FIXME
  // determine the longest extension of the bounding box which will yield the splitting axis (= axis where we have the lowest probability to have overlapping bodies)
  Axis  = CurBBox.size().dominant();

  Pivot = CurBBox.centroid( Axis );

  // let's split according to the respective axis
  size_t  MidPoint = QSplit( ppNodeList, NrOfSubBranchNodes, Pivot, Axis);

  if( m_bVerbose && m_LogLevel >= 1 ) {
    std::cerr << "BVH: Splitting axis = " << Axis << ", pivot = " << Pivot << " splitting " << MidPoint << "/" << NrOfSubBranchNodes - MidPoint << std::endl;
  }

  // assemble a new bounding volume while iterating through all axes
  //VRegion* pLeftLeaf  = BuildBranch( ppNodeList, MidPoint, (Axis + 1) % 3 );
  //VRegion* pRightLeaf = BuildBranch( &ppNodeList[MidPoint], NrOfSubBranchNodes - MidPoint, (Axis + 1) % 3 );

  // assemble a new bounding volume based on axis with the largest coverage
  CBoundingVolHierarchy* pLeftLeaf = BuildBranch( ppNodeList, MidPoint, Axis );
  CBoundingVolHierarchy* pRightLeaf = BuildBranch( &ppNodeList[MidPoint], NrOfSubBranchNodes - MidPoint, Axis );


  return new CBoundingVolHierarchy( m_pEngine, pLeftLeaf, pRightLeaf, CurBBox );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t  CBoundingVolHierarchy::QSplit( VRegion** ppNodeList, size_t NrOfNodes, double PivotVal, int Axis)

// determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid
// coordinate < PivotVal with respect to a specific axis.
// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector   BoxMin, BoxMax;
  double   Centroid = 0.0;
  size_t   RetVal  = 0;

  for( size_t i = 0; i < NrOfNodes; ++i) {

    BBox CurBBox = ppNodeList[i]->region()->bbox();

    Centroid = CurBBox.centroid( Axis );
    if( Centroid < PivotVal ) {
      // iter_swap( ppNodeList.begin() + i, ppNodeList.begin() + RetVal );
      std::swap( ppNodeList[i], ppNodeList[RetVal] );
      RetVal++;
    }
  }

  if( RetVal == 0 || RetVal == NrOfNodes )
    RetVal = NrOfNodes / 2;

  return RetVal;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::Contains( const Vector& CurPoint, const Vector& CurDir, double eps )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  VZone* pLeftHitZone = nullptr;
	VZone* pRightHitZone = nullptr;

  // check if we have are region primitive, otherwise branch further
  if( m_pGeoRegion ) {
    // FIXME: it might be easier and certainly more performant if the GZone object that is returned from GRegion*
    //        would have a pointer to its VZone* that needs to be returned for visualization.
    return m_pEngine->getRegion( m_pGeoRegion->id() )->inside( m_pEngine, CurPoint, CurDir, eps );
  }

  // branch into further BHVs
	if( !m_BoundingBox.insideInclusive( CurPoint, eps ) )
		return nullptr;

  if( m_pLeftLeaf ) {
    pLeftHitZone = static_cast<CBoundingVolHierarchy*>(m_pLeftLeaf)->Contains( CurPoint, CurDir, eps );
  }

  if( m_pRightLeaf  ) {
    pRightHitZone = static_cast<CBoundingVolHierarchy*>(m_pRightLeaf)->Contains( CurPoint, CurDir, eps );
  }

  return pLeftHitZone ? pLeftHitZone : pRightHitZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::inside( const Vector& CurPoint, const Vector& CurDir, double eps /* = SMALL3D4 */ )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  if( m_NrOfLeaves == 0 )
    return nullptr;

  return Contains( CurPoint, CurDir, eps );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBoundingVolHierarchy::PrintTree()

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  std::cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
  std::cerr << "|                 Bounding volume hierarchies              |" << std::endl;
	std::cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
  std::cerr << " Nr of nodes = " << m_NodeList.size() << ", nr of leaves = " << m_NrOfLeaves << std::endl << std::endl;

	if( m_pLeftLeaf ) {
		m_pLeftLeaf->PrintLeaf();
	}

	if( m_pRightLeaf ) {
      m_pRightLeaf->PrintLeaf();
	}

	if( m_pGeoRegion ) {
		std::cerr << "--> #" << 0 << " - L-GEO: " << m_pGeoRegion->name() << " bmin =" << m_pGeoRegion->region()->bbox().low() << ", bmax = " << m_pGeoRegion->region()->bbox().high() << std::endl;
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBoundingVolHierarchy::PrintLeaf( int Level /*= 0*/ )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::cerr << "+";

  Indent( Level );
  std::cerr << "--> #" << Level << "  BB " << " bmin =" << m_BoundingBox.low() << ", bmax = " << m_BoundingBox.high() << std::endl;

  if( m_pLeftLeaf ) {
		Indent( Level + 1 );
    std::cerr << "|" << std::endl;
    m_pLeftLeaf->PrintLeaf( Level + 1 );
  }

  if( m_pRightLeaf ) {
    Indent( Level + 1 );
		std::cerr << "|" << std::endl;
    m_pRightLeaf->PrintLeaf( Level + 1 );
  }

	if( m_pGeoRegion ) {
		Indent( Level + 1 );
		std::cerr << "--> #" << Level + 1 << " - L-GEO: " << m_pGeoRegion->name() << " bmin =" << m_pGeoRegion->region()->bbox().low() << ", bmax = " << m_pGeoRegion->region()->bbox().high() << std::endl;
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersectRay( Ray* pCurRay )

// [11/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tmin = pCurRay->T();
  Vector InvDir( pCurRay->dir().x != 0.0 ? 1.0 / pCurRay->dir().x : 0.0,
                 pCurRay->dir().y != 0.0 ? 1.0 / pCurRay->dir().y : 0.0,
                 pCurRay->dir().z != 0.0 ? 1.0 / pCurRay->dir().z : 0.0 );

  // check if the current ray already has a potential intersection candidate

	VZone* pCurZone = pCurRay->segment().zone;
  if( pCurZone && tmin > SMALL3D4) {
    if( pCurZone->bbox().isValid() and
        const_cast<BBox&>(pCurZone->bbox()).intersectRay( pCurRay->pos(), InvDir, tmin, pCurRay->segment().tmax ) and
        pCurZone->inside( m_pEngine, pCurRay->pos().x, pCurRay->pos().y, pCurRay->pos().z, pCurRay->dir().x, pCurRay->dir().y, pCurRay->dir().z, tmin ) ) {
      return pCurZone;
    }
  }

  return intersectRay( pCurRay, pCurRay->pos(), InvDir, tmin, pCurRay->segment().tmax );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersectRay( Ray* pCurRay, const Vector& Pos, const Vector& InvDir, double& tmin, double tmax )

// optimized intersection that avoids pre-calculating constants like the inverse direction
// [12/2/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminLeft = 0, tmaxLeft = 0, tminRight = 0, tmaxRight = 0;
	VZone* pLeftZone = nullptr;
	VZone* pRightZone = nullptr;

	if( m_BoundingBox.intersectRay( Pos, InvDir, tmin, tmax ) )
		return nullptr;

	// check if we have an actual region to intersect
	if( m_pGeoRegion ) {
		VZone* pZone = nullptr;
		double tminOld = 0;

		// loop over zones to find the closest hit, if any
		for( int i = 0; m_pGeoRegion->size(); ++i ) {
			auto pCurBody = m_pGeoRegion->zone( i )->distance( m_pEngine, Pos, pCurRay->dir(), &tmin, tmax );
			if( pCurBody && tmin < tminOld ) {
				pZone = m_pGeoRegion->zone( i );
				tminOld = tmin;
			}
		}

		if( pZone ) {
			pCurRay->segment().tmin = tminOld;
			pCurRay->segment().zone = pZone;
      tmin = tminOld;
		}

		return pZone;
	}

	// branch into tree
	if( m_pLeftLeaf ) {
		pLeftZone = m_pLeftLeaf->intersectRay( pCurRay, Pos, InvDir, tminLeft, tmaxLeft );
	}

	if( m_pRightLeaf ) {
		pRightZone = m_pRightLeaf->intersectRay( pCurRay, Pos, InvDir, tminRight, tmaxRight );
	}

	// return the closer hit
	if( pRightZone && tminRight < tminLeft ) {
		pCurRay->segment().zone = pRightZone;
		pCurRay->segment().tmin = tminRight;
		return pRightZone;
	}

	pCurRay->segment().zone = pLeftZone;
	pCurRay->segment().tmin = tminLeft;

	return pLeftZone;
}

/*
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CBoundingVolHierarchy::Intersect( CRay& Ray )

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  bool bIntersectLeftLeaf = false;
  bool bIntersectRightLeaf = false;
  CRay BoundsRay( Ray ), LeftLeafRay( Ray ), RightLeafRay( Ray );
  SGdouble tLeft = 0.0, tRight = 0.0;
  CHierarchyNode* pLeftHitNode = nullptr;
  CHierarchyNode* pRightHitNode = nullptr;
  CHierarchyNode* pHitPrimitive = nullptr;

  if( !IntersectBounds( BoundsRay ) )
    return false;

  if( m_pLeftLeaf ) {
    bIntersectLeftLeaf  = m_pLeftLeaf->Intersect( LeftLeafRay );
    tLeft  = LeftLeafRay.GetMint();
  }
  if( m_pRightLeaf ) {
    bIntersectRightLeaf = m_pRightLeaf->Intersect( RightLeafRay );
    tRight = RightLeafRay.GetMint();
  }

  // determine the closest hit
  if( bIntersectLeftLeaf  ) {
    Ray.SetMint( tLeft );
    if( (pHitPrimitive = LeftLeafRay.GetHitPrimitive()) ) {
      Ray.SetHitPrimitive( pHitPrimitive );
      Ray.SetHitNormal( LeftLeafRay.GetHitNormal() );
    }
  }

  if( bIntersectRightLeaf && tRight < tLeft ) {
    Ray.SetMint( tRight );
    if( (pHitPrimitive = RightLeafRay.GetHitPrimitive()) ) {
      Ray.SetHitPrimitive( pHitPrimitive );
      Ray.SetHitNormal( RightLeafRay.GetHitNormal() );
    }
  }

  return (bIntersectLeftLeaf || bIntersectRightLeaf );
}
*/

/*
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CBoundingVolHierarchy::IntersectBounds( CRay& Ray )

// A BVH node has no real geometry. Thus, it has to reset the hit primitive
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  bool bRetVal = CHierarchyNode::IntersectBounds( Ray );
  Ray.SetHitPrimitive( NULL );

  return bRetVal;
}
*/
// vim: ts=2
