/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "cylbody.h"

#include "mesh.h"
#include "line.h"

// =============================== GCylinderBase ===============================
int GCylinderBase::N_CYLINDER = 16;

/** make cone quadrics from 3 vectors and 2 positions */
void GCylinderBase::makeConeQuads()
{
	Matrix4 M, Minv, T;

	Minv.make(X,Y,Z);
	Minv.inverse();
	T.translate(P);
	M.multiply(T, Minv);
	Minv.inverse(M);

	// transform first quad
	_Q[0].transform(Minv, M);
	//_Q[0].normalize();

	// two planes for the caps
	//        Cx     Cy     Cz      C
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);	// base
	addQuad( Z.x, Z.y, Z.z,-Z*Po);	// apex
} // makeConeQuads

/** createEllConeMesh - create mesh of an elliptical cone
 * @param Rbx	base radius along X vector
 * @param Rax	apex radius along X vector
 * @param Rby	base radius along Y vector
 * @param Ray	apex radius along Y vector
 */
void GCylinderBase::createEllConeMesh(	const double rbx, const double rax,
					const double rby, const double ray,
					const bool isinfinite)
{
	GBody::createMesh();
	Point A, B; //, C;
	if (isinfinite) {
		B = P - infinite*Z;
		A = P + infinite*Z;
		//C = P;		// additional ring!
	} else {
		B = P;
		A = Po;
	}
	if (_mesh->isEmpty())
		_mesh->createEllipticalCone(B,A, X,rbx,rax, Y,rby,ray, N_CYLINDER);
	else
		_mesh->updateEllipticalCone(B,A, X,rbx,rax, Y,rby,ray, N_CYLINDER);
} // createEllConeMesh

/** bbox_addRotatedEllipse
 * add to the bounding box an arbitrary rotated ellipse
 */
void GCylinderBase::bbox_addRotatedEllipse(const Point& pos, double ra, double rb)
{
	Vector VX = vectorX();
	Vector VY = vectorY();
	for (int i = 0; i < 3; i++) {
		double a = ra * VX[i];
		double b = rb * VY[i];
		double a2b2 = a*a + b*b;

		if (a2b2 < SMALL) continue;

		double sint = sqrt(b*b/a2b2);
		double cost = 0.0;
		if (Abs(sint) > SMALL)
			cost = a / b * sint;
		else {
			cost = 1.0;
			sint = 0.0;
		}

		Point e;
		e.x = VX.x * ra * cost + VY.x * rb * sint;
		e.y = VX.y * ra * cost + VY.y * rb * sint;
		e.z = VX.z * ra * cost + VY.z * rb * sint;

		_bbox.add(pos.x+e.x, pos.y+e.y, pos.z+e.z);
		_bbox.add(pos.x-e.x, pos.y-e.y, pos.z-e.z);
	}
} // bbox_addRotatedEllipse

// ================================= GRCCBody ==================================
/** set */
void GRCCBody::set(const double x,  const double y,  const double z,
		   const double hx, const double hy, const double hz,
		   const double r)
{
	P.set( x,  y,  z);
	Z.set(hx, hy, hz);
	R(Max(r,0.0));

	Po = P + Z;
	zlen = Z.normalize();

	findXYZ();
	if (r<TOOSMALL) throw std::string("Negative or Zero radius");
} // set

/** set */
void GRCCBody::set(int /*n*/, double* what)
{
	resetBBox();
	round0(what, 7, SMALL2);
	set(what[0], what[1], what[2],
	    what[3], what[4], what[5],
	    what[6]);
} // set

/* --- get --- */
int GRCCBody::get(double* what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = zlen*Z.x;
	what[4] = zlen*Z.y;
	what[5] = zlen*Z.z;
	what[6] = R();
	return 7;
} // get

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GRCCBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;

	switch (item) {
		case 1:	// Find distance from Z-axis
			R(Sqrt(T.length2() - Sq(T*Z)));
			break;

		case 2: {	// Move base
				double zp = T*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
			}
			break;

		case 3:		// Move apex
			zlen = T*SZ;
			if (zlen<0.0) {
				Z = -SZ;
				zlen = -zlen;
			} else
				Z = SZ;
			Po = SP + zlen*Z;
			break;

		default:
			GBody::move(item, r, w);
	}
} // move

/** createQuads */
void GRCCBody::createQuads()
{
	GBody::createQuads();
	addQuad(Quad(Quad::Type::CYLINDER, P, Z, R()));
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);	// base
	addQuad( Z.x, Z.y, Z.z,-Z*Po);	// apex
} // createQuads

/** @return bounding box of body */
void GRCCBody::_updateBBox()
{
	Point p = position();
	Point po = p + (vectorZ()*zlen);

	_bbox.add(p);
	_bbox.add(po);

	bbox_addRotatedEllipse(p,  R(), R());
	bbox_addRotatedEllipse(po, R(), R());
} // bbox

#ifdef USE_OBBOX
/** update oriented bounding box of body */
void GRCCBody::_updateOBB(bool in)
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	if (!in) {
		bb.low()  = Point(-R(), -R(), 0.0);
		bb.high() = Point( R(),  R(), zlen);
	} else {
		double sc = R()/Sqrt2;
		bb.low()  = Point(-sc, -sc, 0.0);
		bb.high() = Point( sc,  sc, zlen);
	}
} // _updateOBB
#endif

// ================================= GRECBody ==================================
/** set */
void GRECBody::set(const double x,  const double y,  const double z,
		   const double wx, const double wy, const double wz,
		   const double ux, const double uy, const double uz,
		   const double vx, const double vy, const double vz)
{
	P.set( x,  y,  z);
	Z.set(wx, wy, wz);
	X.set(ux, uy, uz);
	Y.set(vx, vy, vz);

	Po = P + Z;

	// correct for perpendicularity
	xlen = X.normalize();
	ylen = Y.normalize();
	zlen = Z.normalize();

	if (xlen<TOOSMALL) throw std::string("X-length is zero");
	if (ylen<TOOSMALL) throw std::string("Y-length is zero");
	checkOrthogonal();
} // set

/** set */
void GRECBody::set(int /*n*/, double* what)
{
	resetBBox();
	round0(what, 12, SMALL2);
	set(what[0], what[ 1], what[ 2],
	    what[3], what[ 4], what[ 5],
	    what[6], what[ 7], what[ 8],
	    what[9], what[10], what[11]);
} // set

/* --- get --- */
int GRECBody::get(double* what) const
{
	what[ 0] = P.x;
	what[ 1] = P.y;
	what[ 2] = P.z;
	what[ 3] = zlen*Z.x;
	what[ 4] = zlen*Z.y;
	what[ 5] = zlen*Z.z;
	what[ 6] = xlen*X.x;
	what[ 7] = xlen*X.y;
	what[ 8] = xlen*X.z;
	what[ 9] = ylen*Y.x;
	what[10] = ylen*Y.y;
	what[11] = ylen*Y.z;
	return 12;
} // get

/**  @return closest point, edge, face */
int GRECBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);
	if (close == 1) {
		double px = Abs(Y*(r-P));
		double py = Abs(X*(r-P));
		if (px>py) return 4;
	}
	return close;
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GRECBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;
	switch (item) {
		case 1:	{ // X-radius
				// Find a solution where the surface cross the r
				Point rp = P + (T*Z)*Z;	// projection on Z axis
				double x = (r-rp)*X;		// x,y components on X,Y axis
				double y = (r-rp)*Y;
				// equation
				//  x^2    y^2
				//  ---  + ---  = 1
				//  xl^2   yl^2
				double d = 1.0 - Sq(y)/Sq(sylen);
				if (d > SMALL)
					xlen = sqrt(Sq(x) / d);
				else	// On failure find distance from Y-axis
					xlen = Sqrt(T.length2() - Sq(T*Y));
			}
			break;

		case 2: {	// Move base
				double zp = T*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
			}
			break;

		case 3:		// Move apex
			zlen = T*SZ;
			if (zlen<0.0) {
				Z = -SZ;
				zlen = -zlen;
			} else
				Z = SZ;
			Po = P + zlen*Z;
			break;

		case 4:	{ // X-radius
				// Find distance from Z-axis
				Point rp = P + (T*Z)*Z;	// projection on Z axis
				double x = (r-rp)*X;	// x,y components on X,Y axis
				double y = (r-rp)*Y;
				double d = 1.0 - Sq(x)/Sq(sxlen);
				if (d > SMALL)
					ylen = Sqrt(Sq(y) / d);
				else	// On failure find distance from X-axis
					ylen = Sqrt(T.length2() - Sq(T*X));
			}
			break;

		default:
			GBody::move(item, r, w);
	}
} // move

/** createQuads */
void GRECBody::createQuads()
{
	GBody::createQuads();
	if (ISZERO(xlen) || ISZERO(ylen)) return;

	// Create the cylinder around Z then transform
	//          Cxx             Cyy      Czz  Cxy  Cxz  Cyz  Cx   Cy   Cz    C
	addQuad(1.0/Sq(xlen),  1.0/Sq(ylen), 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,-1.0);
	makeConeQuads();
} // createQuads

/** @return bounding box of body */
void GRECBody::_updateBBox()
{
	Point p = position();
	Point po = p + (vectorZ()*zlen);

	_bbox.add(p);
	_bbox.add(po);

	bbox_addRotatedEllipse(p,  xlen, ylen);
	bbox_addRotatedEllipse(po, xlen, ylen);
} // bbox

#ifdef USE_OBBOX
/** update oriented bounding box of body */
void GRECBody::_updateOBB(bool in)
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	if (!in) {
		bb.low()  = Point(-xlen, -ylen, 0.0);
		bb.high() = Point( xlen,  ylen, zlen);
	} else {
		double sc = 1.0/Sqrt2;
		bb.low()  = Point(-xlen*sc, -ylen*sc, 0.0);
		bb.high() = Point( xlen*sc,  ylen*sc, zlen);
	}
} // _updateOBB
#endif

// ============================== GInfEllCylBody ===============================
/** set */
void GInfEllCylBody::set(const double a, const double b, const double ra, const double rb)
{
	if (ra<TOOSMALL or rb<-TOOSMALL)
		throw std::string("Negative radius");

	Rx(Max(ra,0.0));
	Ry(Max(rb,0.0));

	switch (type()) {
		case Type::CX:
		case Type::C_X:
		case Type::XCC:
		case Type::XEC:
			P.set(0.0, a, b);
			if (type() != Type::XEC) ylen = xlen;
			if (Eq0(a,SMALL) and Eq0(b,SMALL) and type()==Type::C_X)
				_type = Type::CX;
			Z = Vector::Xo;
			X = Vector::Yo;
			Y = Vector::Zo;
			break;

		case Type::CY:		// mcnp order: x z
		case Type::C_Y:
			P.set(a, 0.0, b);
			if (type() != Type::YEC) ylen = xlen;
			if (Eq0(a,SMALL) and Eq0(b,SMALL) and type()==Type::C_Y)
				_type = Type::CY;
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			break;

		case Type::YCC:		// fluka order: z x
		case Type::YEC:
			P.set(b, 0.0, a);
			if (type() != Type::YEC) ylen = xlen;
			if (Eq0(a,SMALL) and Eq0(b,SMALL) and type()==Type::C_Y)
				_type = Type::CY;
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			break;

		case Type::CZ:
		case Type::C_Z:
		case Type::ZCC:
		case Type::ZEC:
			P.set(a, b, 0.0);
			if (type() != Type::ZEC) ylen = xlen;
			if (Eq0(a,SMALL) and Eq0(b,SMALL) and type()==Type::C_Z)
				_type = Type::CZ;
			X = Vector::Xo;
			Y = Vector::Yo;
			Z = Vector::Zo;
			break;
		default:
			assert(false);
	}
} // set

/** set */
void GInfEllCylBody::set(int /*n*/, double* what)
{
	resetBBox();
	if (In(type(), Type::CX, Type::CY, Type::CZ)) {
		round0(what, 1, SMALL2);
		set(0., 0., what[0], what[0]);
	} else {
		round0(what, 4, SMALL2);
		set(what[0], what[1], what[2], what[3]);
	}
} // set

/* --- get --- */
int GInfEllCylBody::get(double* what) const
{
	if (In(type(), Type::CX, Type::CY, Type::CZ)) {
		what[0] = Rx();
		return 1;
	}

	what[2] = Rx();
	what[3] = Ry();		// For [XYZ]EC only

	switch (type()) {
		case Type::C_X:
		case Type::XCC:
		case Type::XEC:
			what[0] = P.y;
			what[1] = P.z;
			if (type()==Type::XEC) return 4;
			return 3;

		case Type::C_Y:		// mcnp order: x z
			what[0] = P.x;
			what[1] = P.z;
			return 3;

		case Type::YCC:		// fluka order: z x
		case Type::YEC:
			what[0] = P.z;
			what[1] = P.x;
			if (type()==Type::YEC) return 4;
			return 3;

		case Type::C_Z:
		case Type::ZCC:
		case Type::ZEC:
			what[0] = P.x;
			what[1] = P.y;
			if (type()==Type::ZEC) return 4;
			return 3;

		default: return -1;
	}
} // get

/**  @return closest point, edge, face */
int GInfEllCylBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);

	switch (type()) {
		case Type::XEC:
			if (Abs(r.y-P.y) < Abs(r.z-P.z))
				return 2;
			return close;
		case Type::YEC:
			if (Abs(r.z-P.z) < Abs(r.x-P.x))
				return 2;
			return close;
		case Type::ZEC:
			if (Abs(r.x-P.x) < Abs(r.y-P.y))
				return 2;
			return close;
		default:
			return close;
	}
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GInfEllCylBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;

	double x, y;
	bool ell = true;

	switch (type()) {
		case Type::XCC:
		case Type::C_X:
		case Type::CX:
			ell = false;
			/* FALLTHRU */
		case Type::XEC:
			x = T.y;
			y = T.z;
			break;

		case Type::YCC:
		case Type::C_Y:
		case Type::CY:
			ell = false;
			/* FALLTHRU */
		case Type::YEC:
			x = T.z;
			y = T.x;
			break;

		case Type::ZCC:
		case Type::C_Z:
		case Type::CZ:
			ell = false;
			/* FALLTHRU */
		case Type::ZEC:
			x = T.x;
			y = T.y;
			break;

		default:
			x = y = 0.0;	// Never comes here but keeps compiler happy
			assert(false);
	}

	if (item==1) {
		if (ell) {
			double d = 1.0 - Sq(y)/Sq(ylen);
			if (d > SMALL)
				Rx(Sqrt(Sq(x) / d));
			else	// On failure find distance from Y-axis
				Rx(x);
		} else
			R(Sqrt(Sq(x) + Sq(y)));
	} else
	if (item==2) {
		double d = 1.0 - Sq(x)/Sq(Rx());
		if (d > SMALL)
			ylen = Sqrt(Sq(y) / d);
		else
			ylen = y;
	} else {
		GBody::move(item, r, w);

		switch (type()) {
			case Type::CX:
				if (P.y!=0.0 or P.z!=0.0)
					_type = Type::C_X;
				break;
			case Type::CY:
				if (P.x!=0.0 or P.z!=0.0)
					_type = Type::C_Y;
				break;
			case Type::CZ:
				if (P.x!=0.0 or P.y!=0.0)
					_type = Type::C_Z;
				break;
			case Type::C_X:
				if (P.y==0.0 and P.z==0.0)
					_type = Type::CX;
				break;
			case Type::C_Y:
				if (P.x==0.0 and P.z==0.0)
					_type = Type::CY;
				break;
			case Type::C_Z:
				if (P.x==0.0 and P.y==0.0)
					_type = Type::CZ;
				break;
			default: NOP;
		}
	}
} // move

/** rotate body around an axis */
void GInfEllCylBody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle, axis);

	/* find closest axis to align to */
	double ax = Abs(Z.x);
	double ay = Abs(Z.y);
	double az = Abs(Z.z);

	if (az>ax && az>ay) {
		Z = Vector::Zo;
		if (In(type(), Type::XEC, Type::YEC, Type::ZEC))
			_type = Type::ZEC;
		else
		if (In(type(), Type::XCC, Type::YCC, Type::ZCC))
			_type = Type::ZCC;
		else
			_type = Type::C_Z;
	} else
	if (ay>ax && ay>az) {
		Z = Vector::Yo;
		if (In(type(), Type::XEC, Type::YEC, Type::ZEC))
			_type = Type::YEC;
		else
		if (In(type(), Type::XCC, Type::YCC, Type::ZCC))
			_type = Type::YCC;
		else
			_type = Type::C_Y;
	} else {
		Z = Vector::Zo;
		if (In(type(), Type::XEC, Type::YEC, Type::ZEC))
			_type = Type::XEC;
		else
		if (In(type(), Type::XCC, Type::YCC, Type::ZCC))
			_type = Type::XCC;
		else
			_type = Type::C_X;
	}

	findXYZ();
} // rotate

/** createQuads */
void GInfEllCylBody::createQuads()
{
	GBody::createQuads();
	double a = 1.0, b = 1.0, c = 1.0;
	double r2 = Sq(xlen);

	switch (type()) {
		case Type::XCC:
		case Type::C_X:
			addQuad(Quad(Quad::Type::CYLINDER_X, P, R()));
			return;
		case Type::CX:
			addQuad(Quad(Quad::Type::CYLINDER_XO, R()));
			return;

		case Type::YCC:
		case Type::C_Y:
			addQuad(Quad(Quad::Type::CYLINDER_Y, P, R()));
			return;
		case Type::CY:
			addQuad(Quad(Quad::Type::CYLINDER_YO, R()));
			return;

		case Type::ZCC:
		case Type::C_Z:
			addQuad(Quad(Quad::Type::CYLINDER_Z, P, R()));
			return;
		case Type::CZ:
			addQuad(Quad(Quad::Type::CYLINDER_ZO, R()));
			return;

		case Type::XEC:
			a   = 0.0;
			c   = r2;
			b   = Sq(ylen);
			r2 *= b;
			break;
		case Type::YEC:
			b   = 0.0;
			a   = r2;
			c   = Sq(ylen);
			r2 *= c;
			break;
		case Type::ZEC:
			c   = 0.0;
			b   = r2;
			a   = Sq(ylen);
			r2 *= a;
			break;
		default: assert(0);
	}

	// add quad for [XYZ]EC
	//      Cxx Cyy Czz Cxy  Cxz  Cyz
	addQuad( a,  b,  c, 0.0, 0.0, 0.0,
		// Cx            Cy            Cz
		-2.0*a*P.x,  -2.0*b*P.y,  -2.0*c*P.z,
		 a*Sq(P.x) + b*Sq(P.y) + c*Sq(P.z) - r2);
} // createQuads

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 */
GBody::Location GInfEllCylBody::_locationWrt(const GBody* body) const
{
	//if (body->type() == Type::RCC) {
	//} else
	if (isInfCylinder() and body->isInfCylinder()) {
		Vector thisZ = vectorZ();
		Vector bodyZ = body->vectorZ();

		// Minimum distance between the two axes
		double dist = lineLineDistance(position(), thisZ, body->position(), bodyZ);

		// If they are parallel.
		// FIXME could be that after a rotation thisZ = +/- bodyZ
		if (Eq(Abs(thisZ * bodyZ), 1.0, SMALL)) {
			if (dist + ((const GInfEllCylBody*)body)->R() <  R())
				return Location::BinA;
			else
			if (dist - ((const GInfEllCylBody*)body)->R() < -R())
				return Location::AinB;
			else
			if (dist - ((const GInfEllCylBody*)body)->R() >  R())
				return Location::OUTSIDE;
			else
				return Location::OVERLAP;
		} else
		// any other skew orientation (after rotation)
		if (dist - ((const GInfEllCylBody*)body)->R() >  R())
			return Location::OUTSIDE;
		else
			return Location::OVERLAP;
	}

	return GBody::_locationWrt(body);
} // _locationWrt

/** @return bounding box of body */
void GInfEllCylBody::_updateBBox()
{
	switch (type()) {
		case Type::XCC:
		case Type::C_X:
		case Type::CX:
			_bbox.low(-INFINITE, P.y-R(), P.z-R());
			_bbox.high(INFINITE, P.y+R(), P.z+R());
			break;
		case Type::YCC:
		case Type::C_Y:
		case Type::CY:
			_bbox.low( P.x-R(), -INFINITE, P.z-R());
			_bbox.high(P.x+R(),  INFINITE, P.z+R());
			break;
		case Type::ZCC:
		case Type::C_Z:
		case Type::CZ:
			_bbox.low( P.x-R(), P.y-R(), -INFINITE);
			_bbox.high(P.x+R(), P.y+R(),  INFINITE);
			break;

		case Type::XEC:
			_bbox.low(-INFINITE, P.y-Rx(), P.z-Ry());
			_bbox.high(INFINITE, P.y+Rx(), P.z+Ry());
			break;
		case Type::YEC:
			_bbox.low( P.x-Ry(), -INFINITE, P.z-Rx());
			_bbox.high(P.x+Ry(),  INFINITE, P.z+Rx());
			break;
		case Type::ZEC:
			_bbox.low( P.x-Rx(), P.y-Ry(), -INFINITE);
			_bbox.high(P.x+Rx(), P.y+Ry(),  INFINITE);
			break;

		default:
			assert(0);
	}
	if (_hasMatrix) _bbox.transform(_matrix);
} // bbox

#ifdef USE_OBBOX
/** update oriented bounding box of body */
void GInfEllCylBody::_updateOBB(bool in)
{
	OBBox& bb = _obbox[(int)in];
	bb.infinite();

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();

	if (!in) {
		bb.low().x  = -Rx();
		bb.low().y  = -Ry();
		bb.high().x =  Rx();
		bb.high().y =  Ry();
	} else {
		double sc = 1.0/Sqrt2;
		bb.low().x  = -Rx()*sc;
		bb.low().y  = -Ry()*sc;
		bb.high().x =  Rx()*sc;
		bb.high().y =  Ry()*sc;
	}
} // _updateOBB
#endif
