/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	10-Oct-2018
 */

#ifndef __GEOMETRY_CYLGBODY_H
#define __GEOMETRY_CYLGBODY_H

#include "gbody.h"

// =============================== GCylinderBase ===============================
// Base class of all cylinder like bodies RCC, REC, TRC, ... to create the mesh
class GCylinderBase : public GBody {
public:
	static int N_CYLINDER;

public:
	GCylinderBase(const std::string& aname, const Type atype) : GBody(aname, atype) {}
protected:
	void	makeConeQuads();
	// mesh functions
	void	createEllConeMesh(const double rbx, const double rax,
				  const double rby, const double ray,
				  const bool isinfinite=false);
	void	bbox_addRotatedEllipse(const Point& pos, double ra, double rb);
}; // class GCylinderBase

// ================================= GRCCBody ==================================
class GRCCBody : public GCylinderBase {
public:
	GRCCBody(const std::string& aname) : GCylinderBase(aname, Type::RCC) {}

	double	R()	const		{ return xlen; }
	void	R(const double r)	{ xlen = ylen = r; }

	void	set(const double x,  const double y,  const double z,
		    const double hx, const double hy, const double hz,
		    const double r);
	void	set(const Vector& pos, const Vector& h, const double r)
			{ set(pos.x, pos.y, pos.z, h.x, h.y, h.z, r); }
	virtual	void	set(int n, double*);
	virtual	int	get(double*) const;
	virtual	int	nodes()		const	{ return 2; }
	virtual void	move(int item, const Point& r, const Vector& w);

	// FIXME
	//virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const { return "R"; }
	virtual const char*	showY()	const { return NULL; }
	virtual const char*	showZ()	const { return "H"; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh()	{ createEllConeMesh(xlen,xlen, ylen,ylen); }
	virtual	void	_updateBBox();
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool in=false);
#endif
}; // GRCCBody

// ================================= GRECBody ==================================
class GRECBody : public GCylinderBase {
public:
	GRECBody(const std::string& aname) : GCylinderBase(aname, Type::REC) {}
	void	set(const double x,  const double y,  const double z,
		    const double wx, const double wy, const double wz,
		    const double ux, const double uy, const double uz,
		    const double vx, const double vy, const double vz);
	void	set(const Vector& pos, const Vector& w,
		    const Vector& u,   const Vector& v)
			{ set(pos.x, pos.y, pos.z,
			      w.x, w.y, w.z,
			      u.x, u.y, u.z,
			      v.x, v.y, v.z); }
	virtual	void	set(int n, double*);
	virtual	int	get(double*) const;
	virtual	int	nodes()		const	{ return 4; }
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);

	// FIXME
	//virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const { return "Rx"; }
	virtual const char*	showY()	const { return "Ry"; }
	virtual const char*	showZ()	const { return "H"; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh()	{ createEllConeMesh(xlen,xlen, ylen,ylen); }
	virtual	void	_updateBBox();
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool in=false);
#endif
}; // GRECBody

// ============================== GInfEllCylBody ===============================
class GInfEllCylBody : public GCylinderBase {
public:
	GInfEllCylBody(const std::string& aname, const Type atype)
		: GCylinderBase(aname, atype)
			{ zlen = infinite; }

	Vector	direction() const;

	double	R()	const		{ return xlen; };
	double	Rx()	const		{ return xlen; };
	double	Ry()	const		{ return ylen; };
	void	R(const double r)	{ xlen = ylen = r; }
	void	Rx(const double r)	{ xlen = r; }
	void	Ry(const double r)	{ ylen = r; }

	void	set(const double a, const double b, const double ra, const double rb=0.0);

	virtual	void	set(int n, double*);
	virtual	int	get(double*) const;
	virtual	int	closest(const Point& r, const double d, const Vector& w) const;
	virtual void	move(int item, const Point& r, const Vector& w);
	virtual	void	rotate(const double angle, const Vector& axis);

	// FIXME
	//virtual	double	distance(const double x, const double y, const double z) const;

	// Dimension labels and show flags
	// @return NULL not to show anything otherwise label
	virtual	const char*	showX()	const {
				switch (type()) {
					case Type::XEC: return "Ry";
					case Type::YEC: return "Rz";
					case Type::ZEC: return "Rx";
					default:
						return "R";
				}
			}
	virtual const char*	showY()	const {
				switch (type()) {
					case Type::XEC: return "Rz";
					case Type::YEC: return "Rx";
					case Type::ZEC: return "Ry";
					default:
						return NULL;
				}
			}
	virtual const char*	showZ()	const { return NULL; }

protected:
	virtual	void	createQuads();
	virtual void	createMesh() { createEllConeMesh(xlen,xlen, ylen,ylen,true); }
	virtual	void	_updateBBox();
#ifdef USE_OBBOX
	virtual	void	_updateOBB(bool in=false);
#endif
	virtual	Location _locationWrt(const GBody* body) const;
}; // GInfEllCylBody

#endif
