/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "gregion.h"
#include "geometry.h"
#include "math/obbox.h"

using std::endl;

/* ================================ GREGION ================================= */
/** FLUKA region definition */
GRegion::GRegion(const std::string& aname) :
	_name(aname),
	_id(-1),
	_type(Type::NORMAL),
	_generation(0),
	_hash(0),
	_material(NULL),
	_hasMatrix(false),
	show(0)
{
} // GRegion

/** matrix - set transformation matrix */
void GRegion::matrix(const Matrix4& M)
{
	if (!_hasMatrix) {
		// First time set the matrix
		_hasMatrix = true;
		_matrix.copy(M);
	} else {
		// Multiple with the existing matrix
		// _matrix = M * _matrix
		Matrix4 N;
		N.multiply(M, _matrix);
		_matrix.copy(N);
	}
	_matrix.fix();
	_invMatrix.inverse(_matrix);		// Invert matrix
	_invMatrix.fix();
} // matrix

/** @return memory used by region */
size_t GRegion::memory() const
{
	size_t mem = sizeof(GRegion);
	for (auto zone : _zones)
		mem += zone->memory();
	return mem;
} // memory

/** addZone */
GZone* GRegion::addZone(ZoneFormat aformat)
{
	GZone* zone = new GZone(this);

	zone->id(_zones.size());
	zone->format(aformat);

	_zones.push_back(zone);
	return zone;
} // addZone

/** addZone */
GZone* GRegion::addZone(GZone* z)
{
	GZone* zone = addZone(z->format());
	zone->reserve(z->size());
	for (int i=0; i<z->size(); i++)
		zone->add((*z)[i]);
	return zone;
} // addZone

/** insert a new zone at position pos */
GZone* GRegion::insertZone(int pos, GZone* z)
{
	GZone* zone = new GZone(this);
	_zones.insert(_zones.begin()+pos, zone);
	zone->id(pos);			// warning subsequent ids are wrong!
	zone->format(z->format());
	zone->reserve(z->size());
	for (int i=0; i<z->size(); i++)
		zone->add((*z)[i]);
	return zone;
} // insertZone

/** delZone */
void GRegion::delZone(int i)
{
	delete _zones[i];
	_zones.erase(_zones.begin()+i);
} // delZone

/** add2exp add token/body to last zone
 * @return true on success
 */
bool GRegion::add2exp(const char* token, GBody* body)
{
	assert(!_zones.empty());
	return _zones.back()->add(token, body);
} // add2exp

/** convert a region with one zone in normal format
 * to multiple zones in either standard or rpn format
 */
void GRegion::convertFromNormal()
{
	for (int iz=0; iz<size(); iz++) {
		auto zone = _zones[iz];
		if (not zone->normal()) {
			zone->id(iz);
			continue;
		}

		_zones.erase(_zones.begin()+iz);	// remove from vector

		int iadd = 0;
		int from = 0;
		// skip leading |
		if (zone->size()>0 and (*zone)[from]==&GBody::tunion) from++;
		for (int to=0; ; from=to+1) {
			to = zone->findZone(from);
			if (to<0) break;
			GZone subzone;
			zone->subZone(&subzone, from, to);
			insertZone(iz+iadd, &subzone);
			iadd++;
		}
		delete zone;	// to remove all references
	}
} // convertFromNormal

/** call setup on all zones */
void GRegion::setup()
{
	for (auto zone : _zones)
		zone->setup();
} // setup

/** delete zone if it is a duplicate of another
 * @return true if the zone is removed
 */
bool GRegion::removeDup(int iz)
{
	GZone* zone = _zones[iz];
	if (zone->empty()) return false;
	for (auto izone : _zones) {
		if (zone == izone) continue;		// skip self
		if (zone->isSubsetOf(*izone)) {
			//std::cout << "+++ Duplicate Zone: " << *zone << std::endl;
			//std::cout << "              with: " << *izone << std::endl;
			zone->clear();
			zone->updateBBox();
			return true;
		}
	}
	return false;
} // removeDup

/** merge current zone defines by iz with other zones in the region
 * @param iz	index of zone to merge with others
 */
bool GRegion::merge(int iz)
{
	GZone* zone = _zones[iz];
	if (not zone->std() or zone->empty()) return false;
	bool changed = false;
	for (auto izone : _zones) {
		if (zone == izone) continue;		// skip self
		if (not izone->std()) continue;		// skip non STD zones
		if (zone->merge(*izone))
			changed = true;
	}
	return changed;
} // merge

/** @return true if region contains zone as expression of subset */
bool GRegion::contains(const GZone& zone)
{
	for (auto izone : _zones)
		if (zone.isSubsetOf(*izone))
			return true;
	return false;
} // contains

/** expand possible #regions present in the zones */
bool GRegion::expandRegions(Geometry* geometry)
{
	convertFromNormal();
	for (auto zone : _zones)
		if (zone->format() == GZone::Format::RPN)
			zone->expandRegions(geometry);
	return true;
} // expandRegions

/** optimize zones */
bool GRegion::optimize()
{
	bool changed = false;
	//cout << "----- Before --------" << endl;
	//for (int j=0; j<size(); j++)
	//	cout << *(_zones[j]) << endl;
	bool mod1;
	do {	// loop until no more changes are done
		mod1 = false;
		for (int i=0; i<size(); i++) {
			GZone* zone = _zones[i];
			bool mod2;		// modified within the loop
			do {
				mod2 = false;
				if (not zone->bbox().isValid()) {
					changed = true;
					zone->clear();
					break;
				}
				// 1st try to geometrically optimize
				if (zone->std())
					mod2 = zone->optimize();
				else
				if (zone->rpn())
					mod2 = zone->optimizeProducts();

				// 2nd try to remove possible duplicate
				if (removeDup(i)) mod2 |= true;

				// 3rd try to merge with other zones
				if (merge(i)) mod2 |= true;
				mod1 |= mod2;
				changed |= mod2;
			} while (mod2 and not zone->empty());
			// printout
			//cout << "++++ Step " << i << endl;
			//for (int j=0; j<size(); j++)
			//	cout << *(_zones[j]) << endl;
		}
	} while (mod1 and size());
	return changed;
} // optimize

/** @return hash value of expression */
dword GRegion::hash() const
{
	if (_hash>0) return _hash;
	_hash = 0;
	int i=1;
	for (auto zone : _zones) {
		_hash += i * zone->hash();
		i += 2;
	}
	return _hash;
} // hash

/** clear */
void GRegion::clear(bool fast)
{
	// Clean the zones by popping the elements in order
	for (auto zone : _zones) {
		if (fast) zone->clear(fast);
		delete zone;
	}
	_zones.erase(_zones.begin(), _zones.end());
	_hash = 0;
	//clearBB();
} // clear

/** update bbox of region zones */
void GRegion::updateBBox()
{
	for (auto zone : _zones)
		zone->updateBBox();
} // updateBbox

/** guess of 3D bounding box
 * @return bounding box of zone
 */
BBox GRegion::bbox() const
{
	BBox bb;

	for (auto zone : _zones)
		if (bb.isValid())
			bb |= zone->bbox();
		else
			bb = zone->bbox();
	return bb;
} // bbox

#ifdef USE_OBBOX
/** guess of 3D oriented bounding box
 * @return bounding box of zone
 */
OBBox GRegion::obbox() const
{
	OBBox obb;

	for (auto zone : _zones)
		if (obb.isValid())
			obb.Union(zone->obbox());
		else
			obb = zone->obbox();
	return obb;
} // obbox
#endif

/** check if point (x,y,z) with (u,v,w) direction is inside the zone or not
 * @param x,y,z    location of point to search
 * @param u,v,w    direction in case on boundary
 * @return zone    return zone that location is inside otherwise NULL
 *
 * @warning	NON-Caching function
 */
GZone* GRegion::inside(const double x, const double y, const double z,
		       const double u, const double v, const double w) const
{
	for (auto zone : _zones)
		if (zone->inside(x,y,z, u,v,w))
			return zone;
	return NULL;
} // inside

/** inside */
GZone* GRegion::inside(const Point& p, const Vector& d,
			const double t, Cache<CBody3D>& cache) const
{
	for (auto zone : _zones)
		if (zone->inside(p,d,t,cache))
			return zone;
	return nullptr;
} // inside

/** calculate track length in a region and the number of boundary crossings
 * given a position (pos) an direction (dir)
 * @param pos		initial position
 * @param dir		direction
 * @param tlength	track length inside region
 * @param cache		caching information of bodies
 * @return		number of boundary crossings
 */
unsigned GRegion::boundaryCrossings(const Point& pos, const Point& dir,
			double* tlength, Cache<CBody3D>& cache) const
{
	double tmin = 0;
	//double tmax = INFINITE; //*tlength;
	double tmax = *tlength;

	unsigned nbdx = 0;	// boundary crossings counter
	*tlength = 0;		// reset track length

	GZone* zone = inside(pos, dir, tmin, cache);
	assert(zone==nullptr); // we must always be outside!!!

	// assert check intersection with bounding box
	if (!bbox().distance(pos, dir, tmin, tmax))
		return nbdx;

	while (tmin<tmax) {
		if (zone==nullptr) {
			// find zone that we are entering
			zone = nullptr;
			double tmax2 = tmax;
			for (auto zzz : _zones) {
				double tmin2 = tmin;
				if (zzz->distance2In(pos,dir, &tmin2,tmax2, cache)) {
					zone  = zzz;
					tmax2 = tmin2;
				}
			}
			if (zone == nullptr) return nbdx;
			nbdx++;	// tally boundary crossing
			tmin = tmax2;
			assert(zone->inside(pos,dir,tmin,cache));
			DUMP(std::cout << "In: " << pos+dir*tmin << std::endl);
		}

		// we are inside zone
		*tlength -= tmin;		// tally negative enter distance

		// find distance to exit
		DEBUG(GBody* body =)
		zone->distance2Out(pos,dir, &tmin,tmax, cache);
		assert(body != nullptr);	// we should always exit

		*tlength += tmin;		// tally positive exit distance

		// find next zone at current location
		assert(!zone->inside(pos,dir,tmin,cache));
		zone = inside(pos,dir, tmin, cache);
		if (zone==nullptr) {
			DUMP(std::cout << "Out: " << pos+dir*tmin << std::endl);
			nbdx++;	// tally boundary crossing
		}
	}
	return 0;
} // boundaryCrossings

/** operator << */
std::ostream& operator << (std::ostream& s, const GRegion& region)
{
	s << region.name() << std::endl;

	for (int i=0; i<region.size(); i++)
		s << "    " << *region[i] << std::endl;
	return s;
} /* operator << */
