/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 */

#include "voxel.h"
#include <sys/stat.h>

#include "math/eps.h"
#include "tool/fortran.h"

using std::cout;
using std::cerr;
using std::endl;

// ============================== ROICombination ===============================
/** parse buffer for combinations */
void ROICombination::parse(void* buffer, int buflen)
{
	FortranParser parser(buffer, buflen);
	length = parser.readInt();
	roi    = new short[length];
	for (int i=0; i<length; i++)
		roi[i] = parser.readWord();
} // parse

#ifdef EXPERIMENTAL
// ================================= ROIPlanar =================================
/** ROIPlanar */
ROIPlanar::~ROIPlanar()
{
	for (int i=0; i<nz; i++)
		if (slice[i]) delete slice[i];
	delete [] slice;
} // ROIPlanar

/** allocate structure slices */
void ROIPlanar::alloc(int anz)
{
	assert(nz==0);
	nz = anz;
	slice = new ROIPlanarSlice*[nz];
	bzero(slice, sizeof(ROIPlanarSlice*)*nz);
} // alloc

/** addZslice */
ROIPlanarSlice& ROIPlanar::addZslice(int z, int n)
{
	assert(slice[z]==NULL);
	slice[z] = new ROIPlanarSlice(n);
	return *slice[z];
} // addZslice
#endif

// ================================== GVoxel ===================================
/** GVoxel */
GVoxel::GVoxel()
	: _data(NULL),
	  _vxl2reg(NULL),
	  _roi(NULL),
	  _roiName(NULL),
	  _roiColor(NULL),
	  _roiVolume(NULL),
	  _roiComb(NULL)
#ifdef EXPERIMENTAL
	, _roiPlanar(NULL)
#endif
{
	nx = ny = nz = 0;
	xlow = ylow = zlow = 0;
	nregions = nvoxels = 0;
	_mtime  = 0;
	_nroi = _roiCombN = 0;
} // GVoxel

/** cleanup */
void GVoxel::cleanup()
{
	if (_data)	delete [] _data;
	if (_vxl2reg)	delete [] _vxl2reg;
	if (_roiName)	delete [] _roiName;
	if (_roiColor)	delete [] _roiColor;
	if (_roiVolume)	delete [] _roiVolume;
	if (_roiComb)	delete [] _roiComb;
#ifdef EXPERIMENTAL
	if (_roiPlanar)	delete [] _roiPlanar;
#endif
	if (_roi)	delete [] _roi;
	filename.clear();
	_data      = NULL;
	_vxl2reg   = NULL;
	_roiName   = NULL;
	_roiColor  = NULL;
	_roiVolume = NULL;
	_roiComb   = NULL;
#ifdef EXPERIMENTAL
	_roiPlanar = NULL;
#endif
	_roi       = NULL;
	nx = ny = nz = 0;
	nregions = nvoxels = 0;
	_mtime  = 0;
	_hasMatrix = false;
	_nroi = _roiCombN = 0;
} // cleanup

/** @return maximum region in kreg */
word GVoxel::maxRegion() const
{
	word maxreg = 0;
	for (int i=0; i<=nvoxels; i++)
		maxreg = Max(maxreg, _vxl2reg[i]);
	return maxreg;
} // maxRegion

/** addKreg */
word GVoxel::addKreg(word region)
{
	assert(_vxl2reg);
	word maxreg = maxRegion();
	if (region > maxreg)
		region = maxreg + 1;

	// Increase the kreg array
	nvoxels++;
	nregions++;
	word* kregnew = new word[nvoxels+1];
	memcpy(kregnew, _vxl2reg, sizeof(word)*nvoxels);
	kregnew[nvoxels] = region;

	delete [] _vxl2reg;
	_vxl2reg = kregnew;

	return region;
} // addKreg

/** load
 * @return	true on success
 *		false on failure
 */
bool GVoxel::load(const char* fn, bool loadCards)
{
	char	buffer[5120];
	int	length, readin, buflen;
	struct stat st;
	FortranParser parser(buffer,sizeof(buffer));

	// Check if information is valid
	if (stat(fn, &st)) {
		cleanup();
		return false;
	}
	if (!filename.compare(fn) && st.st_mtime==_mtime) return true;

	cleanup();

	FortranFile f(fn, "rb");
	if (!f) return false;

	// Remember modification name and time
	filename = fn;
	_mtime = st.st_mtime;

	/* read title */
	memset(title, 0, sizeof(title));
	if (f.read(title,sizeof(title))<0) {
		cerr << "ERROR reading voxel title" << endl;
		goto ERROR;
	}
	FortranParser::strip(title);

	/* read arrays header  */
	if ((buflen=f.read(parser))<0 || (buflen<5*4)) {
		cerr << "ERROR reading voxel header size=" << buflen << endl;
		goto ERROR;
	}
	nx = parser.readInt();
	ny = parser.readInt();
	nz = parser.readInt();
	nregions = parser.readInt();
	nvoxels  = parser.readInt();
	if (parser.more()) {
		_nroi       = parser.readInt();
		_roiCombN   = parser.readInt();
		_roiMaxComb = parser.readInt();
#ifdef EXPERIMENTAL
		if (parser.more())
			_roiPlanarN = parser.readInt();
#endif
	}

	/* read voxel dimensions */
	if (f.read(parser) != 24) {
		cerr << "ERROR reading voxel dimensions" << endl;
		goto ERROR;
	}
	dx = parser.readDouble();
	dy = parser.readDouble();
	dz = parser.readDouble();

	/* allocate arrays */
#ifdef _DEBUG
	printf("VOXEL\tTitle: %s\n",title);
	printf("\tnx=%d ny=%d nz=%d\n",nx,ny,nz);
	printf("\tdx=%g dy=%g dz=%g\n",dx,dy,dz);
	printf("\tDx=%g Dy=%g Dz=%g\n",(double)nx*dx,(double)ny*dy,(double)nz*dz);
	printf("\tnregions=%d nvoxels=%d\n",nregions,nvoxels);
#endif
	nynx = ny*nx;

	/* Read data */
	_data  = new word[size()];
	length = size()*sizeof(word);
	readin = f.read(_data, length);
	if (readin != length) {
		cerr << "ERROR wrong data length=" << readin << " expected=" << length << endl;
		goto ERROR;
	}
	/* Read region definitions */
	_vxl2reg = new word[nvoxels+1];	// include also the vxl=0 cage organ
	_vxl2reg[0] = 0;		// make base:1 for fortran
	readin = f.read(_vxl2reg+1, nvoxels*sizeof(word));
	if (readin != nvoxels*(int)sizeof(word)) {
		cerr << "ERROR wrong kreg length=" << readin
			<< " expected=" << nvoxels*(int)sizeof(word) << endl;
		goto ERROR;
	}

	/* ROI structures */
	if (_nroi) {
		/* Read the names */
		_roiName   = new std::string[_nroi+1];
		_roiColor  = new dword[_nroi+1];
		_roiVolume = new double[_nroi+1];
		for (int i=1; i<=_nroi; i++) {
			const int ROIMAXLEN = 64;
			char roiname[ROIMAXLEN+1];
			//memset(buffer, 0, sizeof(buffer));
			if (f.read(parser) != 76) {
				cerr << "ERROR reading voxel ROI names" << endl;
				goto ERROR;
			}
			_roiColor[i]  = parser.readInt();
			_roiVolume[i] = parser.readDouble();
			parser.read(roiname, ROIMAXLEN);
			_roiName[i] = roiname;
			DUMP(if (buffer[0]) cout << "ROI: " << i+1
				<< ": " << _roiName[i]
				<< ' '  << _roiColor[i]
				<< ' '  << _roiVolume[i]
				<< endl;)
		}

		/* Read the combination dictionary */
		_roiComb = new ROICombination[_roiCombN+1];
		for (int i=1; i<=_roiCombN; i++) {
			if ((buflen=f.read(buffer,sizeof(buffer)))<0) {
				cerr << "ERROR reading voxel ROI combination dictionary" << endl;
				goto ERROR;
			}
			// roiComb[0] is the basic combination with nothing ink:W
			_roiComb[i].parse(buffer, buflen);
			DUMP(printf("ROICMB %d ",i); _roiComb[i].print();)
		}

		/* Read the roi struct data */
		_roi = new word[size()];
		length = size()*sizeof(word);
		readin = f.read(_roi, length);
		if (readin != length) {
			cerr << "ERROR wrong data length=" << readin << " expected=" << length << endl;
			goto ERROR;
		}

//		memset(stat,0,sizeof(stat));
//		for (int i=0; i<length; i++)
//			stat[_roi[i]]++;
//		for (int i=0; i<_roiCombN; i++)
//			cout << i << " : " << stat[i] << endl;

#ifdef EXPERIMENTAL
		/* Read the planar information */
		if (_roiPlanarN)
			_roiPlanar = new ROIPlanar[_nroi+1];
		for (int i=0; i<_roiPlanarN; i++) {
			if (f.read(parser) != 8) {
				cerr << "ERROR reading planar header" << endl;
				goto ERROR;
			}
			/*int r =*/ parser.readInt();
			int n = parser.readInt();
#if _DEBUG>2
			cout << "ROI:" << r;
			cout << " r=" << r << endl;
			cout << " n=" << n << endl;
#endif

//			_roiPlanar[r].alloc(nz);

			for (int j=0; j<n; j++) {
//				buflen=f.read(buffer,sizeof(buffer));
				int buflen2;
				byte* buffer2 = f.readBuffer(&buflen2);
				if (buffer2==NULL) {
					cerr << "ERROR reading planar information" << endl;
					goto ERROR;
				}
				parser(buffer2, buflen2);
				/*int zi =*/ parser.readInt();
				/*int m  =*/ parser.readInt();

				//assert(buflen == m*2*sizeof(float) + 2*sizeof(int));
#if _DEBUG>2
				cout << '[' << i << "] z=" << zi;
				cout << " m=" << m;
				for (int k=0; k<m; k++) {
					float x = parser.readFloat();
					float y = parser.readFloat();
					cout << " (" << x << ',' << y << ')';
				}
				cout << endl;
#endif
				delete [] buffer2;
			}
		}
#endif
	}

	/* XXX maybe read internal cards */
	if (loadCards) {
		char card[81];
		while (f) {
			f.read(card, sizeof(card));
			card[80] = 0;
			cards.push_back(card);
			DUMP(cout << "Voxel: " << card << endl;)
		}
	}
	return true;

ERROR:
	cerr << "ERROR reading voxel file \"" << filename << "\"" << endl;
	cleanup();
	return false;
} // load

/** save
 * @return	true on success
 *		false on failure
 */
bool GVoxel::save(const char* fn)
{
	char	buffer[512];
	if (fn) filename = fn;

	FortranFile f(filename.c_str(), "wb");
	if (!f) return false;

	FortranParser parser(buffer, sizeof(buffer));

	// Title
	parser.write(title, 80);
	f.write(parser);

	// Dimensions
	parser.reset();
	parser.write(nx);
	parser.write(ny);
	parser.write(nz);
	parser.write(nregions);
	parser.write(nvoxels);
//	if (_nroi) {
//		parser.write(_nroi);
//		parser.write(_roiCombN);
//		parser.write(_roiMaxComb);
//	}
	f.write(parser);

	// Size
	parser.reset();
	parser.write(dx);
	parser.write(dy);
	parser.write(dz);
	f.write(parser);

	// write data
	f.write(_data, size()*sizeof(word));

	// write _vxl2reg (skip first)
	f.write(_vxl2reg+1, nvoxels*sizeof(word));

	for (auto card : cards) {
		parser.reset();
		parser.write(card.c_str(), 80);
		f.write(parser);
	}

	f.close();
	return true;
} // save

/** calculate voxel limits */
void GVoxel::calcLimits()
{
	xhigh = voxelx(nx);
	yhigh = voxely(ny);
	zhigh = voxelz(nz);
} // calcLimits

/** voxelijk */
bool GVoxel::voxelijk(double x, double y, double z, int* i, int* j, int* k) const
{
	if (hasMatrix()) matrix().transform(&x,&y,&z);
	// be more flexible on the limits
	double chk = dx*SMALL;

	if (Abs(x-xlow) < chk)
		*i = 0;
	else
	if (Abs(x-xhigh) < chk)
		*i = nx-1;
	else {
		*i = voxeli(x);
		if (*i<0 || *i>=nx)
			return false;
	}

	chk = dy*SMALL;
	if (Abs(y-ylow) < chk)
		*j = 0;
	else
	if (Abs(y-yhigh) < chk)
		*j = ny-1;
	else {
		*j = voxelj(y);
		if (*j<0 || *j>=ny)
			return false;
	}

	chk = dz*SMALL;
	if (Abs(z-zlow) < chk)
		*k = 0;
	else
	if (Abs(z-zhigh) < chk)
		*k = nz-1;
	else {
		*k = voxelk(z);
		if (*k<0 || *k>=nz)
			return false;
	}
	return true;
} // voxelijk

/** get _vxl2reg of data of location
 * @param x,y,z	location to return value
 * @return _vxl2reg[data[x,y,z]] at location or -1 when error
 */
int GVoxel::get(double x, double y, double z) const
{
	int i, j, k;

	if (hasMatrix()) matrix().transform(&x,&y,&z);

	// be more flexible on the limits
	double chk = dx*SMALL;

	if (Abs(x-xlow) < chk)
		i = 0;
	else
	if (Abs(x-xhigh) < chk)
		i = nx-1;
	else {
		i = voxeli(x);
		if (i<0 || i>=nx)
			return -1;
	}

	chk = dy*SMALL;
	if (Abs(y-ylow) < chk)
		j = 0;
	else
	if (Abs(y-yhigh) < chk)
		j = ny-1;
	else {
		j = voxelj(y);
		if (j<0 || j>=ny)
			return -1;
	}

	chk = dz*SMALL;
	if (Abs(z-zlow) < chk)
		k = 0;
	else
	if (Abs(z-zhigh) < chk)
		k = nz-1;
	else {
		k = voxelk(z);
		if (k<0 || k>=nz)
			return -1;
	}

	return get(i,j,k);
} // get

/** return normal vector
 * @param r	location of normal
 * @return	returned normal vector
 */
Vector GVoxel::normal(const Point& p, const Vector& d) const
{
	/* find the closest boundary */
	double remX, remY, remZ;

	Point pos;
	Vector dir;
	if (hasMatrix()) {
		// Convert to voxel coordinates
		pos = matrix() * p;
		dir = matrix() * d;
		dir.renormalize();
	} else {
		pos = p;
		dir = d;
	}

	/* First on X */
	double di = (pos.x-xlow)/dx;
	int i = Round(di);
	if (i>=0 && i<=nx)
		remX = Abs((double)i - di);
	else
		remX = INFINITE;

	/* Then on Y */
	di = (pos.y-ylow)/dy;
	i = Round(di);
	if (i>=0 && i<=ny)
		remY = Abs((double)i - di);
	else
		remY = INFINITE;

	/* Finally on Z */
	di = (pos.z-zlow)/dz;
	i = Round(di);
	if (i>=0 && i<=nz)
		remZ = Abs((double)i - di);
	else
		remZ = INFINITE;

	Vector norm;
	if (remX<remY && remX<remZ) {
		if (dir.x>0.0)
			norm = -Vector::Xo;
		else
			norm = Vector::Xo;
	} else
	if (remY<remX && remY<remZ) {
		if (dir.y>0.0)
			norm = -Vector::Yo;
		else
			norm = Vector::Yo;
	} else
	if (dir.z>0.0)
		norm = -Vector::Zo;
	else
		norm = Vector::Zo;

	if (hasMatrix()) {
		return matrix().multT(norm);
	} else
		return norm;
} // normal

/** distance find the distance to the next voxel with different
 * material
 * @param p	initial position
 * @param d	direction
 * @param tmin	minimum distance of intersection
 * @return true if an intersection is found
 */
bool GVoxel::distance(const Point& /*p*/, const Vector& /*d*/, double* /*tmin*/) const
{
	// FIXME can be cached for faster searching
	// position, direction and step
	assert(0);	// should use the VVoxel one for the moment
	return false;
} // distance

// ================================== VVoxel ===================================
/** cleanup */
void VVoxel::cleanup()
{
	roiShowClear();
	resetColor2Value();

	if (_color) delete [] _color;
	_color    = NULL;
	_nregions = 0;
} // cleanup

/** allocate color array if needed */
void VVoxel::allocate()
{
	if (_color==NULL || _nregions != nregions()) {
		cleanup();
		_nregions = nregions();
		_color = new dword[_nregions+1];
		memset(_color, 0xFF, (_nregions+1)*sizeof(dword));
	}
} // allocate

/** translate a region index to i,j,k indices
 * @param reg voxel region number
 * @param i index in X
 * @param j index in Y
 * @param k index in Z
 * @return true if translation is successful
 */
bool VVoxel::index2ijk(int reg, int& i, int& j, int& k)
{
	if (reg < 0 || reg > _nregions)
		return false;
	k    = reg / _voxel.nynx;
	reg -= k * _voxel.nynx;
	j    = reg / _voxel.nx;
	i    = reg % _voxel.nx;
	return true;
} // index2ijk

/** get color of location
 * @param x,y,z	location to return value
 * @return value at (x,y,z) location
 */
dword VVoxel::color(const double x, const double y, const double z, bool* ok) const
{
	if (_color == NULL) return COLOR_TRANSPARENT;
	int d = _voxel.get(x,y,z);
	if (d<0) {
		*ok = false;
		return COLOR_TRANSPARENT;
	} else {
		*ok = true;
		return _color[d];
	}
} // color

/** set color ith (Base 0) */
void VVoxel::color(int reg, dword c)
{
	if (!_color) allocate();
	if (reg<0 || reg>_nregions) return;
	_color[reg] = c;
} // color

/** @return a histogram of how many voxels use which color */
void VVoxel::colorHistogram(std::map<dword, int>& colorHisto)
{
	colorHisto.clear();

	for (int i=0; i<=_nregions; i++) {
		dword curCol = _color[i];
		if (colorHisto.find(curCol) == colorHisto.end())
			colorHisto.insert(std::make_pair(curCol, 1));
		else
			colorHisto[curCol] += 1;
	}
} // colorHistogram

/**
 * associate a color with a value
 */
void VVoxel::value2color(double val, dword curCol)
{
	if (curCol == COLOR_TRANSPARENT)
		val = -1e9;
	_color2valueLUT[curCol] = val;
} // value2color

/**
 * @return the value for a passed color
 */
double VVoxel::color2value(dword curCol)
{
	auto iter = _color2valueLUT.find(curCol);
	if (iter == _color2valueLUT.end())
		return -1e9;
	return iter->second;
} // color2value

/** roiColor */
dword VVoxel::roiColor(const int reg)
{
	if (_lastRoi==0 || _roiShow==NULL || _roiAlpha==0)
		return color(reg);

	Color32 cmat(color(reg));
	Color32 croi(_voxel._roiColor[_lastRoi]);
	cmat.alpha(0xff);
	croi.alpha(_roiAlpha);
	croi.blend(cmat);
	return croi;
} // roiColor

/** roiShowClear */
void VVoxel::roiShowClear()
{
	if (_roiShow)	delete [] _roiShow;
	_roiShow = NULL;
	_lastRoi = 0;
} // roiShowClear

/** roiShow */
void VVoxel::roiShow(int idx, bool b)
{
	if (idx<0 || idx>_voxel._nroi) return;
	if (_roiShow == NULL) {
		_roiShow = new bool[_voxel._nroi+1];
		memset(_roiShow, 1, sizeof(bool)*(_voxel._nroi+1));
	}
	_roiShow[idx] = b;
} // roiShow

/** return material at position ptr.
 * For the moment return the color at position ptr
 */
inline dword VVoxel::material(int ptr)
{
	dword mat = color(_voxel.get(ptr));
	if (_roiShow==NULL || mat == COLOR_TRANSPARENT) return mat;

	int comb = _voxel._roi[ptr];
	if (comb==0) return COLOR_TRANSPARENT; //mat;
	//_lastRoi = 0;
	// FIXME maybe I should keep the original structure....
	for (int i=0; i<_voxel._roiComb[comb].length; i++) {
		int roi = Abs(_voxel._roiComb[comb][i]);
		//int roi = -_voxel._roiComb[comb][i];
		if (roi>0 && !_roiShow[roi]) {
			_lastRoi = roi;
			return mat;
		}
	}
	return COLOR_TRANSPARENT;
} // material

/** distance find the distance to the next voxel with different
 * material
 * @param p	initial position
 * @param d	direction
 * @param tmin	minimum distance of intersection
 * @return true if an intersection is found
 */
bool VVoxel::distance(const Point& p, const Vector& d, double* tmin)
{
	if (_color == NULL) return false;
	//_voxel.prefetch();
	//__builtin_prefetch(_voxel._vxl2reg);
	//__builtin_prefetch(_voxel._data);

	// FIXME can be cached for faster searching
	// position, direction and step
	Point pos = p + d*(*tmin);
	Vector dir;
	if (hasMatrix()) {
		// Convert to voxel coordinates
		pos = matrix() * pos;
		dir = matrix() * d;
		dir.renormalize();
	} else
		dir = d;

	// Find location
	int i = voxeli(pos.x);
	if (i<0) {
		if (dir.x<=0.0) return false;
		else i = 0;	// correct for possible rounding error
	} else
	if (i>=nx()) {
		if (dir.x>=0.0) return false;
		else i = nx()-1;
	}

	int j = voxelj(pos.y);
	if (j<0) {
		if (dir.y<=0.0) return false;
		else j = 0;	// correct for possible rounding error
	} else
	if (j>=ny()) {
		if (dir.y>=0.0) return false;
		else j = ny()-1;
	}

	int k = voxelk(pos.z);
	if (k<0) {
		if (dir.z<=0.0) return false;
		else k = 0;	// correct for possible rounding error
	} else
	if (k>=nz()) {
		if (dir.z>=0.0) return false;
		else k = nz()-1;
	}

	int ptr = k*_voxel.nynx + j*_voxel.nx + i;
	dword mat = material(ptr);
	if (mat != COLOR_TRANSPARENT) return true;

	double tx,    ty,    tz;		// distance to next intersection with x/y/z plane
	double tdx,   tdy,   tdz;		// distance increase for every x/y/z plane
	int    stepx, stepy, stepz;		// step direction -1/0/+1
	#define ptrx stepx
	int           ptry,  ptrz;		// ptr change ptrx=stepx
	int    outx,  outy,  outz;		// ending condition

	if (dir.x > SMALL) {
		tx    = (voxelx(i+1) - pos.x) / dir.x;
		tdx   = dx() / dir.x;
		stepx = 1;
		outx  = nx();
	} else
	if (dir.x < -SMALL) {
		tx    = (voxelx(i) - pos.x) / dir.x;
		tdx   = -dx()/dir.x;
		stepx = -1;
		outx  = -1;
	} else {
		tx    = INFINITE;
		tdx   = 0.0;
		stepx = 0;
		outx  = 0;
	}

	if (dir.y > SMALL) {
		ty    = (voxely(j+1) - pos.y) / dir.y;
		tdy   = dy() / dir.y;
		stepy = 1;
		ptry  = nx();
		outy  = ny();
	} else
	if (dir.y < -SMALL) {
		ty    = (voxely(j) - pos.y) / dir.y;
		tdy   = -dy()/dir.y;
		stepy = -1;
		ptry  = -nx();
		outy  = -1;
	} else {
		ty    = INFINITE;
		tdy   = 0.0;
		stepy = 0;
		ptry  = 0;
		outy  = 0;
	}

	if (dir.z > SMALL) {
		tz    = (voxelz(k+1) - pos.z) / dir.z;
		tdz   = dz() / dir.z;
		stepz = 1;
		ptrz  = _voxel.nynx;
		outz  = nz();
	} else
	if (dir.z < -SMALL) {
		tz    = (voxelz(k) - pos.z) / dir.z;
		tdz   = -dz()/dir.z;
		stepz = -1;
		ptrz  = -_voxel.nynx;
		outz  = -1;
	} else {
		tz    = INFINITE;
		tdz   = 0.0;
		stepz = 0;
		ptrz  = 0;
		outz  = 0;
	}

	// Loop forever until we find a different material
	while (true) {
		if (tx<ty && tx<tz) {
			//__builtin_prefetch(_voxel._data+ptr+ptrx);
			i += stepx;
			if (i==outx) return false;
			ptr += ptrx;
			if (material(ptr) != mat) {
				*tmin += tx;
				return true;
			}
			tx += tdx;
		} else
		if (ty<tz) {
			//__builtin_prefetch(_voxel._data+ptr+ptry);
			j += stepy;
			if (j==outy) return false;
			ptr += ptry;
			if (material(ptr) != mat) {
				*tmin += ty;
				return true;
			}
			ty += tdy;
		} else {
			//__builtin_prefetch(_voxel._data+ptr+ptrz);
			k += stepz;
			if (k==outz) return false;
			ptr += ptrz;
			if (material(ptr) != mat) {
				*tmin += tz;
				return true;
			}
			tz += tdz;
		}
	}
} // distance
