/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2011
 */

#include "vzone.h"

class GZone;

/** FLUKA viewport region definition */
void VRegion::init(GRegion* reg, GeometryKernel* k)
{
	_region     = reg;
	_generation = -1;
	_kernel     = k;
	_location   = false;
	_in         = NULL;

	clear();	// FIXME clears the zones.
			// Not really needed we could have reused the zones!
	if (_region) {
		_hash = _region->hash();
		for (int i=0; i<reg->size(); i++)
			_zones.add(new VZone(reg->_zones[i],this));
	}
} // init

/** clear */
void VRegion::clear()
{
	for (int i=0; i<_zones.size(); i++)
		if (_zones[i]) delete _zones[i];
	_zones.clear();
} // clear

/** updateLocation
 *
 * find the location of the Region on the Z=0 plane and within the viewport
 * coordinates
 *
 * FIXME: Improvement on STD expressions we can ignore directly the zones
 *        with a negative result, and remove the bodies
 *        Probably add a reference count to the bodies!
 *
 * We have 3 possible locations inside(1) / outside(0) / unknown(?:2)
 *		  a | b
 *	0 | 0 = 0	? | 0 = ?
 *	0 | 1 = 1	? | 1 = 1
 *	1 | 0 = 1	0 | ? = ?
 *	1 | 1 = 1	1 | ? = 1
 */
void VRegion::updateLocation(const ViewPort& view)
{
	_location = false;
	_in       = NULL;
	bool allknown = true;

	for (int iz = 0; iz < _zones.size(); iz++) {
		VZone* z = _zones[iz];
		z->updateLocation(view);
		if (z->location()) {
			if (z->in()) {
				_location = true;
				_in = z;
			}
		} else
			allknown = false;
	}
	if (allknown && !_location) {
		_location = true;
		_in = NULL;
	}
} // updateLocation

/** inside2D
 * WARNING this routine is optimized for points close to the 2D viewport ONLY!
 * @param x,y,z     location of point to search
 * @param u,v,w     direction in case on boundary
 * @return zone     return zone that location is inside otherwise NULL
 */
VZone* VRegion::inside2D(GeometryEngine* engine,
			 const double x, const double y, const double z,
			 const double u, const double v, const double w) const
{
	if (_location) return _in;
	for (int iz = 0; iz < _zones.size(); iz++)
		if (_zones[iz]->inside2D(engine, x,y,z, u,v,w))
			return _zones[iz];
	return NULL;
} // inside2D

/** inside */
VZone* VRegion::inside(GeometryEngine* engine,
			  const double x, const double y, const double z,
			  const double u, const double v, const double w,
			  const double t) const
{
	for (int iz = 0; iz < _zones.size(); iz++) {
		if (_zones[iz]->inside(engine, x,y,z, u,v,w, t))
			return _zones[iz];
	}
	return NULL;
} // inside

/** @return memory used by region */
size_t VRegion::memory() const
{
	size_t zmem = sizeof(VRegion);
	for (int iz = 0; iz < _zones.size(); iz++)
		zmem += _zones[iz]->memory();
	return zmem;
} // memory

/** operator << */
std::ostream& operator << (std::ostream& s, const VRegion& region)
{
	s << "VRegion: " << region.name() << std::endl;
#if 0
	for (int i=0; i<region.n; i++) s << region[i];
	//for (int i=0; i<region.expr.count(); i++) {
		VBody* body = (VBody*)region.expr[i];
		if (body == tplus)		s << "+ " << std::endl;
		else if (body == tminus)	s << "- " << std::endl;
		else if (body == tunion)	s << "| " << std::endl;
		else if (body == tuniverse)	s << "@ " << std::endl;
		else
			s << body->name() << " ";
	}
#endif
	return s;
} /* operator << */
