/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2010
 */

#include <Python.h>

#include <stdio.h>
#include <fnmatch.h>

#include <sstream>

#include "os.h"
#include "gmesh.h"
#include "gobject.h"
#include "pyutils.h"
#include "tool/token.h"
#include "math/obbox.h"
#include "math/matrix4.h"
#include "geometry/geo.h"
#include "geometry/gbody.h"
#include "geometry/voxel.h"
#include "geometry/gregion.h"
#include "geometry/geometry.h"

#include "geometryobject.h"	// it has to be the first include!

#ifdef MEM
#include "tool/memory.h"
#endif

#if 0
//#ifdef MEM
// activate them only for memory leaks debugging
#include "data/eventbin.h"
#include "data/userdump.h"
#include "data/usrbin.h"
#include "geometry/arbbody.h"
#include "geometry/boxbody.h"
#include "geometry/bvh.h"
#include "geometry/cbody.h"
#include "geometry/csg.h"
#include "geometry/cylbody.h"
#include "geometry/edge.h"
#include "geometry/engine.h"
#include "geometry/face.h"
#include "geometry/filter.h"
//#include "geometry/gbody.h"
//#include "geometry/geo.h"
#include "geometry/geocache.h"
//#include "geometry/geometry.h"
//#include "geometry/gregion.h"
#include "geometry/gzone.h"
#include "geometry/kernel.h"
#include "geometry/light.h"
#include "geometry/line.h"
#include "geometry/material.h"
#include "geometry/mesh.h"
#include "geometry/planebody.h"
#include "geometry/polyline.h"
#include "geometry/quabody.h"
#include "geometry/ray.h"
#include "geometry/segment.h"
#include "geometry/spherebody.h"
#include "geometry/stl.h"
#include "geometry/tetra.h"
#include "geometry/tetramesh.h"
#include "geometry/tetrameshio.h"
#include "geometry/torusbody.h"
#include "geometry/vbody.h"
#include "geometry/vertex.h"
//#include "geometry/voxel.h"
#include "geometry/vregion.h"
#include "geometry/vzone.h"
//#include "geometryobject.h"
#include "gldraw/glmesh.h"
//#include "gmesh.h"
//#include "gobject.h"
#include "graph/bfont.h"
#include "graph/color.h"
#include "graph/dxfexport.h"
#include "graph/exportbase.h"
#include "graph/painter.h"
#include "graph/palette.h"
#include "graph/svgexport.h"
#include "graph/viewport.h"
#include "graph/zpainter.h"
#include "math/bbox.h"
#include "math/bit.h"
#include "math/bmath.h"
#include "math/conic.h"
#include "math/eps.h"
#include "math/format.h"
#include "math/histogram.h"
#include "math/histogram2d.h"
#include "math/matrix.h"
#include "math/matrix2.h"
#include "math/matrix3.h"
//#include "math/matrix4.h"
//#include "math/obbox.h"
#include "math/plane.h"
#include "math/point.h"
#include "math/polynomial.h"
#include "math/quad.h"
#include "math/quaternion.h"
#include "math/random.h"
#include "math/randsampler.h"
#include "math/scatter.h"
#include "math/sphere.h"
#include "math/spline.h"
#include "math/surface.h"
#include "math/vector.h"
#include "math/vector2d.h"
//#include "os.h"
#include "physics/particle.h"
#include "physics/units.h"
//#include "pyutils.h"
#include "tool/array.h"
#include "tool/bintree.h"
#include "tool/bstring.h"
#include "tool/cache.h"
#include "tool/fortran.h"
#include "tool/list.h"
#include "tool/sortvector.h"
#include "tool/stack.h"
#include "tool/stat.h"
#include "tool/stream.h"
#include "tool/threadpool.h"
#include "tool/timer.h"
#include "tool/token.h"
#include "viewer/d2layer.h"
#include "viewer/d3layer.h"
#include "viewer/decorationlayer.h"
#include "viewer/exportlayer.h"
#include "viewer/imagelayer.h"
#include "viewer/latticelayer.h"
#include "viewer/layer.h"
#include "viewer/nox.h"
#include "viewer/palettelayer.h"
#include "viewer/userdumplayer.h"
#include "viewer/usrbinlayer.h"
#include "viewer/viewer.h"
#include "viewer/voxellayer.h"
#include "viewerobject.h"
#include "xdraw/ocube.h"
#include "xdraw/xdraw.h"
#endif

using std::cout;
using std::endl;

/* _deleteObjects */
static void _deleteObjects(GeometryObject* self)
{
	// delete all objects
	for (auto object : self->objects)
		delete object;
	self->objects.clear();
} // _deleteObjects

/* Py_GBody
 * @param self	Geometry object
 * @param obj	object to convert to body can be string or id
 * @return body
 */
GBody* Py_GBody(GeometryObject* self, PyObject* obj)
{
	if (obj==NULL) return NULL;
	if (PyUnicode_Check(obj)) {
		const char* name = PyUnicode_AsUTF8(obj);
		auto body = self->geometry->getBody(name);
		if (body) return body;
		PyErr_Format(PyExc_KeyError, "Body \'%s\' not found", name);
	} else
	if (PyLong_Check(obj)) {
		size_t id = PyLong_AsSsize_t(obj);
		if (id < self->geometry->nBodies()) {
			auto body = self->geometry->getBody(id);
			if (body) return body;
		}
		PyErr_Format(PyExc_IndexError, "Body #%d not found",id);
	} else
		PyErr_SetString(PyExc_TypeError, "Invalid body type, string or integer expected");
	return NULL;
} // Py_GBody

/* Py_GRegion
 * @param self	Geometry object
 * @param obj	object to convert to region can be string or id
 * @return region
 */
GRegion* Py_GRegion(GeometryObject* self, PyObject* obj)
{
	if (obj==NULL) return NULL;
	if (PyUnicode_Check(obj)) {
		const char* name = PyUnicode_AsUTF8(obj);
		auto region = self->geometry->getRegion(name);
		if (region) return region;
		PyErr_Format(PyExc_KeyError, "Region \'%s\' not found", name);
	} else
	if (PyLong_Check(obj)) {
		size_t id = PyLong_AsLong(obj);
		if (id < self->geometry->nRegions()) {
			auto region = self->geometry->getRegion(id);
			if (region) return region;
		}
		PyErr_Format(PyExc_IndexError, "Region #%d not found",id);
	} else
		PyErr_SetString(PyExc_TypeError, "Invalid region type, string or integer expected");
	return NULL;
} // Py_GRegion

/* Py_Material
 * @param self	Geometry object
 * @param obj	object to convert to material can be string or id
 * @return material
 */
Material* Py_Material(GeometryObject* self, PyObject* obj)
{
	if (obj==NULL) return NULL;
	if (PyUnicode_Check(obj)) {
		const char* name = PyUnicode_AsUTF8(obj);
		auto material = self->geometry->getMaterial(name);
		if (material) return material;
		PyErr_Format(PyExc_KeyError, "Material \'%s\' not found", name);
	} else
	if (PyLong_Check(obj)) {
		size_t id = PyLong_AsLong(obj);
		if (id < self->geometry->nMaterials()) {
			auto material = self->geometry->getMaterial(id);
			if (material) return material;
		}
		PyErr_Format(PyExc_IndexError, "Material #%d not found",id);
	} else
		PyErr_SetString(PyExc_TypeError, "Invalid material type, string or integer expected");
	return NULL;
} // Py_Material

/* Py_Object
 * @param self	Geometry object
 * @param obj	object to convert to body can be string or id
 * @return object
 */
GObject* Py_Object(GeometryObject* self, PyObject* obj)
{
	if (obj==NULL) return NULL;
	if (PyUnicode_Check(obj)) {
		const char* name = PyUnicode_AsUTF8(obj);
		for (auto object : self->objects)
			if (object->name() == name)
				return object;
		PyErr_Format(PyExc_KeyError, "Object \'%s\' not found", name);
	} else
	if (PyLong_Check(obj)) {
		size_t id = PyLong_AsLong(obj);
		if (id<self->objects.size()) {
			auto object = self->objects[id];
			if (object) return object;
		}
		PyErr_Format(PyExc_IndexError, "Object #%d not found",id);
	} else
		PyErr_SetString(PyExc_TypeError, "Invalid object type, string or integer expected");
	return NULL;
} // Py_Object

/** @return zone as a tuple (format, list of tokens) */
PyObject* Py_ZoneExpr(GZone* zone, Geometry::Mode /*mode*/)
{
	if (!zone->size()) Py_RETURN_NONE;

	PyObject* expr = PyList_New(zone->size());
	for (int j=0; j<zone->size(); j++) {
		const GBody* body = (*zone)[j];
		const char* token = body->cname();
		//if (In(mode, Geometry::Mode::Mcnp, Geometry::Mode::Phits)) {
		//	if (body==&GBody::tplus) {
		//		if ((*zone)[j+1] == &GBody::tleft)
		//			continue;	// no need of + in front of (-paren
		//		token = "-";
		//	} else
		//	if (body==&GBody::tminus) {
		//		if ((*zone)[j+1] == &GBody::tleft)
		//			token = "#";
		//		else
		//			token = "+";
		//	} else
		//	if (body==&GBody::tunion) {
		//		token = ":";
		//	}
		//}
		PyList_SetItem(expr, j, PyUnicode_FromString(token));
	}
	return Py_BuildValue("iO",(int)zone->format(),expr);
} // Py_ZoneExpr

/** @return a region expression as a list of zones @see Py_ZoneExpr */
PyObject* Py_RegionExpr(GRegion* region, Geometry::Mode mode)
{
	// return selected zone
	if (!region->size()) Py_RETURN_NONE;

	PyObject* zones = PyList_New(0);
	for (auto zone : region->zones()) {
		PyObject* z = Py_ZoneExpr(zone, mode);
		if (z) PyList_Append(zones, z);
	}
	return zones;
} // Py_RegionExpr

// ================================= Geometry ==================================
/* --- Geometry_new --- */
static PyObject* Geometry_new(PyTypeObject* type, PyObject* /*args*/,
				PyObject* /*kwds*/)
{
	GeometryObject* self;
	self = (GeometryObject*)type->tp_alloc(type, 0);
	if (self == NULL) return NULL;

	self->geometry       = new Geometry();
	self->cursor         = Vector::O;
	self->cursorType     = CursorType::CCursor;
	self->cursorId       = -1;
	self->cursorSize     = 40;
	self->cursorMoveSize = 10;
	self->trackballSize  = 0.80;
	self->snapDistance   = 4.0;
	self->snapAngle      = Rad(5.0);

	self->ruler[0] = new GRuler("X");
	self->ruler[0]->color  = 0xFF00FF; //COLOR_X_DARK;
	self->ruler[0]->anchor = Anchor_C;

	self->ruler[1] = new GRuler("Y");
	self->ruler[1]->color  = 0xFF00FF; //COLOR_Y_DARK;
	self->ruler[1]->anchor = Anchor_C;

	self->ruler[2] = new GRuler("Z");
	self->ruler[2]->color  = 0xFF00FF; //COLOR_Z_DARK;
	self->ruler[2]->anchor = Anchor_C;

	return (PyObject*)self;
} // Geometry_new

/* --- Geometry_dealloc --- */
static void Geometry_dealloc(GeometryObject* self)
{
	for (int i=0; i<self->rulers; i++)
		delete self->ruler[i];
	_deleteObjects(self);
	delete self->geometry;

	Py_TYPE(self)->tp_free((PyObject*)self);
} // Geometry_dealloc

#if 0
//#ifdef MEM
/* --- Geometry_destroy --- */
static PyObject* Geometry_destroy(GeometryObject* self, PyObject* Py_UNUSED(ignored))
{
#define PRINT_SIZE(a)	printf("  %-16s %5ld\n",#a,sizeof(a));
	printf("class size for reference:\n");

	PRINT_SIZE(ArrayEmpty);
	PRINT_SIZE(BBox);
	PRINT_SIZE(BFont);
	PRINT_SIZE(BaseSplineNode);
	PRINT_SIZE(BinLeaf);
	PRINT_SIZE(BinTree);
	PRINT_SIZE(Body3DFeeder);
	PRINT_SIZE(Body3DWorker);
	PRINT_SIZE(BodyFeeder);
	PRINT_SIZE(BodyWorker);
	PRINT_SIZE(CBody);
	PRINT_SIZE(CBodyOrderAccel);
	PRINT_SIZE(CSG);
//	PRINT_SIZE(Caloron);
	PRINT_SIZE(CardinalSpline);
	PRINT_SIZE(Color);
	PRINT_SIZE(Conic);
	PRINT_SIZE(D2Layer);
	PRINT_SIZE(D3Layer);
	PRINT_SIZE(DXFExport);
	PRINT_SIZE(DecorationLayer);
	PRINT_SIZE(Edge);
	PRINT_SIZE(Eventbin);
	PRINT_SIZE(ExportBase);
	PRINT_SIZE(ExportLayer);
	PRINT_SIZE(Face);
	PRINT_SIZE(FortranFile);
	PRINT_SIZE(FortranParser);
	PRINT_SIZE(GARBBody);
	PRINT_SIZE(GArrow);
	PRINT_SIZE(GBOXBody);
	PRINT_SIZE(GBeam);
	PRINT_SIZE(GBody);
	PRINT_SIZE(GELLBody);
	PRINT_SIZE(GERRBody);
	PRINT_SIZE(GInfEllCylBody);
	PRINT_SIZE(GLMesh);
	PRINT_SIZE(GLight);
	PRINT_SIZE(GMesh);
	PRINT_SIZE(GOPRBody);
	PRINT_SIZE(GObject);
	PRINT_SIZE(GPLABody);
	PRINT_SIZE(GPoint);
	PRINT_SIZE(GQUABody);
	PRINT_SIZE(GRCCBody);
	PRINT_SIZE(GRECBody);
	PRINT_SIZE(GRegion);
	PRINT_SIZE(GRotdefi);
	PRINT_SIZE(GRuler);
	PRINT_SIZE(GSPHBody);
	PRINT_SIZE(GSpline);
	PRINT_SIZE(GTRCBody);
	PRINT_SIZE(GTorusBody);
	PRINT_SIZE(GVoxel);
	PRINT_SIZE(GWEDBody);
	PRINT_SIZE(GZone);
	PRINT_SIZE(Geometry);
	PRINT_SIZE(GeometryEngine);
	PRINT_SIZE(GeometryKernel);
	PRINT_SIZE(GeometryViewer);
	PRINT_SIZE(H1D);
	PRINT_SIZE(Histogram);
	PRINT_SIZE(IPoint);
	PRINT_SIZE(ImageLayer);
	PRINT_SIZE(KahanSum);
	PRINT_SIZE(LatticeLayer);
	PRINT_SIZE(Layer);
	PRINT_SIZE(Material);
	PRINT_SIZE(Matrix);
	PRINT_SIZE(Matrix2);
	PRINT_SIZE(Matrix3);
	PRINT_SIZE(Matrix4);
	PRINT_SIZE(Mesh);
//	PRINT_SIZE(Neutron);
	PRINT_SIZE(OBBox);
	PRINT_SIZE(Painter);
	PRINT_SIZE(Palette);
	PRINT_SIZE(PaletteLayer);
//	PRINT_SIZE(Particle);
	PRINT_SIZE(Plane);
	PRINT_SIZE(Point);
	PRINT_SIZE(Point2D);
	PRINT_SIZE(Polyline);
	PRINT_SIZE(Polynomial);
	PRINT_SIZE(PolynomialSolver);
	PRINT_SIZE(Quad);
	PRINT_SIZE(ROICombination);
	//PRINT_SIZE(ROIPlanar);		// FIXME why is not working?
	//PRINT_SIZE(ROIPlanarSlice);
	PRINT_SIZE(Random);
	PRINT_SIZE(Ray);
	PRINT_SIZE(RaySegment);
	PRINT_SIZE(STL);
	PRINT_SIZE(SVGExport);
	PRINT_SIZE(Scatter);
	PRINT_SIZE(SplineNode);
	PRINT_SIZE(Stream);
	PRINT_SIZE(Tetra);
	PRINT_SIZE(TetraMesh);
	PRINT_SIZE(TetraMeshIO);
	PRINT_SIZE(ThreadPool);
	PRINT_SIZE(ThreadPoolFeeder);
	PRINT_SIZE(ThreadPoolWorker);
	PRINT_SIZE(Timer);
	PRINT_SIZE(Token);
	PRINT_SIZE(UserDump);
	PRINT_SIZE(UserDumpLayer);
	PRINT_SIZE(Usrbin);
	PRINT_SIZE(UsrbinLayer);
	PRINT_SIZE(VBody);
	PRINT_SIZE(VRegion);
	PRINT_SIZE(VVoxel);
	PRINT_SIZE(VZone);
	PRINT_SIZE(Vector);
	PRINT_SIZE(Vector2D);
	PRINT_SIZE(VectorSplineNode);
	PRINT_SIZE(Vertex);
	PRINT_SIZE(Segment);
	PRINT_SIZE(ViewPort);
	PRINT_SIZE(VolumeFeeder);
	PRINT_SIZE(VolumeWorker);
	PRINT_SIZE(VoxelLayer);
	PRINT_SIZE(XYPair);
	PRINT_SIZE(ZoneOfPoint);
// struct
	printf("\nstruct size for reference:\n");
	PRINT_SIZE(GeometryObject);
	PRINT_SIZE(GeometryViewer);
	PRINT_SIZE(ViewportLine);

	PRINT_SIZE(UserDumpTrackPos);
	PRINT_SIZE(UserDumpSourceParticle);
	PRINT_SIZE(Light);
	PRINT_SIZE(VLight);
	PRINT_SIZE(ClipRegion);
	PRINT_SIZE(drand48_data);
//	PRINT_SIZE(Memory);
	PRINT_SIZE(Color32);

	fflush(stdout);
	Memory::dump();
	Geometry_dealloc(self);

	if (!Memory::fini())
		printf("*** No memory leak\n");
	exit(0);
	Py_RETURN_NONE;
} // Geometry_destroy
#endif

/* --- Geometry_bbox --- */
static PyObject* Geometry_bbox(GeometryObject* self, PyObject* args)
{
	char* type;
	PyObject* obj = NULL;
	BBox bbox;
	int n=-1;

	if (!PyArg_ParseTuple(args, "sO|i", &type, &obj, &n)) return NULL;
	if (type[0]==0) {
		PyErr_SetString(PyExc_TypeError, "Invalid object type body, zone, region expected");
		return NULL;
	}

	if (type[0]=='b' || type[0]=='B') {
		GBody* body = Py_GBody(self,obj);
		if (body==NULL) return NULL;
		bbox = body->bbox();
	} else
	if (type[0]=='r' || type[0]=='R' ||
	    type[0]=='z' || type[0]=='Z') {
		GRegion* region = Py_GRegion(self,obj);
		region->updateBBox();
		if (region==NULL) return NULL;
		if (type[0]=='r' || type[0]=='R' || n==-1)
			bbox = region->bbox();
		else {
			if (n<0 || n>=region->size()) {
				PyErr_SetString(PyExc_IndexError, "Invalid zone index");
				return NULL;
			}
			bbox = (*region)[n]->bbox();
		}
	} else
	if (type[0]=='o' || type[0]=='O') {
		GObject* object = Py_Object(self,obj);
		bbox = object->bbox();
	}

	if (bbox.isValid())
		return Py_BuildValue("[dddddd]",
				bbox.low().x,  bbox.low().y,  bbox.low().z,
				bbox.high().x, bbox.high().y, bbox.high().z);
	else
		Py_RETURN_NONE;
} // Geometry_bbox

/* --- Geometry_bodyVar --- */
static PyObject* Geometry_bodyVar(GeometryObject* self, GBody* body,
				const char* var, PyObject* value)
{
	if (!strcmp(var, "name")) {
		if (value==NULL)
			return PyUnicode_FromString(body->cname());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set body name");
			return NULL;
		}
	} else
	if (!strcmp(var, "id")) {
		if (value==NULL)
			return PyLong_FromLong(body->id());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set body id");
			return NULL;
		}
	} else
	if (!strcmp(var, "show")) {
		if (value==NULL)
			return PyLong_FromLong(body->show);
		else
			body->show = (int)PyLong_AsLong(value);
	} else
	if (!strcmp(var, "type")) {
		if (value==NULL)
			return PyUnicode_FromString(body->typeStr());
		else {
			PyErr_Format(PyExc_TypeError, "\'%s\' cannot be set", var);
			return NULL;
		}
	} else
	if (!strcmp(var, "color")) {
		if (value==NULL)
			return PyLong_FromLong(body->color());
		else
			body->color((dword)PyLong_AsUnsignedLongMask(value));
	} else
	if (!strcmp(var, "linewidth")) {
		if (value==NULL)
			return PyLong_FromLong(body->lineWidth());
		else
			body->lineWidth((dword)PyLong_AsLong(value));
	} else
	if (!strcmp(var, "what")) {
		double what[30];
		if (value==NULL) {
			int n = body->get(what);
			assert(n>=0);
			PyObject* obj = PyList_New(n);
			for (int i=0; i<n; i++)
				PyList_SetItem(obj, i, PyFloat_FromDouble(what[i]));
			return obj;
		} else {
			if (!PyList_Check(value)) {
				PyErr_SetString(PyExc_TypeError,
					"Invalid type, list expected");
				return NULL;
			}

			if (PyList_GET_SIZE(value) > (ssize_t)SIZE(what)) {
				PyErr_SetString(PyExc_IndexError,
					"Too large what list");
				return NULL;
			}
			ssize_t i;
			memset(what, 0, sizeof(what));
			for (i=0; i<PyList_GET_SIZE(value); i++)
				what[i] = PyFloat_AsDouble(PyList_GetItem(value, i));

			try {
				body->set(PyList_GET_SIZE(value), what);
			} catch (std::string& err) {
				PyErr_SetString(PyExc_Warning, err.c_str());
			}
			try {
				body->create();
			} catch (...) {
				PyErr_SetString(PyExc_SyntaxError,
					"Error in body parameters");
			}
			if (PyErr_Occurred()) return NULL;
		}
	} else
	if (!strcmp(var, "matrix")) {
		if (value) {
			if (PyList_Check(value)) {
				Matrix4 matrix;
				PyList_AsMatrix4(value, matrix);
				try {
					body->matrix(matrix);
					body->transform();
				} catch (...) {
					PyErr_SetString(PyExc_SyntaxError,
						"Invalid body rotation matrix");
					return NULL;
				}
			} else
			if (value==Py_None) {
				body->clearMatrix();
				body->transform();
			} else {
				PyErr_SetString(PyExc_TypeError,
					"Invalid body rotation matrix");
				return NULL;
			}
		} else
			return PyList_FromMatrix4(body->matrix());
	} else
	if (!strcmp(var, "bbox")) {	// Get or set user defined bbox
		if (value==NULL) {
			BBox bb = body->bbox();
			Point lw = bb.low();
			Point hg = bb.high();

			PyObject* bbox  = PyList_New(6);
			PyList_SetItem(bbox, 0, PyFloat_FromDouble(lw.x));
			PyList_SetItem(bbox, 1, PyFloat_FromDouble(hg.x));
			PyList_SetItem(bbox, 2, PyFloat_FromDouble(lw.y));
			PyList_SetItem(bbox, 3, PyFloat_FromDouble(hg.y));
			PyList_SetItem(bbox, 4, PyFloat_FromDouble(lw.z));
			PyList_SetItem(bbox, 5, PyFloat_FromDouble(hg.z));

			return bbox;
		} else
		if (PyList_Check(value)) {
			double xmin = PyFloat_AsDouble(PyList_GetItem(value,0));
			double xmax = PyFloat_AsDouble(PyList_GetItem(value,1));
			double ymin = PyFloat_AsDouble(PyList_GetItem(value,2));
			double ymax = PyFloat_AsDouble(PyList_GetItem(value,3));
			double zmin = PyFloat_AsDouble(PyList_GetItem(value,4));
			double zmax = PyFloat_AsDouble(PyList_GetItem(value,5));
			body->bbox(BBox(xmin,xmax, ymin,ymax, zmin,zmax));
		}
	} else
	if (!strcmp(var, "pos")) {	// Move absolute
		if (value==NULL)
			return Py_Vector(body->position());
		else {
			Point r = Py_GetVector(value);
			if (PyErr_Occurred()) return NULL;
			body->position(r);
			body->create(true);
			self->geometry->invalidateBody(body);
		}
	} else
	if (!strcmp(var, "save"))
		body->save();
	else
	if (!strcmp(var, "restore")) {
		body->restore();
		body->create(true);
		self->geometry->invalidateBody(body);
	} else
	if (!strcmp(var, "savedpos"))	// return saved position
		return Py_Vector(body->savedPosition());
	else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", var);
		return NULL;
	}
	Py_RETURN_NONE;
} // Geometry_bodyVar

/* --- Geometry_body --- */
static PyObject* Geometry_body(GeometryObject* self, PyObject* args)
{
	PyObject* obj;
	char* var;
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "Os|O", &obj, &var, &value)) return NULL;

	if (!strcmp(var,"add")) {
		if (!PyUnicode_Check(obj)) {
			PyErr_SetString(PyExc_TypeError,"String expected as body name");
			return NULL;
		}
		if (!PyUnicode_Check(value)) {
			PyErr_SetString(PyExc_TypeError,"String expected as body type");
			return NULL;
		}
		const char* name = PyUnicode_AsUTF8(obj);
		const char* type = PyUnicode_AsUTF8(value);
		// Create body
		try {
			GBody* body = self->geometry->addBody(name, type);
			return PyLong_FromLong(body->id());
		} catch (std::string err) {
			PyErr_SetString(PyExc_SyntaxError, err.c_str());
			return NULL;
		}
	} else
	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (auto body : self->geometry->bodies) {
			if (!fnmatch(pattern, body->cname(), 0)) {
				PyObject* ret = Geometry_bodyVar(self, body, var, value);
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else
	if (PyList_Check(obj)) {
		for (ssize_t i=0; i<PyList_GET_SIZE(obj); i++) {
			GBody* body = Py_GBody(self, PyList_GetItem(obj,i));
			if (body==NULL) return NULL;
			PyObject* ret = Geometry_bodyVar(self, body, var, value);
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
	} else
	if (PyTuple_Check(obj)) {
		for (ssize_t i=0; i<PyTuple_GET_SIZE(obj); i++) {
			GBody* body = Py_GBody(self, PyTuple_GetItem(obj,i));
			if (body==NULL) return NULL;
			PyObject* ret = Geometry_bodyVar(self, body, var, value);
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
	} else  {
		GBody* body = Py_GBody(self, obj);
		if (body==NULL) return NULL;
		return Geometry_bodyVar(self, body, var, value);
	}
	Py_RETURN_NONE;		// To keep compiler happy
} // Geometry_body

/* --- Geometry_cleanup --- */
static PyObject* Geometry_cleanup(GeometryObject* self, PyObject* Py_UNUSED(ignored))
{
	_deleteObjects(self);	// delete all objects

	// clean up geometry
	//self->geometry->voxel.cleanup();	Don't clean up heavy objects, needs time
	self->geometry->cleanup();
	Py_RETURN_NONE;
} // Geometry_cleanup

/* --- Geometry_cursor --- */
static PyObject* Geometry_cursor(GeometryObject* self, PyObject* args)
{
	char* name;
	PyObject* value=NULL;

	if (!PyArg_ParseTuple(args, "s|O", &name, &value)) return NULL;
	if (!strcmp(name, "object")) {	// Move relative from the last save position
		if (value) {
			self->cursorShow = false;
			if (value!=Py_None) {
				GObject* object = Py_Object(self,value);
				if (object==NULL) return NULL;
				self->cursorShow = true;
				self->cursorType = CursorType::CObject;
				self->cursor     = object->position();
				self->cursorId   = object->id();
			}
		}
	} else
	if (!strcmp(name, "body")) {	// Move relative from the last save position
		if (value) {
			if (value==Py_None) {
				self->geometry->editBody = NULL;
				self->cursorShow = false;
			} else {
				self->geometry->editBody = Py_GBody(self, value);
				if (self->geometry->editBody==NULL) {
					self->cursorShow = false;
					return NULL;
				}
				self->cursorShow = true;
				self->cursorType = CursorType::CBody;
				self->cursor     = self->geometry->editBody->position();
				self->cursorId   = self->geometry->editBody->id();
			}
		} else
		if (self->geometry->editBody)
			return PyLong_FromLong(self->geometry->editBody->id());
	} else
	if (!strcmp(name, "region")) {	// Move relative from the last save position
		if (value) {
			if (value==Py_None)
				self->cursorShow = false;
			else {
				GRegion* region  = Py_GRegion(self, value);
				if (region==NULL) {
					self->cursorShow = false;
					return NULL;
				}
				self->cursorShow = true;
				self->cursorType = CursorType::CRegion;
				self->cursor     = region->bbox().centroid();
				self->cursorId   = region->id();
#ifdef _DUMP
				cout << "Region:"   << region->name() << endl;
				cout << "\tBBox:"   << region->bbox() << endl;
				//cout << "\tOBBox:"  << region->obbox() << endl;
				cout << "\tCursor:" << self->cursor << endl;
#endif
			}
		}
	} else
	if (!strcmp(name, "cursor")) {
		if (value) {
			if (value==Py_None || value==Py_False)
				self->cursorShow = false;
			else
			if (value==Py_True)
				self->cursorShow = true;
			else {
				self->cursor     = Py_GetVector(value);
				self->cursorType = CursorType::CCursor;
				if (PyErr_Occurred()) {
					self->cursorShow = false;
					return NULL;
				}
				self->cursorShow = true;
				self->cursorId   = -1;
			}
		} else
		if (self->cursorShow)
			return Py_Vector(self->cursor);
	} else
	if (!strcmp(name, "type")) {
		switch (self->cursorType) {
			case CursorType::CCursor:
				return PyUnicode_FromString("cursor");
			case CursorType::CBody:
				return PyUnicode_FromString("body");
			case CursorType::CRegion:
				return PyUnicode_FromString("region");
			case CursorType::CZone:
				// FIXME if it returns "zone" then the GeometryViewer:loadProject()
				// will fail to set it back
				return PyUnicode_FromString("zone");
			case CursorType::CObject:
				return PyUnicode_FromString("object");
		}
	} else
	if (!strcmp(name, "id")) {
		return PyLong_FromLong(self->cursorId);
	} else
	if (!strcmp(name, "clear")) {
		self->cursorShow = false;
		self->geometry->editBody = NULL;
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", name);
		return NULL;
	}
	Py_RETURN_NONE;
} // Geometry_cursor

/* --- Geometry_color --- */
static PyObject* Geometry_color(GeometryObject* self, PyObject* args)
{
#define NOTDEFINED	-999999999
	char* name;
	int   value=NOTDEFINED;

	if (!PyArg_ParseTuple(args, "s|i", &name, &value)) return NULL;

	if (!strcmp(name,"background")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->backgroundColor() & FLAG_COLORMASK);
		else {
			self->geometry->backgroundColor(value);
		}
	} else
	if (!strcmp(name,"bodybboxin")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->bodyBBoxInColor());
			//return PyLong_FromLong(self->geometry->bodyBBoxInColor);
		else
			self->geometry->bodyBBoxInColor(value);
			//self->geometry->bodyBBoxInColor = value;
	} else
	if (!strcmp(name,"bodybboxout")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->bodyBBoxOutColor());
			//return PyLong_FromLong(self->geometry->bodyBBoxOutColor);
		else
			self->geometry->bodyBBoxOutColor(value);
			//self->geometry->bodyBBoxOutColor = value;
	} else
	if (!strcmp(name,"error")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->errorColor & FLAG_COLORMASK);
		else
			self->geometry->errorColor = value;
	} else
	if (!strcmp(name,"gridtext")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->gridTextColor & FLAG_COLORMASK);
		else
			self->geometry->gridTextColor = value | FLAG_FINAL;
	} else
	if (!strcmp(name,"label")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->labelColor & FLAG_COLORMASK);
		else
			self->geometry->labelColor = value | FLAG_FINAL;
	} else
	if (!strcmp(name,"lattice")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->latticeColor & FLAG_COLORMASK);
		else {
			self->geometry->latticeColor = value;
			self->geometry->makeLatticeColor();
		}
	} else
	if (!strcmp(name,"palettetext")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->paletteTextColor & FLAG_COLORMASK);
		else
			self->geometry->paletteTextColor = value;
	} else
	if (!strcmp(name,"regionbbox")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->regionBBoxColor());
			//return PyLong_FromLong(self->geometry->regionBBoxColor);
		else
			self->geometry->regionBBoxColor(value);
			//self->geometry->regionBBoxColor = value;
	} else
	if (!strcmp(name,"regionerror")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->regionErrorColor & FLAG_COLORMASK);
		else
			self->geometry->regionErrorColor = value;
	} else
	if (!strcmp(name,"region")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->regionColor & FLAG_COLORMASK);
		else
			self->geometry->regionColor = value;
	} else
	if (!strcmp(name,"select")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->selectColor & FLAG_COLORMASK);
		else {
			self->geometry->selectColor = value;
			self->geometry->select3DColor(value);
			//self->geometry->select3DColor = value;
			for (int i=0; i<self->rulers; i++)
				self->ruler[i]->color = value;
			GObject::selectColor = value;
		}
	} else
	if (!strcmp(name,"title")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->titleColor & FLAG_COLORMASK);
		else
			self->geometry->titleColor = value;
	} else
	if (!strcmp(name,"vertex")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->vertexColor & FLAG_COLORMASK);
		else
			self->geometry->vertexColor = value;
	} else
	if (!strcmp(name,"visible")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->visibleColor & FLAG_COLORMASK);
		else
			self->geometry->visibleColor = value;
	} else
	if (!strcmp(name,"voxel")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->voxelColor & FLAG_COLORMASK);
		else {
			self->geometry->voxelColor = value;
			self->geometry->makeLatticeColor();
		}
	} else
	if (!strcmp(name,"wireframe")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->wireframeColor());
			//return PyLong_FromLong(self->geometry->wireframeColor);
		else
			self->geometry->wireframeColor(value);
			//self->geometry->wireframeColor = value;
	} else
	if (!strcmp(name,"zone")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->zoneColor & FLAG_COLORMASK);
		else
			self->geometry->zoneColor = value;
	} else
	if (!strcmp(name,"zonebbox")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->zoneBBoxColor());
			//return PyLong_FromLong(self->geometry->zoneBBoxColor);
		else
			self->geometry->zoneBBoxColor(value);
			//self->geometry->zoneBBoxColor = value;
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Geometry_color

/* --- Geometry_derive --- */
static PyObject* Geometry_derive(GeometryObject* self, PyObject* Py_UNUSED(ignored))
{
	self->geometry->derive();
	Py_RETURN_NONE;
} // Geometry_derive

/* --- Geometry_error --- */
static PyObject* Geometry_error(GeometryObject* self, PyObject* args)
{
	char* cmd=nullptr;
	int n = 0;
	if (!PyArg_ParseTuple(args, "|si", &cmd, &n)) return NULL;

	if (cmd==nullptr or !strcmp(cmd, "n"))		// return number of errors
		return PyLong_FromLong(self->errors.size());
	else
	if (!strcmp(cmd, "clear"))	// clear error points
		self->errors.clear();
	else
	if (!strcmp(cmd, "get"))
		return PyUnicode_FromString(self->errors[n].c_str());
	else {
		PyErr_Format(PyExc_TypeError, "Invalid type '%s' specified", cmd);
		return NULL;
	}
	Py_RETURN_NONE;
} // Geometry_error

/* --- Geometry_fpe --- */
static PyObject* Geometry_fpe(ViewerObject*, PyObject* args)
{
	int flag = true;
	if (!PyArg_ParseTuple(args, "|i", &flag)) return NULL;
	fpetrap(flag);
	Py_RETURN_NONE;
} // Geometry_fpe

/* --- Geometry_materialVar --- */
static PyObject* Geometry_materialVar(GeometryObject* /*self*/,
				Material* material,
				const char* var,
				PyObject* value)
{
	if (!strcmp(var, "name")) {
		if (value==NULL)
			return PyUnicode_FromString(material->cname());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set material name");
			return NULL;
		}
	} else
	if (!strcmp(var, "id")) {
		if (value==NULL)
			return PyLong_FromLong(material->id());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set material id");
			return NULL;
		}
	} else
	if (!strcmp(var, "density")) {
		if (value==NULL)
			return PyFloat_FromDouble(material->density());
		else
			material->density(PyFloat_AsDouble(value));
	} else
	if (!strcmp(var, "Z")) {
		if (value==NULL)
			return PyLong_FromLong(material->Z());
		else
			material->Z(PyLong_AsLong(value));
	} else
	if (!strcmp(var, "A")) {
		if (value==NULL)
			return PyLong_FromLong(material->A());
		else
			material->A(PyLong_AsLong(value));
	} else
	if (!strcmp(var, "weight")) {
		if (value==NULL)
			return PyFloat_FromDouble(material->weight());
		else
			material->weight(PyFloat_AsDouble(value));
	} else
	if (!strcmp(var, "bxdf")) {
		if (value==NULL) {
			return PyUnicode_FromString(material->bxdfType().c_str());
		} else {
			if (PyUnicode_Check(value)) {
				if (!material->bxdf(PyUnicode_AsUTF8(value))) {
					PyErr_Format(PyExc_ValueError,
						"Invalid bxdf \'%s\' value specified",
							PyUnicode_AsUTF8(value));
					return NULL;
				}
			} else
			if (PyLong_Check(value)) {
				if (!material->bxdf(PyLong_AsLong(value))) {
					PyErr_Format(PyExc_ValueError,
						"Invalid bxdf \'%d\' value specified",
						PyLong_AsLong(value));
					return NULL;
				}
			} else {
				PyErr_SetString(PyExc_TypeError, "Invalid bxdf specified");
				return NULL;
			}
		}
	} else
	if (!strcmp(var, "color")) {
		if (value==NULL)
			return PyLong_FromLong((int)material->color());
		else
			material->diffuse(PyLong_AsLong(value));
	} else
	if (!strcmp(var, "diffuse")) {
		if (value==NULL)
			return PyLong_FromLong((int)material->diffuse());
		else
			material->diffuse(PyLong_AsLong(value));
	} else
	if (!strcmp(var, "specular")) {
		if (value==NULL)
			return PyFloat_FromDouble(material->specular());
		else
			material->specular(PyFloat_AsDouble(value));
	} else
	//if (!strcmp(var, "alpha")) {		// TODO: to be checked if the alpha variable should be used!
	//	if (value==NULL)
	//		// return PyLong_FromLong((int)material->transparent());
	//		return PyFloat_FromDouble(material->alpha());
	//	else
	//		// material->transparent(PyLong_AsLong(value));
	//		material->alpha(PyFloat_AsDouble(value));
	//} else
	if (!strcmp(var, "shine")) {
		if (value==NULL)
			return PyFloat_FromDouble(material->shine());
		else
			material->shine(PyFloat_AsDouble(value));
	} else
	//if (!strcmp(var, "shineexp")) {
	//	if (value==NULL)
	//		return PyFloat_FromDouble(material->shineExp());
	//	else
	//		material->shineExp(PyFloat_AsDouble(value));
	//} else
	if (!strcmp(var, "ior")) {
		if (value==NULL)
			return PyFloat_FromDouble(material->ior());
		else
			material->ior(PyFloat_AsDouble(value));
	} else
	if (!strcmp(var, "fuzz")) {
		if (value==NULL)
			return PyFloat_FromDouble(material->fuzz());
		else
			material->fuzz(PyFloat_AsDouble(value));
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", var);
		return NULL;
	}
	Py_RETURN_NONE;
} // Geometry_materialVar

/* --- Geometry_material --- */
static PyObject* Geometry_material(GeometryObject* self, PyObject* args)
{
	PyObject* obj;
	char* var;
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "Os|O", &obj, &var, &value)) return NULL;

	if (!strcmp(var,"add")) {
		if (!PyUnicode_Check(obj)) {
			PyErr_SetString(PyExc_TypeError,"String expected as material name");
			return NULL;
		}
		const char* name = PyUnicode_AsUTF8(obj);
		Material* material = self->geometry->addMaterial(name);
		return PyLong_FromLong(material->id());
	} else
	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (auto material : self->geometry->materials) {
			if (!fnmatch(pattern, material->cname(), 0)) {
				PyObject* ret = Geometry_materialVar(self, material, var, value);
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else  {
		Material* material = Py_Material(self, obj);
		if (material==NULL) return NULL;
		return Geometry_materialVar(self, material, var, value);
	}
	Py_RETURN_NONE;		// To keep compiler happy
} // Geometry_material

/* --- Geometry_memory --- */
static PyObject* Geometry_memory(GeometryObject* self, PyObject* args)
{
	char* dump=NULL;
	if (!PyArg_ParseTuple(args, "|s", &dump)) return NULL;

#ifdef MEM
	cout << "resident="  << Memory::resident() << endl;
	cout << "items="     << Memory::items() << endl;
	cout << "allocated=" << Memory::allocated() << endl;
#endif

	if (dump==NULL)
		return PyLong_FromLong(self->geometry->memory());
	else
		self->geometry->printMemory();

	Py_RETURN_NONE;
} // Geometry_memory

/* --- Geometry_message --- */
static PyObject* Geometry_message(GeometryObject* self, PyObject* args)
{
	char* msg=NULL;
	int  color=-1;
	if (!PyArg_ParseTuple(args, "|si", &msg, &color)) return NULL;

	if (msg==NULL)
		return PyUnicode_FromString(self->geometry->message().c_str());
	else
		self->geometry->message(msg);

	if (color>=0)
		self->geometry->messageColor = color & FLAG_COLORMASK;
	Py_RETURN_NONE;
} // Geometry_message

/* --- Geometry_mode --- */
static PyObject* Geometry_mode(GeometryObject* self, PyObject* args)
{
	const char* value = nullptr;
	if (!PyArg_ParseTuple(args, "|s", &value)) return NULL;
	if (value == nullptr) {
		switch (self->geometry->mode()) {
			case Geometry::Mode::Fluka:
				return PyUnicode_FromString("fluka");
			case Geometry::Mode::Moira:
				return PyUnicode_FromString("moira");
			case Geometry::Mode::Mcnp:
				return PyUnicode_FromString("mcnp");
			case Geometry::Mode::Phits:
				return PyUnicode_FromString("phits");
			case Geometry::Mode::Penelope:
				return PyUnicode_FromString("penelope");
		}
	} else {
		if (!strcmp(value,"fluka"))
			self->geometry->mode(Geometry::Mode::Fluka);
		else
		if (!strcmp(value,"moira"))
			self->geometry->mode(Geometry::Mode::Moira);
		else
		if (!strcmp(value,"mcnp"))
			self->geometry->mode(Geometry::Mode::Mcnp);
		else
		if (!strcmp(value,"phits"))
			self->geometry->mode(Geometry::Mode::Phits);
		else
		if (!strcmp(value,"penelope"))
			self->geometry->mode(Geometry::Mode::Penelope);
		else {
			PyErr_Format(PyExc_ValueError,"invalid mode '%s'", value);
			return NULL;
		}
	}
	Py_RETURN_NONE;
} // Geometry_mode

/* --- Geometry_object --- */
static PyObject* Geometry_object(GeometryObject* self, PyObject* args)
{
	PyObject* obj;
	char const* var = (char const*)"id";
	PyObject*   value=NULL;
	if (!PyArg_ParseTuple(args, "O|sO", &obj, &var, &value)) return NULL;

	if (!strcmp(var,"add")) {
		if (!PyUnicode_Check(obj)) {
			PyErr_SetString(PyExc_TypeError,"String expected as object name");
			return NULL;
		}
		if (!PyUnicode_Check(value)) {
			PyErr_SetString(PyExc_TypeError,"String expected as object type");
			return NULL;
		}
		const char* name = PyUnicode_AsUTF8(obj);
		const char* type = PyUnicode_AsUTF8(value);
		GObject* object = NULL;
		if (!strcmp(type,"arrow"))
			object = new GArrow(name);
		else
		if (!strcmp(type,"beam"))
			object = new GBeam(name);
		else
		if (!strcmp(type,"camera"))
			object = new GCamera(name);
		else
		if (!strcmp(type,"light"))
			object = new GLight(name);
		else
		if (!strcmp(type,"mesh"))
			object = new GMesh(name);
		else
		if (!strcmp(type,"point"))
			object = new GPoint(name);
		else
		if (!strcmp(type,"rotdefi"))
			object = new GRotdefi(name);
		else
		if (!strcmp(type,"ruler"))
			object = new GRuler(name);
		else
		if (!strcmp(type,"spline"))
			object = new GSpline(name);
		else {
			PyErr_Format(PyExc_TypeError, "Invalid object '%s'", name);
			return NULL;
		}

		if (object) {
			object->id(self->objects.size());
			self->objects.push_back(object);
			return PyLong_FromLong(object->id());
		}
	} else
	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (auto object : self->objects) {
			if (!fnmatch(pattern, object->cname(), 0)) {
				PyObject* ret = object->config(var, value);
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else
	if (PyList_Check(obj)) {
		for (ssize_t i=0; i<PyList_GET_SIZE(obj); i++) {
			GObject* object = Py_Object(self, PyList_GetItem(obj,i));
			if (object==NULL) return NULL;
			PyObject* ret = object->config(var, value);
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
		Py_RETURN_NONE;
	} else
	if (PyTuple_Check(obj)) {
		for (ssize_t i=0; i<PyTuple_GET_SIZE(obj); i++) {
			GObject* object = Py_Object(self, PyTuple_GetItem(obj,i));
			if (object==NULL) return NULL;
			PyObject* ret = object->config(var, value);
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
		Py_RETURN_NONE;
	} else {
		GObject* object = Py_Object(self, obj);
		if (object)
			return object->config(var, value);
	}
	return NULL;
} // Geometry_object

/* --- _region_addZone ---
 * add a zone to existing region
 * Zone should be in the format (["RPN"]|"STD"], [zonelist...])
 * @return true on failure, false otherwise
 */
static bool _region_addZone(GeometryObject* self, GRegion* region,
			PyObject* zoneobj)
{
	if (!PyTuple_Check(zoneobj)) {
		PyErr_SetString(PyExc_TypeError, "Invalid region expression, tuple expected");
		return true;
	}

	PyObject* efmt  = PyTuple_GetItem(zoneobj, 0);
	PyObject* zexpr = PyTuple_GetItem(zoneobj, 1);

	if (!PyUnicode_Check(efmt) || !PyList_Check(zexpr)) {
		PyErr_SetString(PyExc_TypeError,
			"Invalid region expression, tuple (expr-format,zone-expression) expected");
		return true;
	}

	const char* sfmt = PyUnicode_AsUTF8(efmt);
	GZone::Format format;
	if (!strcmp(sfmt, "STD") || !strcmp(sfmt,"PRODUCT"))
		format = GZone::Format::PRODUCT;
	else
	if (!strcmp(sfmt, "RPN"))
		format = GZone::Format::RPN;
	else
	if (!strcmp(sfmt, "NORMAL"))
		format = GZone::Format::NORMAL;
	else {
		PyErr_Format(PyExc_TypeError, "Invalid expression format %s", sfmt);
		return true;
	}

	GZone* zone = self->geometry->addZone(region, format);
	zone->reserve(PyList_GET_SIZE(zexpr));
	for (ssize_t j=0; j<PyList_GET_SIZE(zexpr); j++) {
		PyObject* token = PyList_GetItem(zexpr, j);
		if (!PyUnicode_Check(token))
			PyErr_SetString(PyExc_TypeError,
				"Invalid region expression, string expected");
		else
		if (!self->geometry->add2exp(region, PyUnicode_AsUTF8(token)))
			PyErr_Format(PyExc_SyntaxError,
				"Region '%s' invalid token: '%s'",
				region->cname(), PyUnicode_AsUTF8(token));
	}
	if (PyErr_Occurred()) return true;
	if (not zone->verify()) {
		PyErr_SetString(PyExc_TypeError, "Invalid region expression");
		return true;
	}
	return false;
} // _region_addZone

/** add a region expression as string
 * @param region	region
 * @param expr		expression string
 * @param retzones	return zones
 * @return subzones if retzones exists and is true
 *
 * FIXME: merge with the Geometry.parseZone
 */
static PyObject* _region_addExpr(GeometryObject* self, GRegion* region,
				PyObject* expr, PyObject* retzones)
{
	PyObject* zones = nullptr;
	if (retzones and PyObject_IsTrue(retzones))
		zones = PyList_New(0);

	std::stringstream exprstream(PyUnicode_AsUTF8(expr));
	Token token(exprstream);
	token.setAlpha(".$@",Token::AlphaNumeric::SUFFIX);	// set additional identifier symbols

	GZone* zone = nullptr;
	GZone::Format fmt = GZone::Format::PRODUCT;		// initial guess
	int startpos  = 0;
	int endpos    = 0;
	bool err      = false;
	bool hash     = false;
	int depth     = 0;
	int zonecount = 0;
	while (token.next() != Token::Type::ENDOFILE) {
		if (zone==nullptr) {
			if (err and zone) zone->clear();
			zone = region->addZone(GZone::Format::NORMAL);
			fmt  = GZone::Format::PRODUCT;
			startpos = endpos;	// just before the token
			zonecount++;
			err = false;
		}
		switch (token()) {
			case Token::Type::IDENT: {
				GBody* body = nullptr;
				if (hash) {		// # region present
					body = self->geometry->addRegionBody(token.string());
					fmt = GZone::Format::RPN;
				} else {
					body = self->geometry->getBody(token.string());
					if (body==nullptr) {
						self->errors.push_back(
							sformat("Region '%s:%d' invalid body '%s'",
								region->cname(), zonecount,
								token.string()));
						err = true;
					}
				}
				if (body) *zone << body;
				hash = false;
				}
				break;

			case Token::Type::NUMBER: {
				GBody* body = nullptr;
				assert(In(self->geometry->mode(),
						Geometry::Mode::Mcnp,
						Geometry::Mode::Phits));
				if (hash) {		// # region present
					if (token.isInteger()) {
						int n = token.integer();
						if (n<0) {
							self->errors.push_back(
								sformat("Cell '%s:%d' invalid cell '#%s'",
									region->cname(), zonecount,
									token.string()));
							err = true;
						}
						*zone << "-";
					}
					body = self->geometry->addRegionBody(token.string());
					fmt = GZone::Format::RPN;
				} else {		// normal surface
					if (token.isInteger()) {
						int n = token.integer();
						// FIXME do we need to check for MCNP?
						*zone << (n>0? "-" : "+");	// for MCNP inverted!!!
						body = self->geometry->getBody(std::to_string(abs(n)));
					}
					if (body==nullptr) {
						self->errors.push_back(
							sformat("Cell '%s:%d' invalid surface '%s'",
								region->cname(), zonecount,
								token.string()));
						err = true;
					}
				}
				if (body) *zone << body;
				hash = false;
				}
				break;

			case Token::Type::OR:
			case Token::Type::COLON:
				if (depth==0) {		// split zones
					if (zone->size()!=0) {
						if (zones) {
							// FIXME it might create misalignment
							// with the card[Card.ZONE]
							std::string subexpr = exprstream.str().
								substr(startpos,endpos-startpos);
							trim(subexpr);
							PyList_Append(zones,
								PyUnicode_FromString(subexpr.c_str()));
						}
						// Last check for format
						// (+-) +-term => RPN
						// (+-) => PRODUCT
						if (fmt == GZone::Format::PRODUCT and
						    (*zone)[0] == &GBody::tleft   and
						    (*zone)[zone->size()-1] != &GBody::tright)
							fmt = GZone::Format::RPN;
						try {
							zone->convertTo(fmt);
						} catch (std::string& ex) {
							self->errors.push_back(
								sformat("Region '%s:%d' %s",
									region->cname(), zonecount,
									ex.c_str()));
							err = true;
						}
						zone = nullptr;
					}
					fmt = GZone::Format::PRODUCT;
				} else {
					*zone << &GBody::tunion;
					fmt = GZone::Format::RPN;
				}
				break;

			case Token::Type::LEFT_PAREN:
				if (In(self->geometry->mode(),
						Geometry::Mode::Mcnp,
						Geometry::Mode::Phits) and zone->size()>0)
					*zone << (hash? "-" : "+");
				*zone << &GBody::tleft;
				hash = false;
				// set type to RPN if ( is not the first token
				fmt = GZone::Format::RPN;
				depth++;
				break;

			case Token::Type::RIGHT_PAREN:
				*zone << &GBody::tright;
				if (--depth<0) {
					self->errors.push_back(
						sformat("Region '%s:%d' too many closing ) parenthesis",
							region->cname(), zonecount));
					depth = 0;
					err = true;
				}
				break;

			case Token::Type::PLUS:
				if (In(self->geometry->mode(),
				    Geometry::Mode::Fluka, Geometry::Mode::Moira))
					*zone << &GBody::tplus;
				break;

			case Token::Type::MINUS:
				if (In(self->geometry->mode(),
				    Geometry::Mode::Fluka, Geometry::Mode::Moira))
					*zone << &GBody::tminus;
				break;

			case Token::Type::HASH:
				//assert(In(self->geometry->mode(),
				//		Geometry::Mode::Mcnp,
				//		Geometry::Mode::Phits));
				if (hash)
					self->errors.push_back("Double # operator found");
				hash = true;
				break;

			case Token::Type::NEWLINE:	// skip
				break;

			default:
				self->errors.push_back(
					sformat("Region '%s:%d' invalid token '%s'",
						region->cname(), zonecount,
						token.string()));
				err = true;
		}
		endpos = exprstream.tellg();
	}

	if (zones and zone!=nullptr and zone->size()) {
		// FIXME it might create misalignment with the card[Card.ZONE]
		std::string subexpr = exprstream.str().substr(startpos,endpos-startpos);
		trim(subexpr);
		PyList_Append(zones, PyUnicode_FromString(subexpr.c_str()));
	}

	if (err) {
		if (zone) zone->clear();
	} else
	if (depth!=0) {
		self->errors.push_back(
			sformat("Region '%s:%d' too many opening '(' parenthesis",
				region->cname(), zonecount));
		if (zone) zone->clear();
	} else
	if (zone==nullptr) {
		if (region->type() != GRegion::Type::VOXEL)
			self->errors.push_back(
				sformat("Region '%s' empty expression",
					region->cname()));
	} else
	if (zone->size()==0)
		region->delZone(region->size()-1);
	else {
		// Last check for format
		// (+-) +-term => RPN
		// (+-) => PRODUCT
		if (fmt == GZone::Format::PRODUCT and
		    (*zone)[0] == &GBody::tleft   and
		    (*zone)[zone->size()-1] != &GBody::tright)
			fmt = GZone::Format::RPN;
		try {
			zone->convertTo(fmt);
		} catch (std::string& ex) {
			zone->clear();
			self->errors.push_back(
				sformat("Region '%s:%d' %s",
					region->cname(), zonecount,
					ex.c_str()));
		}
	}
	if (zones) return zones;
	Py_RETURN_NONE;
} // _region_addExpr

/* --- Geometry_regionVar --- */
static PyObject* Geometry_regionVar(GeometryObject* self, GRegion* region,
			const char* var, PyObject* value, PyObject* value2)
{
	if (!strcmp(var, "name")) {
		if (value==NULL)
			return PyUnicode_FromString(region->cname());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set region name");
			return NULL;
		}
	} else
	if (!strcmp(var, "id")) {
		if (value==NULL)
			return PyLong_FromLong(region->id());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set region id");
			return NULL;
		}
	} else
	if (!strcmp(var, "expr")) {
		if (value==NULL)
			return Py_RegionExpr(region, self->geometry->mode());
		else
		if (PyList_Check(value)) {
			if (PyList_GET_SIZE(value)>0)
				region->generation(self->geometry->nextGeneration());
			for (ssize_t i=0; i<PyList_GET_SIZE(value); i++)
				if (_region_addZone(self, region, PyList_GetItem(value, i)))
					return NULL;
		} else
		if (PyUnicode_Check(value))
			return _region_addExpr(self, region, value, value2);
		else {
			PyErr_SetString(PyExc_KeyError,
				"Invalid region expression, string or list of tuples expected");
			return NULL;
		}
	} else
	if (!strcmp(var, "zone")) {	// get (ONLY) a specific zone or number of zones
		if (value==NULL)
			return PyLong_FromLong(region->size());
		else
		if (PyLong_Check(value)) {
			int i = PyLong_AsLong(value);
			if (0<=i and i<region->size())
				return Py_ZoneExpr((*region)[i], self->geometry->mode());
		}

		PyErr_SetString(PyExc_TypeError, "Invalid zone specified");
		return NULL;
	} else
	if (!strcmp(var, "material")) {
		if (value==NULL) {
			const Material* mat = region->material();
			if (mat)
				return PyUnicode_FromString(mat->cname());
			else
				Py_RETURN_NONE;
		} else
			region->material(Py_Material(self, value));
	} else
	if (!strcmp(var, "matrix")) {
		if (value) {
			if (PyList_Check(value)) {
				Matrix4 matrix;
				PyList_AsMatrix4(value, matrix);
				try {
					region->matrix(matrix);
				} catch (...) {
					PyErr_SetString(PyExc_SyntaxError,
						"Invalid region rotation matrix");
					return NULL;
				}
			} else
				region->clearMatrix();
		} //else return matrix FIXME
	} else
	if (!strcmp(var, "hasmatrix")) {
		return PyBool_FromLong(region->hasMatrix());
	} else
	if (!strcmp(var, "show")) {
		if (value==NULL)
			return PyLong_FromLong(region->show);
		else
		if (PyLong_Check(value))
			region->show = PyLong_AsLong(value);
		else {
			PyErr_SetString(PyExc_TypeError, "Invalid region 'show' value specified");
			return NULL;
		}
	} else
	if (!strcmp(var, "type")) {
		if (value==NULL)
			return PyLong_FromLong((int)region->type());
		else
		if (PyLong_Check(value))
			region->type((GRegion::Type)PyLong_AsLong(value));
		else
		if (value != Py_None) {
			PyErr_SetString(PyExc_TypeError, "Invalid region 'type' value specified");
			return NULL;
		}
	} else
	if (!strcmp(var, "optimize"))
		return PyBool_FromLong(region->optimize());
	else
	if (!strcmp(var, "expand")) {
		if (value==NULL) {
			assert(false);
		} else
		if (PyLong_Check(value)) {
			int i = PyLong_AsLong(value);
			if (i<0 or i>=region->size()) Py_RETURN_FALSE;

			GZone* zone = (*region)[i];
			if (zone->std())
				return PyBool_FromLong(zone->optimize());
			else
			if (zone->rpn()) {
				unsigned limit = 0;
				if (value2 and PyLong_Check(value2))
					limit = PyLong_AsUnsignedLong(value2);
				return PyBool_FromLong(zone->expand(limit));
			} else {
				PyErr_SetString(PyExc_TypeError, "Invalid zone 'format'");
				return NULL;
			}
		}
		PyErr_SetString(PyExc_TypeError, "expand: invalid zone specified");
		return NULL;
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", var);
		return NULL;
	}
	Py_RETURN_NONE;
} // Geometry_regionVar

/* --- Geometry_region --- */
static PyObject* Geometry_region(GeometryObject* self, PyObject* args)
{
	PyObject* obj;
	char* var;
	PyObject* value=NULL;
	PyObject* value2=NULL;
	if (!PyArg_ParseTuple(args, "Os|OO", &obj, &var, &value, &value2)) return NULL;

	if (!strcmp(var,"add")) {
		if (!PyUnicode_Check(obj)) {
			PyErr_SetString(PyExc_TypeError,"String expected as region name");
			return NULL;
		}
		const char* name = PyUnicode_AsUTF8(obj);
		GRegion* region = self->geometry->addRegion(name);
		return PyLong_FromLong(region->id());
	} else
	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (auto region : self->geometry->regions) {
			if (!fnmatch(pattern, region->cname(), 0)) {
				PyObject* ret = Geometry_regionVar(self, region, var, value, value2);
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else  {
		GRegion* region = Py_GRegion(self, obj);
		if (region==NULL) return NULL;
		return Geometry_regionVar(self, region, var, value, value2);
	}
	Py_RETURN_NONE;		// To keep compiler happy
} // Geometry_region

/* --- Geometry_set --- */
static PyObject* Geometry_set(GeometryObject* self, PyObject* args)
{
#define NOTDEFINED	-999999999
	char* name;
	int   value=NOTDEFINED;

	if (!PyArg_ParseTuple(args, "s|i", &name, &value)) return NULL;

	if (!strcmp(name,"axislen")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->axisLen);
		else
			self->geometry->axisLen = value;
	} else
	if (!strcmp(name,"cursor")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->cursorSize);
		else {
			self->cursorSize = value;
			if (self->cursorSize < 2*self->cursorMoveSize)
				self->cursorSize = 2*self->cursorMoveSize;
		}
	} else
	if (!strcmp(name,"developer")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->developer);
		else
			self->geometry->developer = value;
	} else
	if (!strcmp(name,"lighterlevel")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(self->geometry->lighterLevel);
		else
			self->geometry->lighterLevel = value;
	} else
	if (!strcmp(name,"snapdistance")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(Round(self->snapDistance));
		else
			self->snapDistance = value;
	} else
	if (!strcmp(name,"snapangle")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(Round(Deg(self->snapAngle)));
		else
			self->snapAngle = Rad(value);
	} else
	if (!strcmp(name,"trackballsize")) {
		if (value==NOTDEFINED)
			return PyLong_FromLong(Round(self->trackballSize*100.0));
		else
			self->trackballSize = (double)value/100.0;
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Geometry_set

/* --- Geometry_select --- */
static PyObject* Geometry_select(GeometryObject* self, PyObject* args)
{
	char* name=NULL;
	dword mask=BIT_SELECT | BIT_FREEZE;

	if (!PyArg_ParseTuple(args, "s|i", &name, &mask)) return NULL;

	if (name[0]==0) Py_RETURN_NONE;

	if (name[0]=='b' || name[0]=='B') {
		int count = 0;
		for (auto body : self->geometry->bodies) {
			if (body->show & mask)
				count++;
		}
		return PyLong_FromLong(count);
	} else
	if (name[0]=='r' || name[0]=='R') {
		int count = 0;
		for (auto region : self->geometry->regions) {
			if (region->show & mask)
				count++;
		}
		return PyLong_FromLong(count);
	} else
	if (name[0]=='o' || name[0]=='O') {
		int count = 0;
		for (auto object : self->objects)
			if (object->select)
				count++;
		return PyLong_FromLong(count);
	} else
	Py_RETURN_NONE;
} // Geometry_select

/* --- Geometry_setLights --- */
static PyObject* Geometry_setLights(GeometryObject* self, PyObject* Py_UNUSED(ignored))
{
	// Scan objects for the first MAXLIGHT active lights
	self->geometry->delLights();
	for (auto object : self->objects) {
		Light light;
		if (object->isA() != GLightClass) continue;
		static_cast<GLight*>(object)->toLight(light);
		self->geometry->addLight(light);
	}
	Py_RETURN_NONE;
} // Geometry_setLights

/* --- Geometry_volume --- */
static PyObject* Geometry_volume(GeometryObject* self, PyObject* args)
{
	// FIXME not static move to self
	PyObject* obj = NULL;
	int samples = -1;
	char* cmd = NULL;

	if (!PyArg_ParseTuple(args, "|sOi", &cmd, &obj, &samples)) return NULL;

	if (cmd==NULL || !strcmp(cmd, "status"))
		goto STATUS;
	else
	if (!strcmp(cmd, "stop")) {
		// FIXME Needs to be protected by semaphore
		if (self->volumeJob.isRunning()) {
			self->volumeJob.stop();
			self->debug3DJob.stop();		// CT: debug test
		}
	} else
	if (!strcmp(cmd, "run")) {
		if (obj==NULL) {
			PyErr_SetString(PyExc_ValueError,"Region expected");
			return NULL;
		}
		if (self->volumeJob.isRunning() || self->debug3DJob.isRunning()) {	// CT: debug test
			PyErr_SetString(PyExc_ValueError,
				"Cannot submit a volume job while the previous hasn't finished");
			return NULL;
		} else {
			if (self->volumeJob.init(self->geometry,
						Py_GRegion(self, obj),
						samples))
				self->volumeJob.spawn();
			/* CT: Enable to test 3D debugging
				if( self->debug3DJob.init( self->geometry,
						Py_GRegion( self, obj )->bbox(),
						samples ) )
					self->debug3DJob.spawn();
			*/
			goto STATUS;
		}
	}
	Py_RETURN_NONE;		// To keep compiler happy

STATUS:
	return Py_BuildValue("sdddi",
				(self->volumeJob.isRunning()? "running" : "finished"),
				self->volumeJob.volume,
				self->volumeJob.area,
				self->volumeJob.error,
				self->volumeJob.nsamples);

	/*
	// CT
	////////// TEST of 3D debugging
	//
	// FIXME not static move to self
	PyObject* obj = NULL;
	int samples = -1;
	char* cmd = NULL;

	if (!PyArg_ParseTuple(args, "|sOi", &cmd, &obj, &samples)) return NULL;

	if (cmd==NULL || !strcmp(cmd, "status"))
		goto STATUS;
	else
	if (!strcmp(cmd, "stop")) {
		// FIXME Needs to be protected by semaphore
		if (self->debug3DJob.isRunning())
			self->debug3DJob.stop();
	} else
	if (!strcmp(cmd, "run") && cmd != nullptr) {
		if (obj==NULL) {
			PyErr_SetString(PyExc_ValueError,"Region expected");
			return NULL;
		}
		if (self->debug3DJob.isRunning()) {
			PyErr_SetString(PyExc_ValueError,
				"Cannot submit a debug job while the previous hasn't finished");
			return NULL;
		} else {
			if (self->debug3DJob.init(self->geometry,
						Py_GRegion(self, obj)->bbox(),
						samples))
				self->debug3DJob.spawn();
			goto STATUS;
		}
	}
	Py_RETURN_NONE;		// To keep compiler happy

STATUS:
	return Py_BuildValue("siii",
				(self->debug3DJob.isRunning()? "running" : "finished"),
				(int)self->debug3DJob.ambiguousPoints.size(),
				(int)self->debug3DJob.undefinedPoints.size(),
				self->debug3DJob.nsamples);
*/
} // Geometry_volume

/* --- Geometry_voxel --- */
static PyObject* Geometry_voxel(GeometryObject* self, PyObject* args)
{
	char* name;
	PyObject* value=NULL, *opt=NULL;

	if (!PyArg_ParseTuple(args, "s|OO", &name, &value, &opt)) return NULL;

	if (!strcmp(name,"load")) {	// load voxel detector
		if (value != NULL) {
			const char* filename = PyUnicode_AsUTF8(value);
			if (!self->geometry->voxel.load(filename)) {
				PyErr_Format(PyExc_IOError, "Unable to read voxel file \'%s\'",
						filename);
				return NULL;
			}
		}
	} else
	if (!strcmp(name,"xlow")) {	// set x-low position
		if (value != NULL)
			self->geometry->voxel.xlow = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->geometry->voxel.xlow);
	} else
	if (!strcmp(name,"ylow")) {	// set y-low position
		if (value != NULL)
			self->geometry->voxel.ylow = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->geometry->voxel.ylow);
	} else
	if (!strcmp(name,"zlow")) {	// set z-low position
		if (value != NULL)
			self->geometry->voxel.zlow = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->geometry->voxel.zlow);
	} else
	if (!strcmp(name,"xhigh")) {	// set x-high position
		if (value != NULL)
			self->geometry->voxel.xhigh = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->geometry->voxel.xhigh);
	} else
	if (!strcmp(name,"yhigh")) {	// set y-high position
		if (value != NULL)
			self->geometry->voxel.yhigh = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->geometry->voxel.yhigh);
	} else
	if (!strcmp(name,"zhigh")) {	// set z-high position
		if (value != NULL)
			self->geometry->voxel.zhigh = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->geometry->voxel.zhigh);
	} else
	if (!strcmp(name,"matrix")) {	// set transformation matrix for voxel
		if (value) {
			if (PyList_Check(value)) {
				Matrix4	matrix;
				PyList_AsMatrix4(value, matrix);
				self->geometry->voxel.matrix(matrix);
			} else
				self->geometry->voxel.clearMatrix();
		} //else return matrix
	} else
	if (!strcmp(name,"get")) {	// get kreg[data[x,y,z]] value (region number)
		if (value != NULL) {
			if (!Py_IsVector(value)) {
				PyErr_SetString(PyExc_TypeError,"Tuple with voxel coordinates was expected");
				return NULL;
			}
			Vector p = Py_GetVector(value);
			return PyLong_FromLong(self->geometry->voxel.get(p.x, p.y, p.z));
		}
		// no setting is allowed
	} else
	if (!strcmp(name,"data")) {	// get data[x,y,z] value (HU number) WARNING!
		if (value != NULL) {
			if (!Py_IsVector(value)) {
				PyErr_SetString(PyExc_TypeError,"Tuple with voxel coordinates was expected");
				return NULL;
			}
			Vector p = Py_GetVector(value);
			return PyLong_FromLong(self->geometry->voxel.data(p.x, p.y, p.z));
		}
		// no setting is allowed
	} else
	if (!strcmp(name,"index")) {	// return (i,j,k) position of (x,y,z)
		if (value != NULL) {
			if (!Py_IsVector(value)) {
				PyErr_SetString(PyExc_TypeError,"Tuple with voxel coordinates was expected");
				return NULL;
			}
			Vector p = Py_GetVector(value);
			return Py_BuildValue("iii",
					self->geometry->voxel.voxeli(p.x),
					self->geometry->voxel.voxelj(p.y),
					self->geometry->voxel.voxelk(p.z));
		}
		// no setting is allowed
	} else
	if (!strcmp(name,"nregions") || !strcmp(name,"no")) {	// return no (number of regions)
		return PyLong_FromLong(self->geometry->voxel.nregions);
	} else
	if (!strcmp(name,"nvoxels") || !strcmp(name,"mo")) {	// return mo (number of unique voxels)
		return PyLong_FromLong(self->geometry->voxel.nvoxels);
	} else
	if (!strcmp(name,"free")) {
		self->geometry->voxel.cleanup();
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid option", name);
		return NULL;
	}

	self->geometry->voxel.calcLimits();

	Py_RETURN_NONE;
} // Geometry_voxel

/** _selectZone
 * @return false on error
 */
static void _selectZone(GeometryObject* self, PyObject* obj)
{
	if (PyUnicode_Check(obj)) {
		GZone* zone = self->geometry->editRegion.addZone(GZone::Format::NORMAL);
		try {
			self->geometry->parseZone(*zone, PyUnicode_AsUTF8(obj));
		} catch (std::string& err) {
			PyErr_SetString(PyExc_Warning, err.c_str());
		}
	} else
	if (PyTuple_Check(obj) && PyTuple_GET_SIZE(obj)==2) {
		// Set zone
		GZone::Format format = GZone::Format::PRODUCT;
		PyObject* mode = PyTuple_GetItem(obj,0);
		if (PyUnicode_Check(mode)) {
			if (!strcmp(PyUnicode_AsUTF8(mode),"RPN"))
				format = GZone::Format::RPN;
		} else
		if (PyLong_Check(mode))
			format = (GZone::Format)PyLong_AsLong(mode);
		PyObject* zoneobj = PyTuple_GetItem(obj,1);

		GZone* zone = self->geometry->editRegion.addZone(format);
		zone->reserve(PyList_GET_SIZE(zoneobj));
		for (ssize_t j=0; j<PyList_GET_SIZE(zoneobj); j++) {
			const char* token = PyUnicode_AsUTF8(PyList_GetItem(zoneobj, j));
			if (PyErr_Occurred()) {
				PyErr_SetString(PyExc_ValueError,
					"Invalid region expression, string expected");
				return;
			}
			GBody* body = self->geometry->getBody(token);
			if (!zone->add(token, body)) {
				if (In(self->geometry->mode(),
						Geometry::Mode::Mcnp,
						Geometry::Mode::Phits))
					PyErr_Format(PyExc_ValueError,"invalid surface '%s'", token);
				else
					PyErr_Format(PyExc_ValueError,"invalid body '%s'", token);
				return;
			}
		}
		self->geometry->editRegion.setup();
	} else
		PyErr_SetString(PyExc_ValueError, "Invalid zone expression");
} // _selectZone

/** Manipulate selected zone */
static PyObject* Geometry_zone(GeometryObject* self, PyObject* args)
{
	char* cmd;
	PyObject* value=NULL;

	if (!PyArg_ParseTuple(args, "s|O", &cmd, &value)) return NULL;

	// check if it is already selected
	if (!strcmp(cmd, "has") && PyTuple_Check(value) && PyTuple_GET_SIZE(value)==2) {
		// construct the hash as in GZone::()
		dword _hash = 0;
		PyObject* zonelist = PyTuple_GetItem(value,1);	// 0 is the type
		for (ssize_t i=0; i<PyList_GET_SIZE(zonelist); i++) {
			PyObject* obj = PyList_GetItem(zonelist,i);
			_hash += (_hash<<5) + i;
			_hash += (_hash<<5) + hash_djb2(PyUnicode_AsUTF8(obj));
		}

		// check all zones in editRegion
		for (auto zone : self->geometry->editRegion.zones()) {
			if (zone->hash() == _hash)
				Py_RETURN_TRUE;
		}
		Py_RETURN_FALSE;
	} else
	if (!strcmp(cmd, "select")) {
		if (value==NULL)
			return Py_RegionExpr(&self->geometry->editRegion, self->geometry->mode());

		self->geometry->lockEdit();
		if (PyList_Check(value)) {
			for (ssize_t i=0; i<PyList_GET_SIZE(value); i++) {
				PyObject* obj = PyList_GetItem(value,i);
				_selectZone(self, obj);
				if (PyErr_Occurred()) {
					self->geometry->editRegion.clear();
					self->geometry->unlockEdit();
					return NULL;
				}
			}
		} else
		if (PyUnicode_Check(value) || PyTuple_Check(value)) {
			_selectZone(self, value);
			if (PyErr_Occurred()) {
				self->geometry->editRegion.clear();
				self->geometry->unlockEdit();
				return NULL;
			}
		}
		self->geometry->editRegion.expandRegions(self->geometry);
		self->geometry->editRegion.setup();
		self->geometry->unlockEdit();
	} else
	if (!strcmp(cmd, "contained")) {
		// check if editRegion zones exist inside a region
		GRegion* region = Py_GRegion(self, value);
		if (region == NULL) return NULL;
		self->geometry->lockEdit();
		for (auto zone : self->geometry->editRegion.zones())
			if (not region->contains(*zone)) {
				self->geometry->unlockEdit();
				Py_RETURN_FALSE;
			}
		self->geometry->unlockEdit();
		Py_RETURN_TRUE;
	} else
	if (!strcmp(cmd, "show")) {
		if (value==NULL)
			return PyLong_FromLong(self->geometry->editRegion.show);
		else
			self->geometry->editRegion.show = PyLong_AsLong(value);
	} else
	if (!strcmp(cmd, "clear")) {
		self->geometry->lockEdit();
		self->geometry->editRegion.clear();
		self->geometry->unlockEdit();
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", cmd);
		return NULL;
	}

	Py_RETURN_NONE;
} // Geometry_zone

/* --- Geometry_lock --- */
static PyObject* Geometry_lock(GeometryObject* self, PyObject* Py_UNUSED(ignored))
{
	self->geometry->lockWrite();
	Py_RETURN_NONE;
} // Geometry_lock

/* --- Geometry_unlock --- */
static PyObject* Geometry_unlock(GeometryObject* self, PyObject* Py_UNUSED(ignored))
{
	self->geometry->unlockWrite();
	Py_RETURN_NONE;
} // Geometry_unlock

/* ----------------- Geometry type ------------------- */
static PyMethodDef Geometry_methods[] = {
	{"bbox",	(PyCFunction)Geometry_bbox,	METH_VARARGS,	"Return bounding box"},
	{"body",	(PyCFunction)Geometry_body,	METH_VARARGS,	"Get/Set body properties"},
	{"cleanup",	(PyCFunction)Geometry_cleanup,	METH_NOARGS,	"cleanup everything"},
	{"cursor",	(PyCFunction)Geometry_cursor,	METH_VARARGS,	"Get/Set cursor object"},
	{"color",	(PyCFunction)Geometry_color,	METH_VARARGS,	"Return/set colors"},
	{"derive",	(PyCFunction)Geometry_derive,	METH_NOARGS,	"Derive"},
	{"error",	(PyCFunction)Geometry_error,	METH_VARARGS,	"Return error"},
	{"fpe",		(PyCFunction)Geometry_fpe,	METH_VARARGS,	"Activate FPE exceptions"},
	{"material",	(PyCFunction)Geometry_material,	METH_VARARGS,	"Set/Get material properties"},
	{"message",	(PyCFunction)Geometry_message,	METH_VARARGS,	"Set/Get message information"},
	{"memory",	(PyCFunction)Geometry_memory,	METH_VARARGS,	"Get or print memory"},
	{"mode",	(PyCFunction)Geometry_mode,	METH_VARARGS,	"Set/Get geometry mode"},
	{"object",	(PyCFunction)Geometry_object,	METH_VARARGS,	"Return/set object parameters"},
	{"region",	(PyCFunction)Geometry_region,	METH_VARARGS,	"Set/Get region properties"},
	{"set",		(PyCFunction)Geometry_set,	METH_VARARGS,	"Return/set flag/parameter"},
	{"selection",	(PyCFunction)Geometry_select,	METH_VARARGS,	"Get number of selected items"},
	{"setLights",	(PyCFunction)Geometry_setLights,METH_NOARGS,	"set lights from objects"},
	{"volume",	(PyCFunction)Geometry_volume,	METH_VARARGS,	"Calculate volume of region"},
	{"voxel",	(PyCFunction)Geometry_voxel,	METH_VARARGS,	"Set/Get voxel information"},
	{"zone",	(PyCFunction)Geometry_zone,	METH_VARARGS,	"Manipulate selected zone"},
	{"lock",	(PyCFunction)Geometry_lock,	METH_NOARGS,	"Lock mutex"},
	{"unlock",	(PyCFunction)Geometry_unlock,	METH_NOARGS,	"UnLock mutex"},
#if 0
//#ifdef MEM
	{"_destroy",	(PyCFunction)Geometry_destroy,	METH_NOARGS,	"Deallocate"},
#endif
	{NULL, NULL, 0, NULL}	// Sentinel
}; // Geometry_methods

#if 1
PyTypeObject GeometryType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	"geoviewer.Geometry",		// tp_name
	sizeof(GeometryObject),		// tp_basicsize
	0,				// tp_itemsize
	(destructor)Geometry_dealloc,	// tp_dealloc
	0,				// tp_print < 3.8 else tp_vectorcall_offset
	0,				// tp_getattr
	0,				// tp_setattr
	0,				// tp_reserved
	0,				// tp_repr
	0,				// tp_as_number
	0,				// tp_as_sequence
	0,				// tp_as_mapping
	0,				// tp_hash
	0,				// tp_call
	0,				// tp_str
	0,				// tp_getattro
	0,				// tp_setattro
	0,				// tp_as_buffer
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	// tp_flags
	"Geometry object",		// tp_doc
	0,				// tp_traverse
	0,				// tp_clear
	0,				// tp_richcompare
	0,				// tp_weaklistoffset
	0,				// tp_iter
	0,				// tp_iternext
	Geometry_methods,		// tp_methods
	0,//Geometry_members,		// tp_members
	0,				// tp_getset
	0,				// tp_base
	0,				// tp_dict
	0,				// tp_descr_get
	0,				// tp_descr_set
	0,				// tp_dictoffset
	0,				// tp_init
	0,				// tp_alloc
	Geometry_new,			// tp_new
	0,				// tp_free
	0,				// tp_is_gc
	0,				// tp_bases
	0,				// tp_mro
	0,				// tp_cache
	0,				// tp_subclasses
	0,				// tp_weaklist
	0,				// tp_del
	0,				// tp_version_tag
	0,				// tp_finalize
#if PY_VERSION_HEX >= 0x03080000
	0,				// tp_vectorcall
#if PY_VERSION_HEX < 0x03090000
	0,				// tp_print (deprecated)
#endif
#endif
}; // GeometryType
#else
PyTypeObject GeometryType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	.tp_name      = "geoviewer.Geometry",
	.tp_basicsize = sizeof(GeometryObject),
	.tp_itemsize  = 0,
	.tp_dealloc   = (destructor)Geometry_dealloc,
	.tp_flags     = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
	.tp_doc       = "Geometry object",
	.tp_methods   = Geometry_methods,
	.tp_new       = Geometry_new,
};
#endif
