/**
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	15-Jul-2020
 */

#ifndef __GRAPH_CLIPREGION_H
#define __GRAPH_CLIPREGION_H

#include <os.h>

// ================================ ClipRegion =================================
/* rectangular clipping region
 * use signed types to avoid problems when clipping negative points
 */
//template <typename T>
class ClipRegion {
public:
	/* out codes for clipping lines */
	union OutCodeUnion {		/* outcodes are represented as bit fields */
		struct {
			unsigned code0 : 1;	/* x < left	*/
			unsigned code1 : 1;	/* y < top	*/
			unsigned code2 : 1;	/* x > right	*/
			unsigned code3 : 1;	/* y > bottom	*/
		} ocs;
		int outcodes;
	};

public:
	int	left;			/* X-upper left corner	*/
	int	top;			/* Y-upper left corner	*/
	int	right;			/* X-lower right corner	*/
	int	bottom;			/* Y-lower right corner	*/

public:
	ClipRegion() = default;
	ClipRegion(int l, int t, int r, int b)
			: left(l),
			  top(t),
			  right(r),
			  bottom(b)
			{}
	ClipRegion(const ClipRegion& c) = default;
	ClipRegion& operator = (const ClipRegion& c) = default;
	~ClipRegion() = default;

	/** return with and height of clipping area */
	int	width() const	{ return right-left; }
	int	height() const	{ return bottom-top; }

	/** set clipping region to full window */
	void	set(int width, int height) {
			left   = 0;
			top    = 0;
			right  = width;
			bottom = height;
		}

	/** set clipping region.
	 * If values are negative then consider them from the other side
	 * @param l	X-left coordinate
	 * @param t	Y-top coordinate
	 * @param r	X-right coordinate
	 * @param b	Y-bottom coordinate
	 */
	void	set(int l, int t, int r, int b, int width, int height) {
			if (l<0) l = width  + l;
			if (t<0) t = height + t;
			if (r<0) r = width  + r;
			if (b<0) b = height + b;
			left   = Range(0, l, width );
			top    = Range(0, t, height);
			right  = Range(0, r, width );
			bottom = Range(0, b, height);
		}

	/** @return true if clipping region is not the full window */
	bool	isclipped(int width, int height) const {
			return	left   != 0     ||
				top    != 0     ||
				right  != width ||
				bottom != height;
		}

	/** @return true if point is inside clipping region */
	bool	inside(const int x, const int y) const {
			return left<=x && x<=right &&
			       top <=y && y<=bottom;
		}

	/** clamp point inside clipping region */
	void	point(int* x, int* y) const {
			*x = Range(left, *x, right);
			*y = Range(top,  *y, bottom);
		}

	/** clip line coordinates to the appropriate clipping region
	 * @param x1,y1		starting point of line
	 * @param x2,y2		end point of line
	 * @return true if a fraction of the line falls inside the clipping region
	 */
	bool	line(int* x1, int* y1, int* x2, int* y2) const;
}; // class ClipRegion

#endif
