/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 */

#include "fortran.h"

#include <stdio.h>
#include <string.h>

/** strip - strip a fortran character block from trailing spaces
 * @param line	fortran character string to be stripped
 */
int FortranParser::strip(char* line)
{
	int length = (int)strlen(line)-1;
	while (length>=0 && line[length]==' ')
		length--;
	line[length+1] = 0;
	return length;
} // strip

/** read a fortran block
 * @param buffer	buffer where data should be stored
 * @param maxSize	maximum length of buffer
 * @return length of data read
 */
int FortranFile::read(void* buffer, const int maxSize)
{
	int length, length2;
	if (fread(&length,sizeof(length),1,handle)<=0)
		return -1;
	if (length>maxSize) {
		fseek(handle,sizeof(length),SEEK_CUR);
		return 0;
	}
	if (fread(buffer,length,1,handle)<=0)
		return -1;
	if (fread(&length2,sizeof(length2),1,handle)<=0)
		return -1;
	if (length!=length2)
		return -1;
	return length;
} // read

/** read a fortran block inside parser
 * @param parser	parser holding the buffer to read data
 * @return length of data read
 */
int FortranFile::read(FortranParser& parser)
{
	int length = read(parser.buffer, parser.size());
	parser.set(length);
	return length;
} // read

/** allocate memory (with new) and read a fortran record
 * WARNING: The calling procedure is responsible for freeing the memory!!
 * @param size	return size of the buffer allocated
 */
byte* FortranFile::readBuffer(int* length)
{
	if (fread(length,sizeof(*length),1,handle)<=0)
		return NULL;

	byte* buffer = new byte[*length];
	if (buffer==NULL) return NULL;

	if (fread(buffer,*length,1,handle)<=0) {
		*length = -1;
		delete [] buffer;
		return NULL;
	}

	int length2;
	if (fread(&length2,sizeof(length2),1,handle)<=0) {
		*length = -1;
		delete [] buffer;
		return NULL;
	}
	if (*length!=length2) {
		*length = -1;
		delete [] buffer;
		return NULL;
	}
	return buffer;
} // readBuffer

/** write
 * @param buffer	write a buffer into file
 * @param len		length of data to write
 * @return length of data written (normally len)
 */
int FortranFile::write(const void* buffer, const int len)
{
	if (fwrite(&len, sizeof(len), 1, handle) != 1) return -1;
	if (fwrite(buffer, len, 1, handle) != 1)       return -1;
	if (fwrite(&len, sizeof(len), 1, handle) != 1) return -1;
	return len;
} // write

/** skip a fortran block
 * @return length of data skipped
 */
int FortranFile::skip()
{
	int length, length2;
	if (fread(&length,sizeof(length),1,handle)<=0)
		return -1;
	fseek(handle, length, SEEK_CUR);
	if (fread(&length2,sizeof(length2),1,handle)<=0)
		return -1;
	if (length!=length2)
		return -1;
	return length;
} // skip

/** backspace a fortran block
 * @return length of data skipped
 */
int FortranFile::backspace()
{
	int length;
	if (ftell(handle)<=8) return 0;
	fseek(handle, -4, SEEK_CUR);
	if (fread(&length,sizeof(length),1,handle)<=0)
		return -1;
	fseek(handle, -(length+8), SEEK_CUR);
	return length;
} // backspace

/** read fortran block size
 * @return size of block data
 */
int FortranFile::blockSize()
{
	int length;
	if (fread(&length,sizeof(length),1,handle)<=0)
		return -1;
	fseek(handle, -4, SEEK_CUR);
	return length;
} // blockSize

/** fortran size must be
 * @param length size of block
 * @return size of block data
 */
bool FortranFile::mustBe(int length)
{
	int length2;
	if (fread(&length2,sizeof(length2),1,handle)<=0)
		return false;
	return length==length2;
} // mustBe
