/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 */

#ifndef __TOOL_FORTRAN_H
#define __TOOL_FORTRAN_H

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "os.h"

class FortranFile;

// =============================== FortranParser ===============================
class FortranParser {
private:
	char*	buffer;			/** buffer pointer		*/
	char*	ptr;			/** current location pointer	*/
	char*	ptrEnd;			/** end of buffer pointer	*/
	size_t	_size;			/** buffer size			*/

public:
	FortranParser() : buffer(NULL), ptr(NULL), ptrEnd(NULL) {}
	FortranParser(void* buf, size_t len=0)		{ set(buf, len); }
	FortranParser(const FortranParser& p) : buffer(p.buffer), ptr(p.ptr), ptrEnd(p.ptrEnd) {}
	void	operator()(void* buf, int len=0)	{ set(buf, len); }
	void*	operator()()				{ return (void*)buffer; }
	void	set(void* buf, int len=0) {
			ptr    = buffer   = (char*)buf;
			ptrEnd = ptr+len;
			_size  = len;
		}
	void	set(int len) {
			ptr = buffer;
			if (len>0)
				ptrEnd = ptr + len;
			else
				ptrEnd = ptr;
		}
	void	reset()		{ ptr = buffer; }
	int	length() const	{ return ptr - buffer; }
	int	size() const	{ return _size; }
	bool	more()		{ return ptr<ptrEnd; }
	char*	getPtr()	{ return ptr; }

	void	skipWord()	{ ptr += sizeof(word); }
	void	skipInt()	{ ptr += sizeof(int); }
	void	skipFloat()	{ ptr += sizeof(float); }
	void	skipDouble()	{ ptr += sizeof(double); }

	// WARNING: the sc buffer should have a length of at least len+1
	int	read(char sc[], int len) {
			assert(ptr+len <= ptrEnd);
			memcpy(sc,ptr,len);
			sc[len] = 0;
			ptr += len;
			return strip(sc);
		}
	void	write(const char* sc, int len) {
			assert(ptr+len <= ptrEnd);
			memset(ptr, ' ', len);		// initialize to spaces
			memcpy(ptr, sc, strlen(sc));
			ptr += len;
		}
#ifdef ALIASING
	int	readWord() {
			char num[2];
			assert(ptr+sizeof(word) <= ptrEnd);
			num[0] = *ptr++; num[1] = *ptr++;
			return *reinterpret_cast<word*>(&num);
		}
	int	readInt() {
			assert(ptr+sizeof(int) <= ptrEnd);
			char num[sizeof(int)];
			memcpy(&num,ptr,sizeof(int));
			return *reinterpret_cast<int*>(&num);
		}
	float	readFloat() {
			char num[sizeof(float)];
			assert(ptr+sizeof(float) <= ptrEnd);
			memcpy(&num,ptr,sizeof(float));
			return *reinterpret_cast<float*>(&num);
		}
	double	readDouble() {
			char num[sizeof(double)];
			assert(ptr+sizeof(double) <= ptrEnd);
			memcpy(&num,ptr,sizeof(double));
			return *reinterpret_cast<double*>(&num);
		}

	void	write(const word a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			memcpy(ptr, &a, sizeof(a));
			ptr += sizeof(a);
		}
	void	write(const int a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			memcpy(ptr, &a, sizeof(a));
			ptr += sizeof(a);
		}
	void	write(const float a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			memcpy(ptr, &a, sizeof(a));
			ptr += sizeof(a);
		}
	void	write(const double a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			memcpy(ptr, &a, sizeof(a));
			ptr += sizeof(a);
		}
#else
	int	readWord() {
			assert(ptr+sizeof(word) <= ptrEnd);
			word a = *reinterpret_cast<word*>(ptr);
			ptr += sizeof(word);
			return a;
		}
	int	readInt() {
			assert(ptr+sizeof(int) <= ptrEnd);
			int a = *reinterpret_cast<int*>(ptr);
			ptr += sizeof(int);
			return a;
		}
	float	readFloat() {
			assert(ptr+sizeof(float) <= ptrEnd);
			float a = *reinterpret_cast<float*>(ptr);
			ptr += sizeof(float);
			return a;
		}
	double	readDouble() {
			assert(ptr+sizeof(double) <= ptrEnd);
			double a = *reinterpret_cast<double*>(ptr);
			ptr += sizeof(double);
			return a;
		}

	void	write(const word a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			*reinterpret_cast<word*>(ptr) = a;
			ptr += sizeof(a);
		}
	void	write(const int a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			*reinterpret_cast<int*>(ptr) = a;
			ptr += sizeof(a);
		}
	void	write(const float a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			*reinterpret_cast<float*>(ptr) = a;
			ptr += sizeof(a);
		}
	void	write(const double a) {
			assert(ptr+sizeof(a) <= ptrEnd);
			*reinterpret_cast<double*>(ptr) = a;
			ptr += sizeof(a);
		}
#endif

static	int	strip(char* line);

friend class FortranFile;
}; // FortranParser

// ================================ FortranFile ================================
class FortranFile {
public:
	FILE*	handle;			/** file handle		*/

public:
	FortranFile()		: handle(NULL)	{}
	FortranFile(const char* filename, const char* mode)	{ handle = fopen(filename, mode); }
	FortranFile(FILE* af)	: handle(af)	{}

	~FortranFile()	{ if (handle) fclose(handle); }

	bool	open(const char* filename, const char* mode) {
			handle = fopen(filename, mode);
			return handle!=NULL;
		}
	void	close() {
			if (handle) fclose(handle);
			handle = NULL;
		}

	//operator int()	{ return handle!=NULL && !feof(handle); }
	operator int()	const { return handle!=NULL; }
	bool	 eof()	const { return feof(handle); }

	int	seek(long offset, int whence=SEEK_SET)	{ return fseek(handle, offset, whence); }
	size_t	tell()					{ return ftell(handle); }
	int	getpos(fpos_t* position)		{ return fgetpos(handle, position); }
	int	setpos(fpos_t* position)		{ return fsetpos(handle, position); }

	int	readraw(void* buffer, int sc, int n)	{ return fread(buffer, sc, n, handle); }
	int	read(void* buffer, const int maxSize);
	byte*	readBuffer(int* length);
	int	read(FortranParser& parser);

	/** write buffer directly to handler */
	int	writeBuffer(const void* buffer, const int len) {
			if (fwrite(buffer, len, 1, handle) != 1) return -1;
			return len;
		}
	/** write a fortran record */
	int	write(const void* buffer, const int len);

	/** write a record from parser */
	int	write(FortranParser& parser)	{ return write(parser(), parser.length()); }
	void	write(float val)		{ fprintf(handle, "%f",  val); }
	void	write(double val)		{ fprintf(handle, "%lf", val); }
	void	write(int val)			{ fprintf(handle, "%d",  val); }
	void	write(char ch)			{ fputc(ch,   handle); }
	void	writeln()			{ fputc('\n', handle); }

	int	backspace();
	int	skip();
	int	blockSize();
	bool	mustBe(int length);
}; // class FortranFile

#endif
