/**
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	30-Jan-2019
 */

#ifndef __TOOL_STAT_H
#define __TOOL_STAT_H

#ifdef STAT
	#define IFSTAT(X)	X
#else
	#define IFSTAT(X)
#endif

#ifdef STAT
#include <assert.h>
#include <pthread.h>

#include <map>
#include <string>
#include <ostream>

#include "timer.h"

//class Stat : public std::map<std::string, size_t> {
class Stat {
private:
	std::map<std::string, size_t>	counter;
	std::map<std::string, CPUTime>	totaltime;
	std::map<std::string, CPUTime>	starttime;
mutable	pthread_mutex_t			mutex;

public:
	Stat()  { pthread_mutex_init(&mutex, NULL); }
	~Stat() { pthread_mutex_destroy(&mutex); }

	/* threads */
	void	lock()	const	{ pthread_mutex_lock(&mutex);   }
	void	unlock() const	{ pthread_mutex_unlock(&mutex); }

	void	reset()	{
			lock();
			counter.clear();
			totaltime.clear();
			starttime.clear();
			unlock();
		}

	// Counter functions
		operator bool()	const			{ return not counter.empty(); }
	size_t	operator ()(const std::string& key)	{
				lock();
				size_t value = _inc(key);
				unlock();
				return value;
			}

	size_t	inc(const std::string& key, const size_t amount=1) {
			lock();
			size_t value = _inc(key, amount);
			unlock();
			return value;
		}
	size_t	operator [](const std::string& key) {
				lock();
				size_t value = counter[key];
				unlock();
				return value;
			}

	// Timer functions
	void	start(const std::string& key) {
				lock();
				_inc(key);			// increase main counter
				assert(starttime[key].tv_nsec==0);
				starttime[key].get();
				unlock();
			}

	void	stop(const std::string& key) {
				lock();
				CPUTime& to = starttime[key];
				assert(to.tv_nsec!=0);
				CPUTime tp(true);
				totaltime[key] += tp-to;
#if _DEBUG>0
				starttime[key].tv_sec = 0;
				starttime[key].tv_nsec = 0;
#endif
				unlock();
			}

	// Merge statistics from another class
	void	merge(const Stat& stat) {
				lock();
				stat.lock();
				for (auto it : stat.counter) {
					_inc(it.first, it.second);
					if (stat.totaltime.find(it.first) != stat.totaltime.end()) {
						try {
							totaltime.at(it.first) += stat.totaltime.at(it.first);
						} catch (const std::out_of_range&) {
							totaltime[it.first] = stat.totaltime.at(it.first);
						}
					}
				}
				stat.unlock();
				unlock();
			}
	Stat&	operator += (const Stat& stat)	{ merge(stat); return *this; }

	/** @return maximum length of all keys */
	size_t	maxlength() const {
				lock();
				size_t maxlen = 0;
				for (auto it : counter)
					maxlen = std::max(maxlen, it.first.length());
				unlock();
				return maxlen;
			}

private:
	/** inc without a lock */
	size_t	_inc(const std::string& key, const size_t amount=1) {
			try {
				counter.at(key) += amount;
				return counter[key];
			} catch (const std::out_of_range&) {
				counter[key] = amount;
				return amount;
			}
		}

friend	std::ostream& operator << (std::ostream&, const Stat&);
}; // class Stat

std::ostream& operator << (std::ostream&, const Stat&);

#endif
#endif
