/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author: Vasilis.Vlachoudis@cern.ch, siu_coders@gmail.com
 * Based on original code from David Sinuela Pastor siu_coders@gmail.com
 * Date:	15-Jan-2013
 *
 */
#ifndef __TOOL_THREADPOOL_H
#define __TOOL_THREADPOOL_H

#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>

class ThreadPool;
class ThreadPoolFeeder;

/* ========================= ThreadPoolWorker ========================= */
/** A work unit for the TreadPool */
class ThreadPoolWorker {
protected:
	ThreadPoolFeeder*	_feeder;	// to recover easier the information...
public:
	ThreadPoolWorker(ThreadPoolFeeder* f=NULL) : _feeder(f) {}
virtual	~ThreadPoolWorker() {}		// Destroy

virtual	void	operator()() = 0;		// Perform the work
virtual	void	feeder(ThreadPoolFeeder* f)	{ _feeder = f; }
virtual	ThreadPoolFeeder* feeder()		{ return _feeder; }
}; // ThreadPoolWorker

/* ========================= ThreadPoolFeeder ========================= */
/** Worker provider unit for the threadpool.
 * It has to be subclassed to send work to the threadpool
 */
class ThreadPoolFeeder {
protected:
	ThreadPool&	pool;	// reference to pool
//	int		njobs;	// number of running jobs

public:
	ThreadPoolFeeder(ThreadPool& p) : pool(p) /*,njobs(0)*/ {}
virtual	~ThreadPoolFeeder() {}

virtual	ThreadPoolWorker*	feed(int threadId) = 0;
const	ThreadPool&		threadpool()	const { return pool; }

//	void	inc()	{ njobs++; }
//	void	dec()	{ njobs--; }
}; // ThreadPoolFeeder

/* ============================ ThreadPool ============================ */
class ThreadPool {
private:
	/** Internal structure for linking the running threads with the threadpool */
	struct _ThreadData {
		int		id;
		pthread_t	thread;
		ThreadPool*	pool;
	}; // _ThreadData

private:
	int	_nthreads;			// Number of threads
	int	activeJobs;			// Number of active jobs
	bool	requestThreadExit;		// request threads to end
volatile bool	stopped;			// flag exited normally or stopped

	_ThreadData*		threads;	// Thread data
	ThreadPoolFeeder*	feeder;		// Feeder providing workers
	pthread_mutex_t		mutex;
	pthread_cond_t		condFeeder;
	pthread_cond_t		condEnd;

static const long	DT = 100000000L;		// by 100ms in ns

public:
	ThreadPool();
	~ThreadPool();

	int	nthreads()	const	{ return _nthreads; }

	bool	init(int n);
	void	end();

	bool	execute(ThreadPoolFeeder* f);
	void	stop();
	bool	isStopped()	const	{ return stopped; }

	ThreadPoolWorker* fetchWork(int thread_id);

	bool	isRunning()	{ return feeder!=NULL || activeJobs>0; }
	void	reset()		{ stopped = false; }

static	int	num_cores()	{ return (int)sysconf(_SC_NPROCESSORS_ONLN); }

private:
	int	lock()		{ return pthread_mutex_lock(&mutex); }
	int	unlock()	{ return pthread_mutex_unlock(&mutex); }
	void	decActiveJobs();

static	void*	threadExecute(void* param);
static	void	getTimeout(struct timespec* ts, long dt=DT);
}; // ThreadPool

#endif
