/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 *
 * TODO
 * - draw with hash lines internal of bodies (cut by clipbody)
 */

#include "d3layer.h"

#include <assert.h>

#include <thread>
#include <future>
#include <utility>

#include "viewer.h"
#include "math/bit.h"
#include "math/morton.h"
#include "geometry/mesh.h"
#include "math/randsampler.h"

#if _DEBUG>0
#	include "tool/timer.h"
#endif
const dword FLAG_IGNORE = FLAG_REGION | FLAG_LATTICE | FLAG_FINAL;

#ifdef THREAD
/* =========================== Body3DWorker =========================== */
/** operator() */
void Body3DWorker::operator()()
{
	if (*ptr & FLAG_IGNORE) return;
	d3->drawPixelAA(engine, ray, *painter, ptr,
			W, H, ii, jj,
			step, &zone);
	*ptr |= FLAG_3D;
} // operator

/* =========================== RayShooterWorker =========================== */
// Worker thread which shoots rays for a specific (sub)pixel
// & returns the zone and color
/** operator() */
//void RayShooterWorker::operator()()
//{
//	color = d3->shootRay(u, v, ray, engine, &zone);
//} // operator

/* =========================== Body3DFeeder =========================== */
/** allocate */
void Body3DFeeder::allocate()
{
	if (nworkers != pool.nthreads()) {
		if (workers) delete [] workers;
		workers  = new Body3DWorker[pool.nthreads()];
		nworkers = pool.nthreads();
	}
} // allocate

/** reset */
void Body3DFeeder::reset(ZPainter* p, Ray& ray)
{
	pool.reset();
	allocate();

	// painter width&height
	painter = p;
	W = painter->width();
	H = painter->height();

	// initialize workers
	for (int k=0; k<nworkers; k++) {
		workers[k].d3      = d3;
		workers[k].painter = painter;
		workers[k].engine  = d3->kernel.engine(k);
		workers[k].ray     = ray;
		workers[k].zone    = nullptr;
		workers[k].W       = W;
		workers[k].H       = H;
		workers[k].color   = p->background();
	}
} // reset

/** init */
void Body3DFeeder::init(int astep)
{
	// nested loops initialization
	step = astep;
	ii   = 0;
	jj   = 0;
	ptr = painter->data();
#ifndef MORTON
	pline = ptr;
#else
	is   = 0;
	js   = 0;
	//step_shift = ...
	Ws   = W / step;
	Hs   = H / step;
	//Ws   = W >> step_shift;
	//Hs   = H >> step_shift;
#endif
} // init

/** loop - for next pixel
 * @return true if we have finished, false otherwise
 */
bool Body3DFeeder::loop()
{
	while (true) {
#ifndef MORTON
		// Original: line scanning
		// (ii) loop
		ii  += step;
		ptr += step;
		if (ii>=W) {
			// (jj) loop
			jj += step;
			if (jj>=H) return true;
			pline += step*W;
			ii = 0;
			ptr = pline;
		}
#else
		// space filling Z-curve to profit from cache coherence
		if (!Morton::next2D(is,js, Ws,Hs)) return true;
		ii = is*step;
		jj = js*step;
		//ii = is << step_shift;
		//jj = js << step_shift;
		assert(0<=ii and ii<W);
		assert(0<=jj and jj<H);
		ptr = painter->pixelPtr(ii,jj);
#endif
		if (*ptr & FLAG_3D) continue;
		return false;
	}
} // loop

/** feed */
ThreadPoolWorker* Body3DFeeder::feed(int threadId)
{
	// Check for ending condition
	if (d3->viewer.stop()) return nullptr;

	// update worker
	workers[threadId].ii   = ii;
	workers[threadId].jj   = jj;
	workers[threadId].ptr  = ptr;
	workers[threadId].step = step;
	workers[threadId].painter = painter;

	if (loop()) return nullptr;
	return (ThreadPoolWorker*)&workers[threadId];
} // feed
#endif

/* ============================== D3Layer ============================= */
D3Layer::D3Layer(const Geometry& g, GeometryKernel& k, GeometryViewer& v) :
		Layer(g, k, v)
#ifdef THREAD
		, feeder(k.threadpool, this)
#endif
{
	deflights     = true;
	ambient       = 0x30;		// ambient intensity
	xray          = 0;
	antialias     = 1;
	drawEdges     = false;
	shadows       = true;
	skip_1stblack = false;
	_step         = 32;		// initial guess
	_maxDepth     = 0;
	_usrbinastexture = true;

	_beautyPassFinished = false;

	// intialize FARM renderer
	m_RenderObj.Init(this, &(kernel.voxel), const_cast<Geometry*>(&geometry));
	m_RenderObj.SetRenderMode(CSurfaceIntegrator::cSpeed);

	drawTime(15);
} // D3Layer

/** nextIntersection
 * @return false if the ray is ended or on error, true otherwise
 */
bool D3Layer::nextIntersection(GeometryEngine* engine, Ray* ray)
{
	Point hit;

INT:	if (engine->intersectRay(ray, false))
		// End of ray, maybe we should return sky color?
		return false;

	RaySegment& segment = ray->segment();
	VZone* zone = segment.zone;

	if (zone == nullptr) {
		ray->error = true;
		return false;
	}

	// Set default normal the opposite direction of the beam
	ray->normal = -ray->segment().dir;

	int n = ray->n;

	if (ray->project_hit)
		/* do nothing */;
	else
	if (zone->region()->type() == GRegion::Type::BLACKHOLE) // Blackhole
		return false;
	else
	if (zone->region()->type() == GRegion::Type::VOXEL) {
		// Inside voxel region
		hit = segment.hit(SMALL3D3);
		ray->voxelreg = viewer.voxel().get(hit);
		if (ray->voxelreg<0 || viewer.voxel().color(ray->voxelreg)==COLOR_TRANSPARENT) {
			// End of the voxel or precision error
			ray->moveby(SMALL3D3);
			goto INT;		// treat as transparent
		} else {
			if (segment.body && ray->clip_hit) {
				ray->normal = segment.body->body()->normal(hit);
				if (ray->normal*segment.dir > 0.0)
					ray->normal.negate();
			} else
			if (!ray->start())
				ray->normal = geometry.voxel.normal(hit, segment.dir);
		}
	} else {
		// Normal region
		// Check if we are at the entry of a lattice
		// in order to avoid unnecessary transformations
		while (n>0 && ray->segment(n).body == nullptr) n--;
		if (ray->segment(n).body != nullptr) {
			hit = ray->segment(n).hit(-SMALL3D3);
			// Find body normal
			ray->normal = ray->segment(n).body->body()->normal(hit);
			// check that the normal is opposite to the direction
			if (ray->normal*ray->segment(n).dir > 0.0)
				ray->normal.negate();
		}
	}
	// FIXME Transform normal due to lattice if any?
	if (n>0) {
		IFSTAT(stat("shade_correctRotdefiNormal"));
		for (; n>=0; n--)
			if (ray->segment(n).region)
				ray->normal = ray->segment(n).invMatrix() * ray->normal;
	}

	return true;
} // nextIntersection

/**
 * @param eng	current geometry engine
 * @param ray	ray object
 * @param step	step size/pixel size
 * @return the final color of the ray
 */
Color D3Layer::shade(GeometryEngine* eng, Ray* ray, const int step)
{
	Color color, dataColor;

	IFSTAT(stat("shade_call_count"));

INT:	if (!nextIntersection(eng, ray))
		return color;	// on end of ray, error or blackbody

	// Get region/voxel color
	if (ray->voxelreg >= 0) {
		if (ray->max_depth < 1)
			//color.set(viewer.voxel().roiColor(ray->voxelreg));	// old: voxel color
			color = m_RenderObj.Shade(eng, ray, step, light0);
		else
			color = m_RenderObj.ShadeVoxel(eng, ray, step, light0);  //  new volume renderer
	} else {
		// use new render object for shading
		color = m_RenderObj.Shade(eng, ray, step, light0);
	}

	// Check for USRBIN
	if (viewer.usrbin.show) {
		if (usrbinAsTexture() || ray->project_hit) {
			ray->project_alpha = viewer.usrbin.shade(ray->hit(SMALL3D3), dataColor);
			if (ray->project_hit && ray->project_alpha == 0xFF) {
				ray->project_hit = false;
				ray->moveby(SMALL3D3);
				goto INT;
			}
			if (ray->project_alpha == 0xFF && ray->voxelreg == -1) {
				// no USRBIN hit -> branch to FARM
				color = m_RenderObj.Shade(eng, ray, step, light0);
			} else
			if (ray->project_alpha != 0xFF && ray->voxelreg == -1) {
				 // Apply shadows to data. By default this is off as
				 // it's clearer for the user but it can be activated
				 // in the farm.ini config file via the option
				 // "usrbin-shadow = 1"
				// USRBIN hit -> branch to FARM to handle shadows
				double shadowIntensity = m_RenderObj.ShadeData(eng, ray, step, light0);
				color = dataColor * shadowIntensity;
			} else
			if (ray->voxelreg >= 0 && ray->project_alpha != 0xFF) {
				// within USRBIN but no data or transparent data
				// pure texture projection = default quality
				if (ray->max_depth < 1) {
					// voxel color
					Color voxelCol(viewer.voxel().roiColor(ray->voxelreg));
					// we scale the returned USRBIN white with the shaded voxel color
					color = dataColor * voxelCol.luminance();
				} else {
					// higher quality
					 // NEW: blending of USRBIN and volume rendering
					color = m_RenderObj.ShadeVoxel(eng, ray, step, light0);
					if (dataColor != COLOR_TRANSPARENT)
						color.blend(dataColor, 0.50 + (float)ray->project_alpha / 255.0f);
				}
			}
		}
	}
	return color;
} // shade

/** shadeXray
 * Computes the color of a pixel
 * for transparent objects it invokes iteratively @see shade()
 * until the color contribution is below the threshold
 */
Color32 D3Layer::shadeXray(GeometryEngine* eng, Ray* ray,
			const double u, const double v,
			const int step, int alpha,
			VRegion* last_region)
{
	// alpha=255 first level
	VZone* zone = ray->segment(0).zone;
	while (true) {
		int transparency;
		bool shadowsChanged;

		// Remove shadows for deeper levels
		if (alpha!=255 && ray->shadow) {
			ray->shadow = false;
			shadowsChanged = true;
		} else
			shadowsChanged = false;

		Color32 thisColor = shade(eng, ray, step).color32();

		// restore shadows
		if (shadowsChanged) ray->shadow = true;

		if ( alpha==255       &&
		    _edgeDetect       &&
		    !ray->project_hit &&
		    ray->hitZone()    &&	// don't call for VOXELS
		    ray->hitRegion()->type() != GRegion::Type::VOXEL) {
			// remember checkid to restore in case of no edge
			int oldBodyCheckId = eng->bodyCheckId();
			if (edgePixel(eng, *ray, u,v, zone)) return 0;
			eng->bodyCheckId(oldBodyCheckId);
		}
		thisColor.alpha(0);

		if (ray->ended()   ||
		   !ray->hitZone() ||
		    ray->hitRegion()->type() == GRegion::Type::BLACKHOLE)
			return geometry._transparentColor;
		else
		if (ray->hitRegion() == last_region &&
		    ray->hitRegion()->type() != GRegion::Type::VOXEL) {
			ray->moveby(SMALL3D3);
			ray->skip_current = true;
			continue;
		}

		if (ray->project_hit)
			transparency = ray->project_alpha;
		else
			transparency = (xray>0)? xray: ray->hitRegion()->alpha();

		int nextAlpha = byteMul(alpha, transparency)-1;
		if (nextAlpha < 1)
			return thisColor;

		ray->skip_transparent = !ray->skip_transparent;
		Color32 nextColor = shadeXray(eng, ray, u, v, step, nextAlpha, ray->hitRegion());
		nextColor.alpha(255);
		thisColor.alpha(transparency ^ 0xff);	// 255-transparency = transparency^0xff
		thisColor.blend(nextColor);
		thisColor.alpha(0);			// "Fix" color.alpha
		return thisColor;
	}
} // shadeXray

/** @return true if pixel is on edge */
bool D3Layer::edgePixel(GeometryEngine* eng, Ray& ray, const double u, const double v, VZone* zone)
{
	// Sample few rays to detect change in depth, normal, region
	Point  hit[5];		// hit position
	Point  projhit[5];	// hit position projected to the hit-plane of center pixel
	Vector normal[5];	// normal at hit position
	CBody* body[5];

	// Do not return edge for zone=nullptr or BLACKBODY
	if (!ray.hitZone() or
	     ray.ended() or
	     ray.hitRegion()->type() == GRegion::Type::BLACKHOLE) return false;

	double depth = ray.T();

	hit[0]    = projhit[0] = ray.hit();
	normal[0] = ray.normal;
	body[0]   = ray.hitBody();

	Ray edgeray;
	edgeray.use_clip      = ray.use_clip;
	edgeray.skip_1stblack = ray.skip_1stblack;

	double x=0.0,y=0.0,z=0.0;
	if (view().projection == ViewPort::Projection::Perspective)
		view().cameraPosition(&x, &y, &z);

	int cid = 1;
	double du = view().Dx();
	double dv = view().Dy();
	double uu = u - du/2.0;
	for (int i=2; i--; uu+=du) {
		double vv = v - dv/2.0;
		for (int j=2; j--; vv+=dv, cid++) {
			double dx, dy, dz;
			if (view().projection!=ViewPort::Projection::Perspective) {
				x = view().uv2x(uu,vv);
				y = view().uv2y(uu,vv);
				z = view().uv2z(uu,vv);
			}
			view().rayDirection(uu,vv, &dx,&dy,&dz);
			eng->incBodyCheckId();
			zone = eng->whereRay(x,y,z, dx,dy,dz, SMALL3D, zone);
			edgeray.init();
			edgeray.push(RaySegment(x,y,z, dx,dy,dz, zone));
			nextIntersection(eng, &edgeray);

			if (!edgeray.hitZone()) return true;
			if (edgeray.ended()) return true;
			if (edgeray.hitRegion()->type() == GRegion::Type::BLACKHOLE) return true;
			if (edgeray.hitRegion() != ray.hitRegion()) return true;

			hit[cid]     = edgeray.hit();
			projhit[cid] = edgeray.dir() * depth + edgeray.pos();
			normal[cid]  = edgeray.normal;
			body[cid]    = edgeray.hitBody();

			double cosf = normal[cid]*ray.dir();
			//if (Abs(cosf)<0.1) return true;

			for (int k=0; k<cid; k++) {
				double dot = normal[cid] * normal[k];
				if ((body[k]==body[cid] && dot<0.5) or
				    (body[k]!=body[cid] && dot<0.99999)) return true;
				else
				if (dot>0.99999) {
					double be = (hit[cid]-projhit[k]).length2();
					double bd = (hit[cid]-hit[k]).length2() * Sq(cosf);
					if (bd > 2.0 * be) return true;
				}
			}
		}
	}
	return false;
} // edgePixel

/** determine the color of a pixel by shooting a ray*/
dword D3Layer::shootRay(double u, double v, Ray& ray,
			GeometryEngine* engine,
			VZone** zone, int step /*= 1*/)
{
	Point  p = view().rayPosition(u, v);
	Vector d = view().rayDirection(u, v);

	// ATTENTION: if we work exclusively with BVH we should not set a starting zone
	// because the whereRay() call would return the VOID/BLACKHOLE if the camera
	// is inside and then we would intersect the ray directly with this region.
	//
	// Using the standard cache works because during the first call the isortedZones()
	// that is used for determining the location is still empty.
	//
	// CT 18/11/2020
	engine->incBodyCheckId();
	*zone = engine->whereRay(p, d, SMALL3D, *zone);

	// prepare the ray
	ray.init();
	ray.push(RaySegment(p, d, *zone));

	// fire ray to determine color
	return shadeXray(engine, &ray, u, v, step);
} // shootRay

/** drawPixel
 * shoot one ray and fill pixel and adjacent square step x step pixels
 * with the corresponding color.
 * Draw with a hash pattern if the region/zone is selected
 */
dword D3Layer::drawPixel(GeometryEngine* engine, Ray& ray,
			ZPainter& painter, dword* ptr,
			const int W, const int H,
			const int i, const int j,
			const int step, const int nsample,
			const double u, const double v,
			VZone** zone)
{
	dword color = shootRay(u, v, ray, engine, zone, step);
	IFSTAT(stat("drawPixel_call"));

	bool region_selected = false;
	bool zone_selected   = false;
	dword color2 = 0;
	if (ray.error && _showErrors && ((i+j)&15)==0) {
		color = geometry.errorColor;
		kernel.setError3D();
	} else
	if (ray.hitZone() && ray.hitRegion()->show() & BIT_SELECT) {
		Color32 col;
		col = ray.hitRegion()->color();
		col.red()   = col.red()  >0x7F? 0:0xFF;
		col.green() = col.green()>0x7F? 0:0xFF;
		col.blue()  = col.blue() >0x7F? 0:0xFF;
		color2 = col;
		region_selected = true;
	}

	// FIXME not the best way but works
	// Maybe I should add a selected flag inside the zone
	if (geometry.editRegion.size()) {
		Point hit = ray.hit();
		Vector d  = ray.dir();
		geometry.lockEdit();
		if (geometry.editRegion.inside(hit, d)) {
			color2 = 0xFF00FF;
			zone_selected = true;
		}
		geometry.unlockEdit();
	}

	if ((color&FLAG_TRANSPARENT) and ((i^j)&8))
		color = Color32::darker(color, 200);

	painter.zpixel(i,j, ray.T());
	for (int jj=0; jj<step; jj++, ptr+=W) {
		if (j+jj>=H) break;
		int step2 = Min(step, W-i);
		for (int ii=0; ii<step2; ii++) {
			dword flag = ptr[ii] & FLAG_INFOMASK;
			if (flag & FLAG_IGNORE) continue;
			Color32 new_color;
			if ((region_selected && (!(((i+ii+j+jj)&3) || ((j^jj)&1)))) or
			    (zone_selected   && ((i^ii^j^jj)&1)))
				new_color = color2;
			else
				new_color = color;
			Color32 avg_color = ptr[ii];
			avg_color.average(new_color, nsample);
			if (ii or jj)
				ptr[ii] = (avg_color&FLAG_COLORMASK) | flag;	// restore flag
			else
				ptr[ii] = avg_color;
		}
	}
	return color;
} // drawPixel

/** drawPixelAA
 * adaptive antialias wrapper to drawPixel calls
 */
void D3Layer::drawPixelAA(GeometryEngine* engine, Ray& ray,
			ZPainter& painter, dword* ptr,
			const int W, const int H,
			const int ii, const int jj,
			const int step,
			VZone** zone)
{
#if 0
	int sample = 1;
	for (int sx=0; sx<_samples; sx++, sample++) {
		// define sub-pixel offset based on the super sample
		double u = view().i2u((double)ii + ((double)sx+0.5)/(double)_samples);
		for (int sy=0; sy < _samples; sy++) {
			// define sub-pixel offset based on the super sample
			double v = view().j2v((double)jj + ((double)sy+0.5)/(double)_samples);
			drawPixel(engine, ray,
					painter, ptr,
					W, H,
					ii, jj,
					step, sample,
					u, v,
					zone);
		}
	}
#else
	// let's start with the original color as a basis
	double uc = view().ic2u(ii);
	double vc = view().jc2v(jj);
	Color accumulatedColor = drawPixel(engine, ray,
					painter, ptr,
					W, H,
					ii, jj,
					step, 1,
					uc, vc,
					zone);

	if (_samples>1) {
		//////////////////////////////////////////////////////////////////////////////
		// New adaptive anti-aliasing based on convergence & QMC series
		// FIXME Ideally all calcs would be done on Color and not on Color32
		// to avoid loss of precision.
		unsigned long long pScramble[2] = { 2346349678, 1340569164 };
		double		 pRandSample[2];
		const double	 cMaxDeltaColor = 0.01;

		// sample pixel center
		double totalFilterWeight = 1.0;

		double du = view().Dx();
		double dv = view().Dy();

		// start minimum sampling
		for (int t=1; t<_samples; ++t) {
			// LD sampling - introducing a bit of variation,
			// that acts similarly to blue noise, increases the AA quality
			CBaseSampler::Sample02Dbl(t, pScramble, pRandSample);
			assert(InRange(0.0,pRandSample[0],1.0));
			assert(InRange(0.0,pRandSample[1],1.0));
			double u = uc + du*(pRandSample[0]-0.5);
			double v = vc + du*(pRandSample[1]-0.5);
			Color current = drawPixel(engine, ray,
						painter, ptr,
						W, H,
						ii, jj,
						step, t+1,
						u, v,
						zone);
			// apply DSP filter
			double signalFilter = pixelFilter().Evaluate(
							pRandSample[0]-0.5,
							pRandSample[1]-0.5);
			// we use Color objects instead of Color32 to retain small nuances
			accumulatedColor += current * signalFilter;

			totalFilterWeight += signalFilter;
		} // end of minimum sampling

		Color prevAccumColor = accumulatedColor;

		// continue adding samples adaptively. we separate this and the
		// previous loop to allow for loop-unrolling and vectorization of the
		// first loop due to the avoided convergence conditional. The second
		// loop goes up to samples^2 considering distribution in U and V
		for (int t=_samples; t < _samples * _samples; ++t) {
			CBaseSampler::Sample02Dbl(t, pScramble, pRandSample);
			assert(InRange(0.0,pRandSample[0],1.0));
			assert(InRange(0.0,pRandSample[1],1.0));
			double u = uc + du*(pRandSample[0]-0.5);
			double v = vc + dv*(pRandSample[1]-0.5);
			Color current = drawPixel(engine, ray,
						painter, ptr,
						W, H,
						ii, jj,
						step, t+1,
						u, v,
						zone);
			// apply DSP filter
			double signalFilter = pixelFilter().Evaluate(
							pRandSample[0]-0.5,
							pRandSample[1]-0.5);
			// we use Color objects instead of Color32 to retain small nuances
			accumulatedColor  += current * signalFilter;
			totalFilterWeight += signalFilter;
			Color delta = accumulatedColor - prevAccumColor;

			// check convergence
			if (Abs(delta.red())   < cMaxDeltaColor &&
			    Abs(delta.green()) < cMaxDeltaColor &&
			    Abs(delta.blue())  < cMaxDeltaColor) {
#if 0
				if (t>2)
				std::cerr << "Bailing out at t = " << t
					<< "/" << _samples * _samples
					<< " delta=" << delta
					<< " accu=" << accumulatedColor
					<< " prev=" << prevAccumColor
					<< " filter=" << signalFilter
					<< " du=" << pRandSample[0]
					<< " dv=" << pRandSample[1]
					<< std::endl;
#endif
				break;
			}
			prevAccumColor = accumulatedColor;
		} // end of maximum sampling loop
		// re-normalize accumulated color
		//accumulatedColor /= totalFilterWeight;
	}
#endif
} // drawPixelAA

/** draw */
void D3Layer::draw(ZPainter& painter, bool checkTime)
{
	// FIXME ray... Should move to workers
	Ray ray;
	int step = 0;

	_beautyPassFinished = false;
	kernel.clearError3D();

#ifdef STAT
	stat.reset();
	kernel.resetStatistics();
#endif

	// Check for the default lights
	int fromlight, tolight;
	if (deflights) {
		fromlight = light0 = 0;
		tolight   = geometry.deflights;
	} else {
		fromlight = light0 = geometry.deflights;
		tolight   = geometry.nlights;
	}

	clock_t starttime = clock();
	clock_t endtime;
	if (checkTime && _drawTime) {
		endtime     = starttime + _drawTime;	// find ending time
		ray.shadow  = false;
		ray.nlights = Min(2, tolight-fromlight);

		_edgeDetect = false;
		_showErrors = false;
		_samples    = 1;
	} else {
		endtime     = 0;
		ray.shadow  = shadows;
		ray.nlights = tolight - fromlight;

		_edgeDetect = drawEdges;
		_showErrors = viewer.showErrors;
		_samples    = antialias;
	}
	ray.use_clip      = kernel.clipBodyCount()    > 0;
	ray.use_project   = kernel.projectBodyCount() > 0;
	ray.max_depth     = _maxDepth;
	ray.skip_1stblack = skip_1stblack;

	// Convert lights from relative to absolute position and direction
	for (int l=0; fromlight<tolight; l++, fromlight++) {
		if (geometry.light[fromlight].relative) {
			light[l].pos = view().matrix() * geometry.light[fromlight].pos;
			light[l].dir = view().matrix() * geometry.light[fromlight].dir;
			light[l].dir.renormalize();
		} else {
			light[l].pos = geometry.light[fromlight].pos;
			light[l].dir = geometry.light[fromlight].dir;
		}
	}

	geometry.lockRead();
#ifdef THREAD
	if (kernel.nthreads()) {
		// Multi threaded version using the threadpool
		if (kernel.nBodies()) {
			assert(!kernel.isRunning());
			feeder.reset(&painter, ray);
			//int step_shift =
			for (step=_step; step>0; step>>=1) {
				feeder.init(step);
				if (kernel.threadpool.execute(&feeder))
					goto END;
				// Stop if we have eaten up our time
				if (endtime and clock()>endtime) break;
			}

			// once we have finished our beauty pass we signal this and postDraw()
			// effects like adaptive antialiasing can be rendered by the viewer
			// in a deferred pass
			if (step <= 1)
				_beautyPassFinished = true;
		}
		assert(!kernel.isRunning());
	} else
#endif
	{	// Single threaded version, active also when #nthreads()=0 (cores=1)
		int H = painter.height();	// save variables that might change
		int W = painter.width();	// during the scanning...
		VZone* zone = nullptr;

		// scan in steps
		for (step=_step; step>0; step>>=1) {
			dword* pline = painter.data();
			for (int j=0; j<H; j+=step) {
				dword* ptr = pline;
				for (int i=0; i<W; i+=step, ptr+=step) {
					if (*ptr & FLAG_IGNORE) continue;
					if (stop()) goto END;
					drawPixelAA(engine(), ray,
							painter, ptr,
							W, H,
							i, j,
							step,
							&zone);
					*ptr |= FLAG_3D;
				}
				pline += step*W;
			}
			// Stop if we have eaten up our time
			if (endtime && clock()>endtime) break;
		}
	}
END:
	geometry.unlockRead();

	// adaptive step adjust _step if needed
	if (endtime) {
		clock_t dt = clock() - starttime;
		if (dt > _drawTime*4) {		// _step goes with the square (surface)
			_step <<= 1;
			if (_step>64) _step = 64;
		} else
		if (dt < _drawTime/4) {
			_step >>= 1;
			if (_step==0) _step = 1;
		}
	}
#if _DEBUG>1
	std::cout << "D3Layer::draw() time="
			<< (double)(clock() - starttime)/(double)CLOCKS_PER_SEC
			<< "s  check=" << checkTime << std::endl;
#endif
#ifdef STAT
	std::cout << "Stats\n" << stat << std::endl;
	kernel.mergeStatistics();
	std::cout << "Engine Statistics\n" << engine()->stat << std::endl;
#endif
} // draw

/** postDraw
 * Render any effect (e.g. adaptive antialiasing) that should be applied in the
 * phase after the beauty pass has been successfully rendered.
 * The postDraw effects are calculated by a background thread and displayed
 * as soon as they are ready.
 * [8/19/2019 theis]
 */
void D3Layer::postDraw(ZPainter& painter)
{
	if (stop()) return;
	if (_beautyPassFinished and antialias>=4) {
		geometry.lockRead();
		drawAdaptiveAAPass(painter);
		geometry.unlockRead();
	}
} // postDraw

/** drawWireframe */
void D3Layer::drawWireframe(ZPainter& painter)
{
	geometry.lockRead();
	for (auto body : kernel.bodies) {
		if (stop()) break;
		if (body->show() & BIT_WIREFRAME)
			drawWireframe(painter, body);
	}
	geometry.unlockRead();
} // drawWireframe

/** drawWireframe */
void D3Layer::drawWireframe(ZPainter& painter, VBody* body)
{
	const Color3D& color = (body->show() & BIT_SELECT)?
	//const dword color = (body->show() & BIT_SELECT)?
			geometry.select3DColor : geometry.wireframeColor;

	/* let's use brute force */
	GBody* gbody = body->body();
	Mesh* mesh = gbody->mesh();
	for (size_t i=0; i<mesh->nedges(); i++) {
		const Edge* edge = mesh->edge(i);
		if (edge->show && draw3Dline(painter, edge->A(), edge->B(), color))
			body->visible = true;
	}

	for (int n=0; n<gbody->nodes(); n++) {
		Point r;
		view().xyz2uvw3D(gbody->node(n), &r);
		if (show || Eq0(r.z,SMALL3D)) {
			int x = view().u2i(r.x);
			int y = view().v2j(r.y);
			if (painter.rectangle(x-1,y-1,x+1,y+1, color.color))
				body->visible = true;
		}
	}

#ifdef OPENGL
#if 0
	// FIXME VERY PRIMITIVE
	gbody->mesh.clearProcessedFlag();
	for (int i=0; i<gbody->mesh.faceCount(); i++) {
		Face* f = gbody->mesh.face(i);
		for (int j=0; j<3; j++) {
			if (!f->show(j)) continue;
			Face* fn = f->neighbor(j);
			if (fn && fn->processed()) continue;
			int jn = (j+1)%3;
			Point* A = f->vertex(j);
			Point* B = f->vertex(jn);
			draw3Dline(painter, *A, *B, color);
		}
		f->processed(true);
	}
#endif
#endif
} // drawWireframe

/** drawBodiesBBox */
void D3Layer::drawBodiesBBox(ZPainter& painter)
{
	if (stop()) return;
	geometry.lockRead();;
	for (size_t i=0; i<kernel.nBodies();i++) {
		VBody* body = kernel.getBody(i);
		if ((body->show() & BIT_SELBBOX) == BIT_SELBBOX) {
			drawBBox(painter, body->body()->bbox(),
					geometry.bodyBBoxOutColor);
#ifdef USE_OBBOX
			drawOBBox(painter, body->body()->obbox(false),
					geometry.bodyBBoxOutColor);
			drawOBBox(painter, body->body()->obbox(true),
					geometry.bodyBBoxInColor);
#endif
		}
	}
	geometry.unlockRead();;
} // drawBodiesBBox

/** drawRegionsBBox */
void D3Layer::drawRegionsBBox(ZPainter& painter)
{
	if (stop()) return;
	geometry.lockRead();;
	// FIXME scan regions before if BBOX is selected
	for (auto region : geometry.regions) {
		if ((region->show & BIT_SELBBOX) == BIT_SELBBOX) {
			drawBBox(painter,  region->bbox(),  geometry.regionBBoxColor);
//			drawOBBox(painter, region->obbox(), geometry.zoneBBoxColor);
		}
	}
#if 1
	if (geometry.editRegion.show&BIT_BBOX && geometry.editRegion.size()) {
		for (auto zone : geometry.editRegion.zones()) {
			drawBBox(painter,  zone->bbox(),  geometry.zoneBBoxColor);
			//drawOBBox(painter, zone->obbox(), geometry.zoneBBoxColor);
		}
	}
#endif
	geometry.unlockRead();;
} // drawRegionsBBox


////////////////////////////////////////////////////////////////////////////////////////////////////
// Check if the gradient in the vicinity of a certain pixel is excessive.
// In that case we might want to trigger adaptive super-sampling
// [8/6/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////
bool D3Layer::isGradientExcessive(ZPainter* pDC, int ii, int jj, const Color32& maxGradient)
{
	// handle corner pixels by simply ignoring them
	if (ii < 1 || ii >= pDC->width()-1 || jj < 1 || jj > pDC->height()-1)
		return false;

	// Test for individual pixel gradients
	// TODO: Check if there is a better way to get the individual components
	// of a Color32 via a loop
	for (int c=0; c<3; c++) {
		int Maxi = Max( Color32(pDC->pixel(ii, jj))[c],
				Color32(pDC->pixel(ii, jj+1))[c]);
		if (Color32(pDC->pixel(ii-1, jj  ))[c] > Maxi) Maxi = Color32(pDC->pixel(ii-1, jj  ))[c];
		if (Color32(pDC->pixel(ii+1, jj  ))[c] > Maxi) Maxi = Color32(pDC->pixel(ii+1, jj  ))[c];
		if (Color32(pDC->pixel(ii  , jj-1))[c] > Maxi) Maxi = Color32(pDC->pixel(ii  , jj-1))[c];
		if (Color32(pDC->pixel(ii-1, jj+1))[c] > Maxi) Maxi = Color32(pDC->pixel(ii-1, jj+1))[c];
		if (Color32(pDC->pixel(ii+1, jj+1))[c] > Maxi) Maxi = Color32(pDC->pixel(ii+1, jj+1))[c];
		if (Color32(pDC->pixel(ii-1, jj-1))[c] > Maxi) Maxi = Color32(pDC->pixel(ii-1, jj-1))[c];
		if (Color32(pDC->pixel(ii+1, jj-1))[c] > Maxi) Maxi = Color32(pDC->pixel(ii+1, jj-1))[c];

		int Mini = Min( Color32(pDC->pixel(ii, jj))[c],
				Color32(pDC->pixel(ii, jj+1))[c]);
		if (Color32(pDC->pixel(ii-1, jj  ))[c] < Mini) Mini = Color32(pDC->pixel(ii-1, jj  ))[c];
		if (Color32(pDC->pixel(ii+1, jj  ))[c] < Mini) Mini = Color32(pDC->pixel(ii+1, jj  ))[c];
		if (Color32(pDC->pixel(ii  , jj-1))[c] < Mini) Mini = Color32(pDC->pixel(ii  , jj-1))[c];
		if (Color32(pDC->pixel(ii-1, jj+1))[c] < Mini) Mini = Color32(pDC->pixel(ii-1, jj+1))[c];
		if (Color32(pDC->pixel(ii+1, jj+1))[c] < Mini) Mini = Color32(pDC->pixel(ii+1, jj+1))[c];
		if (Color32(pDC->pixel(ii-1, jj-1))[c] < Mini) Mini = Color32(pDC->pixel(ii-1, jj-1))[c];
		if (Color32(pDC->pixel(ii+1, jj-1))[c] < Mini) Mini = Color32(pDC->pixel(ii+1, jj-1))[c];

		if (Maxi - Mini > maxGradient[c])
			return true;
	}
	return false;
} // isGradientExcessive

////////////////////////////////////////////////////////////////////////////////////////////////////
// run a pass which adaptively performs anti-aliasing based on the existing image data
// [8/7/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////
void D3Layer::drawAdaptiveAAPass(ZPainter& DC)
{
	DUMP(std::cerr << "Adaptive AA started" << std::endl);
	assert(DC.data());	// function must not be called before the image data is available

	Color32   maxContrast(25, 25, 25);
	const int adaptiveSamples = 2;          // could be configurable. for the time being
						// we fix it to 2
						// (the total nr. per pixel is 1+2*adaptiveSamples)
	dword     bkgColor = DC.background();

	int H = DC.height();	// save variables that might change
	int W = DC.width();	// during the scanning...

	// initialize super sampling buffer for all pixels
	DC.resetSuperSamplingBuf();

	// let's start
	int i=0;
	int j=0;
	do {
		if (i==0 or j==0) continue;
		auto ptr = DC.pixelPtr(i,j);

		if (*ptr & (FLAG_IGNORE|FLAG_TRANSPARENT)) continue;

		// CT: the max tolerable gradient could be made configurable but 25
		// (in RGB) works reasonably well in most cases
		if (isGradientExcessive(&DC, i,j, maxContrast) && !DC.isPixelSuperSampled(i, j)) {
			DUMP(std::cerr << "AAA: excessive gradient found for x = "
					<< i << " and y = " << j << std::endl);
			superSamplePixel(i, j, DC, adaptiveSamples);
			// treat adjacent pixels
			if (!DC.isPixelSuperSampled(i+1, j))   superSamplePixel(i+1, j,   DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i-1, j))   superSamplePixel(i-1, j,   DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i,   j-1)) superSamplePixel(i,   j-1, DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i,   j+1)) superSamplePixel(i,   j+1, DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i-1, j+1)) superSamplePixel(i-1, j+1, DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i-1, j-1)) superSamplePixel(i-1, j-1, DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i+1, j+1)) superSamplePixel(i+1, j+1, DC, adaptiveSamples);
			if (!DC.isPixelSuperSampled(i+1, j-1)) superSamplePixel(i+1, j-1, DC, adaptiveSamples);
		}
		if (stop()) return;
	} while (Morton::next2D(i,j, W-1, H-1));

	// transfer super sampling buffer to view buffer
	dword* pSSLine = DC.superSamplingBuf();
	dword* pViewLine = DC.data();
	for (j=0; j<H; j++) {
		dword* pSSPtr = pSSLine;
		dword* pViewPtr = pViewLine;

		// TODO: check if a merging of the complete super-sampling buffer
		// and the image buffer is not faster than a conditional check
		for (i=0; i<W; i++, pSSPtr++, pViewPtr++)
			if (*pSSPtr != bkgColor)
				*pViewPtr = *pSSPtr;

		pSSLine   += W;
		pViewLine += W;
	}

	DUMP(std::cerr << "Adaptive AA finished" << std::endl);
} // drawAdaptiveAAPass

////////////////////////////////////////////////////////////////////////////////////////////////////
// [8/7/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////
void D3Layer::superSamplePixel(const int i, const int j, ZPainter& DC, int nrOfSamples)
{
	Color    accumulatedColor;
	Color32  curPixelColor;
	VZone*   pZone = nullptr;
	Ray      pixelRay;
	double   u, v;
	float    pRandSample[6*6][2], signalFilter = 0.0, totalFilterWeight = 0.0;
	unsigned pScramble[2] = { 2346349678, 1340569164 };

	// Ignore pixels that are labels or other types of drawing
	if (DC.pixel(i,j) & (FLAG_IGNORE|FLAG_TRANSPARENT)) return;

	// visual debugging: mark points identified for super-sampling
	DC.pixel(i,j) |= FLAG_FINAL;
#ifdef STRONG
	DC.setPixelSuperSampled(i, j, Color32(255, 0, 0));
	return;
#endif

	// Check for the default lights
	int fromlight, tolight;
	if (deflights) {
		fromlight = light0 = 0;
		tolight   = geometry.deflights;
	} else {
		fromlight = light0 = geometry.deflights;
		tolight   = geometry.nlights;
	}

	// set up ray
	pixelRay.use_clip      = kernel.clipBodyCount() > 0;
	pixelRay.use_project   = kernel.projectBodyCount() > 0;
	pixelRay.max_depth     = _maxDepth;
	pixelRay.skip_1stblack = skip_1stblack;
	pixelRay.shadow        = shadows;
	pixelRay.nlights       = tolight - fromlight;

	////////////////////////////////////////////////////////////////////////////////////////////
	// multi-threaded super-sampling
#if defined(THREAD) && !defined(SINGLE_AA)
	// PARALLEL adaptive anti-aliasing
	if (kernel.nthreads()) {
		if (kernel.nBodies()) {
			////////////////////////////////////////////////////////////////////////////
			// C++11 threaded version
			//
			// NOTE:
			//   I resorted to this method as my pthread threadpool implementation,
			//   similar to the one used for rendering the scene, always ran into spurious
			//   race issues that I couldn't figure out. Thus, I deleted that code and
			//   rewrote it via the C++11 thread mechanisms.
			//  CT 16/9/2019
			if (nrOfSamples > 6)
				nrOfSamples = 6;

			std::future<dword> results[6 * 6];
			int nrOfThreads = std::min(kernel.nthreads(), nrOfSamples * nrOfSamples);

			// define task
			auto rayShooter = [](D3Layer& d3,
					     double curU, double curV,
					     Ray curRay,
					     GeometryEngine& geoEngine,
					     VZone& curZone) {
				VZone* pzone = &curZone;
				return d3.shootRay(curU, curV, curRay, &geoEngine, &pzone, 1);
			};

			DUMP(std::cerr << "AA-MT: initializing ("
					<< i << ", " << j << ")" << std::endl);
			for (int t=0; t < nrOfThreads; t++) {
				// LD sampling - introducing a bit of variation,
				// that acts similarly to blue noise, increases the AA quality
				CBaseSampler::Sample02(t, pScramble, pRandSample[t]);
				u = view().i2u((double)i + (pRandSample[t][0] - 0.5));
				v = view().j2v((double)j + (pRandSample[t][1] - 0.5));
				DUMP(std::cerr << "AA-MT: threadId = " << t
						<< ", u = " << u << ", v = " << v << std::endl);

				// NOTE: we provide an engine per thread as the kernel provides
				//       a pool of those engine objects for threading. However,
				//       it might be worth to rather make the GeometryEngine
				//       object thread-safe if possible (?)
				results[t] = std::async(std::launch::async | std::launch::deferred,
							rayShooter,
							std::ref(*this),
							u, v,
							pixelRay,
							std::ref(*engine(t)),
							std::ref(*pZone));
			}

			// gather results which are blocking due to the future mechanism
			//std::cerr << "AA-MT: gathering thread results" << std::endl;
			// Don't forget that the central pixel must be taken into account as well!!
			totalFilterWeight = 1;
			// let's start with the original color as a basis
			accumulatedColor = Color(Color32(DC.pixel(i, j)));
			for (int s=0; s < nrOfThreads; s++) {
				// apply DSP filter
				signalFilter = (float)_pixelFilter.Evaluate(
								(double)pRandSample[s][0] - 0.5,
								(double)pRandSample[s][1] - 0.5);
				// we use Color objects instead of Color32 to retain small nuances
				accumulatedColor += (Color(results[s].get()) * signalFilter);
				totalFilterWeight += signalFilter;
				// std::cerr << "AA-MT: ur = " << pRandSample[s][0] - 0.5
				//           << ", us = " << pRandSample[s][1] - 0.5
				//           << ", f = " << signalFilter << std::endl;
			}
			accumulatedColor /= totalFilterWeight;
			//std::cerr << "AA-MT: gathering finished " << std::endl;
			////////////////////////////////////////////////////////////////////////////
		}
	} else
#endif
	////////////////////////////////////////////////////////////////////////////////////////////
	// STANDARD SEQUENTIAL super-sampling
	{
		DUMP(std::cerr << "[" << i << ", " << j << "] -- Before: " << Color32(DC.pixel(i, j)));

		// super sampling loop
		// Don't forget that the central pixel must be taken into account as well!!
		totalFilterWeight = 1;
		// let's start with the original color as a basis
		accumulatedColor = Color(Color32(DC.pixel(i, j)));

		for (int t=0; t<nrOfSamples * nrOfSamples; ++t) {
			// LD sampling - introducing a bit of variation,
			// that acts similarly to blue noise, increases the AA quality
			CBaseSampler::Sample02(t, pScramble, pRandSample[t]);
			u = view().i2u((double)i + (pRandSample[t][0] - 0.5));
			v = view().j2v((double)j + (pRandSample[t][1] - 0.5));

			curPixelColor = shootRay(u, v, pixelRay, engine(), &pZone);
			// apply DSP filter
			signalFilter = (float)_pixelFilter.Evaluate(
							(double)pRandSample[t][0] - 0.5,
							(double)pRandSample[t][1] - 0.5);
			// we use Color objects instead of Color32 to retain small nuances
			accumulatedColor += Color(curPixelColor) * signalFilter;

			totalFilterWeight += signalFilter;

			if (stop()) return;  // bail out if necessary
		} // end of sampling

		// re-normalize accumulated color
		accumulatedColor /= totalFilterWeight;
	}
	////////////////////////////////////////////////////////////////////////////////////////////

	// store the value of the super-sampled pixel in the SS buffer
	DC.setPixelSuperSampled(i, j, accumulatedColor.color32 ());

	DUMP(std::cerr << "\tAfter: " << (Color32)accumulatedColor.color32() << std::endl);
	assert(DC.pixel(i,j) & FLAG_FINAL);
	return;
} // superSamplePixel
