/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	13-Aug-2013
 */

#include "xdraw.h"

#include "geometry/geo.h"
#include "math/matrix4.h"
#include "graph/painter.h"
#include "graph/viewport.h"

//static const char dottedPattern[2] = {1,1};
//static const char dashedPattern[2] = {3,3};

// =================================== XDraw ===================================
/** XDraw */
XDraw::~XDraw()
{
	XFreePixmap(display, transparent25);
	XFreePixmap(display, transparent50);
	XFreePixmap(display, transparent75);
} // XDraw

/** initialize basic structures */
bool XDraw::init(Drawable drawable)
{
	static char stipple_bits25[] = {0x02, 0x00};
	static char stipple_bits50[] = {0x01, 0x02};
	static char stipple_bits75[] = {0x01, 0x03};

	/* Create the stipple for ellipse fill pattern of gc2 */
	if ((transparent25 = XCreateBitmapFromData(display, drawable, stipple_bits25, 2, 2)) == 0)
		return false;
	if ((transparent50 = XCreateBitmapFromData(display, drawable, stipple_bits50, 2, 2)) == 0)
		return false;
	if ((transparent75 = XCreateBitmapFromData(display, drawable, stipple_bits75, 2, 2)) == 0)
		return false;

	return true;
} // init

/** setFillTransparency */
void XDraw::setFillTransparency(int value)
{
	switch (value) {
		case 0:
			XSetFillStyle(display, gc, FillSolid);
			return;
		case 25:
			XSetStipple(display, gc, transparent25);
			break;
		case 50:
			XSetStipple(display, gc, transparent50);
			break;
		case 75:
			XSetStipple(display, gc, transparent75);
			break;
		default:
			assert(0);
			XSetStipple(display, gc, transparent50);
			break;
	}
	XSetFillStyle(display, gc, FillStippled);
} // setFillTransparency

/** draw a pseudo 3D line */
void XDraw::line3D(Drawable drawable,
			const ViewPort& view,
			const Point& a, const Point& b)
{
	int xx1 = view.u2i(a.x);
	int yy1 = view.v2j(a.y);
	int xx2 = view.u2i(b.x);
	int yy2 = view.v2j(b.y);

	// It should take care of the perspective viewports
	//setLineSolid();
	line(drawable, xx1, yy1, xx2, yy2);
#if 0
	if (Eq0(z1,SMALL) && Eq0(z2,SMALL)) {
		setLineSolid();
		setLineWidth(1)
		line(drawable,xx1, yy1, xx2, yy2);
	} else
	if (z1>=-SMALL && z2>=-SMALL) {
		setLineSolid();
		setLineWidth(0)
		line(drawable, xx1, yy1, xx2, yy2);
	} else
	if (z1<=SMALL && z2<=SMALL) {
		setLineDashed():
		setDashes(0, dottedPattern, SIZE(dottedPattern));
		setLineWidth(0);
		line(drawable, xx1, yy1, xx2, yy2);
	} else {
		double ww = z1 / (z2-z1);
		int xx0 = view.u2i(x1 - (x2-x1) * ww);
		int yy0 = view.v2j(y1 - (y2-y1) * ww);

		setLineSolid();
		setLineWidth(0);
		if (z1>=-SMALL) {
			line(drawable, xx1, yy1, xx0, yy0);
			setLineDashed():
			setDashes(0, dottedPattern, SIZE(dottedPattern));
			line(drawable, xx0, yy0, xx2, yy2);
		} else {
			line(drawable, xx2, yy2, xx0, yy0);
			setLineDashed():
			setDashes(0, dottedPattern, SIZE(dottedPattern));
			line(drawable, xx0, yy0, xx1, yy1);
		}
	}
#endif
} // line3d

/** draw a 3D axes system
 * @param x,y		center position of the axes
 * @param size		length of the axes
 * @param matrix	axes system matrix
 * @param xyz		if true draw with RGB else with the
 *			complementary colors
 */
void XDraw::axes(Drawable drawable,
			const int x, const int y, const int size,
			const Matrix4& matrix, bool xyz)
{
	/* X - axis */
	double r = (double)size;
	int xe = x + Round(r * matrix(0,0));
	int ye = y - Round(r * matrix(0,1));
	dword color;
	if (xyz)
		color = matrix(0,2)>=0.0? COLOR_X : COLOR_X_DARK;
	else
		color = matrix(0,2)>=0.0? COLOR_U : COLOR_U_DARK;

	foreground(color);
	line(drawable, x, y, xe, ye);

	/* Y - axis */
	xe = x + Round(r * matrix(1,0));
	ye = y - Round(r * matrix(1,1));
	if (xyz)
		color = matrix(1,2)>=0.0? COLOR_Y : COLOR_Y_DARK;
	else
		color = matrix(1,2)>=0.0? COLOR_V : COLOR_V_DARK;
	foreground(color);
	line(drawable, x, y, xe, ye);

	/* Z - axis */
	xe = x + Round(r * matrix(2,0));
	ye = y - Round(r * matrix(2,1));
	if (xyz)
		color = matrix(2,2)>=0.0? COLOR_Z : COLOR_Z_DARK;
	else
		color = matrix(2,2)>=0.0? COLOR_W : COLOR_W_DARK;
	foreground(color);
	line(drawable, x, y, xe, ye);
} // axes

/** draw a rotated ellipse */
void XDraw::ellipse(Drawable drawable,
			const int x, const int y,
			const int major, const int minor,
			double rotation , double from, double to)
{
#define SEGMENTS	32
	XPoint pts[SEGMENTS];
	double step = (from-to) / (double)(SEGMENTS-1);
	double crot, srot;
	sincos(rotation, &srot, &crot);

	double angle = from;
	for (int i=0; i<SEGMENTS; i++, angle += step) {
		double s, c;
		sincos(angle, &s, &c);
		double xx = (double)major * c;
		double yy = (double)minor * s;
		pts[i].x = x + (int)( xx*crot + yy*srot);
		pts[i].y = y + (int)(-xx*srot + yy*crot);
	}
	lines(drawable, pts, SEGMENTS, CoordModeOrigin);
} // ellipse

/** draw a 3D track ball */
void XDraw::trackball(Drawable drawable,
			const int x, const int y, const int radius,
			const Matrix4& matrix)
{
	dword col;
//	col=0x7F0000;
//	for (int i=0; i<3; i++, col >>= 8) {
//		foreground(col);
//		double angle = atan2(matrix(i,1), matrix(i,0)) + PIover2;
//		ellipse(drawable, x, y, radius, radius*matrix(i,2), angle, PI, PI2);
//	}
	col = 0xFF0000;
	for (int i=0; i<3; i++, col >>= 8) {
		foreground(col);
		double angle = atan2(matrix(i,1), matrix(i,0)) + PIover2;
		ellipse(drawable, x, y, radius, Round(radius*matrix(i,2)), angle, 0.0, PI);
	}

	foreground(0xAF00AF);

	arc(drawable, x-radius, y-radius, 2*radius, 2*radius, 0, 360*64);
	axes(drawable, x, y, radius, matrix);
} // trackball
